@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package example;
 
@@ -60,6 +60,10 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 	protected boolean copy;
 	protected boolean edit;
 	protected boolean annotate;
+	protected boolean form;
+	protected boolean accessibility;
+	protected boolean assemble;
+	protected boolean printHq;
 	protected boolean reflowable;
 	protected boolean linearized;
 	protected int updates;
@@ -69,7 +73,8 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 
 	protected Pixmap pixmap;
 	protected Rect bbox;
-	protected Link[] links;
+	protected Rect[] links;
+	protected String[] linkURIs;
 	protected Quad[][] hits;
 
 	protected int pixmapWidth;
@@ -960,6 +965,14 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 			buffer.append("edit, ");
 		if (annotate)
 			buffer.append("annotate, ");
+		if (form)
+			buffer.append("form, ");
+		if (accessibility)
+			buffer.append("accessibility, ");
+		if (assemble)
+			buffer.append("assemble, ");
+		if (printHq)
+			buffer.append("print-hq, ");
 		if (buffer.length() > 2)
 			buffer.delete(buffer.length() - 2, buffer.length());
 		String permissions = buffer.length() > 0 ? buffer.toString() : null;
@@ -1894,9 +1907,10 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		if (hadOutline)
 			outlinePanel.setVisible(outline != null);
 	}
-	public void onPageContentsChange(Pixmap pixmap, Link[] links, Quad[][] hits) {
+	public void onPageContentsChange(Pixmap pixmap, Rect[] links, String[] linkURIs, Quad[][] hits) {
 		this.pixmap = pixmap;
 		this.links = links;
+		this.linkURIs = linkURIs;
 		this.hits = hits;
 		redraw();
 	}
@@ -1949,11 +1963,15 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		this.format = format;
 		this.encryption = encryption;
 	}
-	public void onPermissionsChange(boolean print, boolean copy, boolean edit, boolean annotate) {
+	public void onPermissionsChange(boolean print, boolean copy, boolean edit, boolean annotate, boolean form, boolean accessibility, boolean assemble, boolean printHq) {
 		this.print = print;
 		this.copy = copy;
 		this.edit = edit;
 		this.annotate = annotate;
+		this.form = form;
+		this.accessibility = accessibility;
+		this.assemble = assemble;
+		this.printHq = printHq;
 	}
 	public void onLinearizedChange(boolean linearized) {
 		this.linearized = linearized;
