@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 /* PDFObject interface */
 
@@ -66,23 +66,6 @@ FUN(PDFObject_isIndirect)(JNIEnv *env, jobject self)
 	return b ? JNI_TRUE : JNI_FALSE;
 }
 
-JNIEXPORT jboolean JNICALL
-FUN(PDFObject_isNull)(JNIEnv *env, jobject self)
-{
-	fz_context *ctx = get_context(env);
-	pdf_obj *obj = from_PDFObject(env, self);
-	int b = 0;
-
-	if (!ctx || !obj) return JNI_FALSE;
-
-	fz_try(ctx)
-		b = pdf_is_null(ctx, obj);
-	fz_catch(ctx)
-		jni_rethrow(env, ctx);
-
-	return b ? JNI_TRUE : JNI_FALSE;
-}
-
 JNIEXPORT jboolean JNICALL
 FUN(PDFObject_isBoolean)(JNIEnv *env, jobject self)
 {
@@ -436,7 +419,9 @@ FUN(PDFObject_resolve)(JNIEnv *env, jobject self)
 	pdf_obj *ind = NULL;
 	jobject jobj;
 
-	if (!ctx || !obj) return NULL;
+	if (!ctx) return NULL;
+
+	if (!obj) return to_PDFObject_safe(ctx, env, NULL);
 
 	fz_try(ctx)
 		ind = pdf_resolve_indirect(ctx, obj);
@@ -457,7 +442,9 @@ FUN(PDFObject_getArray)(JNIEnv *env, jobject self, jint index)
 	pdf_obj *arr = from_PDFObject(env, self);
 	pdf_obj *val = NULL;
 
-	if (!ctx || !arr) return NULL;
+	if (!ctx) return NULL;
+
+	if (!arr) return to_PDFObject_safe(ctx, env, NULL);
 
 	fz_try(ctx)
 		val = pdf_array_get(ctx, arr, index);
@@ -468,21 +455,26 @@ FUN(PDFObject_getArray)(JNIEnv *env, jobject self, jint index)
 }
 
 JNIEXPORT jobject JNICALL
-FUN(PDFObject_getDictionary)(JNIEnv *env, jobject self, jstring jname)
+FUN(PDFObject_getDictionary)(JNIEnv *env, jobject self, jstring jname, jboolean inheritable)
 {
 	fz_context *ctx = get_context(env);
 	pdf_obj *dict = from_PDFObject(env, self);
 	const char *name = NULL;
 	pdf_obj *val = NULL;
 
-	if (!ctx || !dict) return NULL;
+	if (!ctx) return NULL;
 	if (!jname) jni_throw_arg(env, "name must not be null");
 
+	if (!dict) return to_PDFObject_safe(ctx, env, NULL);
+
 	name = (*env)->GetStringUTFChars(env, jname, NULL);
 	if (!name) jni_throw_run(env, "cannot get name to lookup");
 
 	fz_try(ctx)
-		val = pdf_dict_gets(ctx, dict, name);
+		if (inheritable)
+			val = pdf_dict_gets_inheritable(ctx, dict, name);
+		else
+			val = pdf_dict_gets(ctx, dict, name);
 	fz_always(ctx)
 		(*env)->ReleaseStringUTFChars(env, jname, name);
 	fz_catch(ctx)
@@ -498,7 +490,9 @@ FUN(PDFObject_getDictionaryKey)(JNIEnv *env, jobject self, jint index)
 	pdf_obj *dict = from_PDFObject(env, self);
 	pdf_obj *key = NULL;
 
-	if (!ctx || !dict) return NULL;
+	if (!ctx) return NULL;
+
+	if (!dict) return to_PDFObject_safe(ctx, env, NULL);
 
 	fz_try(ctx)
 		key = pdf_dict_get_key(ctx, dict, index);
@@ -1239,3 +1233,26 @@ FUN(PDFObject_toString)(JNIEnv *env, jobject self, jboolean tight, jboolean asci
 
 	return string;
 }
+
+JNIEXPORT jboolean JNICALL
+FUN(PDFObject_equals)(JNIEnv *env, jobject self, jobject jother)
+{
+	fz_context *ctx = get_context(env);
+	pdf_obj *obj = from_PDFObject_safe(env, self);
+	pdf_obj *other = NULL;
+	int result = 0;
+
+	if (!ctx) return JNI_FALSE;
+
+	if (!(*env)->IsInstanceOf(env, jother, cls_PDFObject))
+		return JNI_FALSE;
+
+	other = from_PDFObject_safe(env, jother);
+
+	fz_try(ctx)
+		result = pdf_objcmp(ctx, obj, other);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return result == 0 ? JNI_TRUE : JNI_FALSE;
+}
