@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,11 +17,13 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package com.artifex.mupdf.fitz;
 
+import java.util.Objects;
+
 public class Quad
 {
 	public float ul_x, ul_y;
@@ -40,6 +42,17 @@ public class Quad
 		this.lr_y = lr_y;
 	}
 
+	public Quad(Rect r) {
+		this.ul_x = r.x0;
+		this.ul_y = r.y0;
+		this.ur_x = r.x1;
+		this.ur_y = r.y0;
+		this.ll_x = r.x0;
+		this.ll_y = r.y1;
+		this.lr_x = r.x1;
+		this.lr_y = r.y1;
+	}
+
 	public Rect toRect() {
 		float x0 = Math.min(Math.min(ul_x, ur_x), Math.min(ll_x, lr_x));
 		float y0 = Math.min(Math.min(ul_y, ur_y), Math.min(ll_y, lr_y));
@@ -85,6 +98,31 @@ public class Quad
 		return this;
 	}
 
+	protected boolean triangleContainsPoint(float x, float y, float ax, float ay, float bx, float by, float cx, float cy) {
+		float s, t, area;
+		s = ay * cx - ax * cy + (cy - ay) * x + (ax - cx) * y;
+		t = ax * by - ay * bx + (ay - by) * x + (bx - ax) * y;
+
+		if ((s < 0) != (t < 0))
+			return false;
+
+		area = -by * cx + ay * (cx - bx) + ax * (by - cy) + bx * cy;
+
+		return area < 0 ?
+			(s <= 0 && s + t >= area) :
+			(s >= 0 && s + t <= area);
+	}
+
+	public boolean contains(float x, float y) {
+		return triangleContainsPoint(x, y, ul_x, ul_y, ur_x, ur_y, lr_x, lr_y) ||
+			triangleContainsPoint(x, y, ul_x, ul_y, lr_x, lr_y, ll_x, ll_y);
+
+	}
+
+	public boolean contains(Point p) {
+		return contains(p.x, p.y);
+	}
+
 	public String toString() {
 		return "["
 			+ ul_x + " " + ul_y + " "
@@ -93,4 +131,19 @@ public class Quad
 			+ lr_x + " " + lr_y
 			+ "]";
 	}
+
+	@Override
+	public boolean equals(Object obj) {
+		if (!(obj instanceof Quad))
+			return false;
+		Quad other = (Quad) obj;
+		return ul_x == other.ul_x && ul_y == other.ul_y && ur_x == other.ur_x &&
+			ur_y == other.ur_y && ll_x == other.ll_x && ll_y == other.ll_y &&
+			lr_x == other.lr_x && lr_y == other.lr_y;
+	}
+
+	@Override
+	public int hashCode() {
+		return Objects.hash(ul_x, ul_y, ur_x, ur_y, ll_x, ll_y, lr_x, lr_y);
+	}
 }
