@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package com.artifex.mupdf.fitz;
 
@@ -154,6 +154,11 @@ public class Rect
 		return (x >= x0 && x < x1 && y >= y0 && y < y1);
 	}
 
+	public boolean contains(Point p)
+	{
+		return contains(p.x, p.y);
+	}
+
 	public boolean contains(Rect r)
 	{
 		if (isEmpty() || r.isEmpty())
@@ -169,7 +174,7 @@ public class Rect
 
 	public boolean isValid()
 	{
-		return (x0 <= x1 || y0 <= y1);
+		return (x0 <= x1 && y0 <= y1);
 	}
 
 	public void union(Rect r)
@@ -194,4 +199,41 @@ public class Rect
 		if (r.y1 > y1)
 			y1 = r.y1;
 	}
+
+	public void inset(float dx, float dy) {
+		if (!isValid() || isInfinite() || isEmpty())
+			return;
+		x0 += dx;
+		y0 += dy;
+		x1 -= dx;
+		y1 -= dy;
+	}
+
+	public void inset(float left, float top, float right, float bottom) {
+		if (!isValid() || isInfinite() || isEmpty())
+			return;
+		x0 += left;
+		y0 += top;
+		x1 -= right;
+		y1 -= bottom;
+	}
+
+	public void offset(float dx, float dy) {
+		if (!isValid() || isInfinite() || isEmpty())
+			return;
+		x0 += dx;
+		y0 += dy;
+		x1 += dx;
+		y1 += dy;
+	}
+
+	public void offsetTo(float left, float top) {
+		if (!isValid() || isInfinite() || isEmpty())
+			return;
+		x1 += left - x0;
+		y1 += top - y0;
+		x0 = left;
+		y0 = top;
+	}
+
 }
