@@ -6,17 +6,13 @@
 #include <limits.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <sys/stat.h>
-
-#ifdef _MSC_VER
-#define stat _stat
-#endif
 
 #ifdef _WIN32
 #include <windows.h>
 #include <direct.h> /* for getcwd */
 #else
 #include <unistd.h> /* for getcwd */
+#include <sys/stat.h> /* for mkdir */
 #endif
 
 #define BEYOND_THRESHHOLD 40
@@ -29,30 +25,73 @@
 #define MAX(a,b) ((a) > (b) ? (a) : (b))
 #endif
 
-time_t
-stat_mtime(const char *path)
+static int
+fz_mkdir(char *path)
 {
-	struct stat info;
+#ifdef _WIN32
+	int ret;
+	wchar_t *wpath = fz_wchar_from_utf8(path);
 
-	if (stat(path, &info) < 0)
-		return 0;
+	if (wpath == NULL)
+		return -1;
 
-	return info.st_mtime;
+	ret = _wmkdir(wpath);
+
+	free(wpath);
+
+	return ret;
+#else
+	return mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO);
+#endif
 }
 
-static int convert_to_accel_path(fz_context *ctx, char outname[], char *absname, size_t len)
+static int create_accel_path(fz_context *ctx, char outname[], size_t len, int create, const char *absname, ...)
+{
+	va_list args;
+	char *s = outname;
+	size_t z, remain = len;
+	char *arg;
+
+	va_start(args, absname);
+
+	while ((arg = va_arg(args, char *)) != NULL)
+	{
+		z = fz_snprintf(s, remain, "%s", arg);
+		if (z+1 > remain)
+			goto fail; /* won't fit */
+
+		if (create)
+			(void) fz_mkdir(outname);
+		if (!fz_is_directory(ctx, outname))
+			goto fail; /* directory creation failed, or that dir doesn't exist! */
+#ifdef _WIN32
+		s[z] = '\\';
+#else
+		s[z] = '/';
+#endif
+		s[z+1] = 0;
+		s += z+1;
+		remain -= z+1;
+	}
+
+	if (fz_snprintf(s, remain, "%s.accel", absname) >= remain)
+		goto fail; /* won't fit */
+
+	va_end(args);
+
+	return 1;
+
+fail:
+	va_end(args);
+
+	return 0;
+}
+
+static int convert_to_accel_path(fz_context *ctx, char outname[], char *absname, size_t len, int create)
 {
 	char *tmpdir;
 	char *s;
 
-	tmpdir = getenv("TEMP");
-	if (!tmpdir)
-		tmpdir = getenv("TMP");
-	if (!tmpdir)
-		tmpdir = "/var/tmp";
-	if (!fz_is_directory(ctx, tmpdir))
-		tmpdir = "/tmp";
-
 	if (absname[0] == '/' || absname[0] == '\\')
 		++absname;
 
@@ -63,17 +102,34 @@ static int convert_to_accel_path(fz_context *ctx, char outname[], char *absname,
 		++s;
 	}
 
-	if (fz_snprintf(outname, len, "%s/%s.accel", tmpdir, absname) >= len)
-		return 0;
-	return 1;
+#ifdef _WIN32
+	tmpdir = getenv("USERPROFILE");
+	if (tmpdir && create_accel_path(ctx, outname, len, create, absname, tmpdir, ".config", "mupdf", NULL))
+		return 1; /* OK! */
+	/* TEMP and TMP are user-specific on modern windows. */
+	tmpdir = getenv("TEMP");
+	if (tmpdir && create_accel_path(ctx, outname, len, create, absname, tmpdir, "mupdf", NULL))
+		return 1; /* OK! */
+	tmpdir = getenv("TMP");
+	if (tmpdir && create_accel_path(ctx, outname, len, create, absname, tmpdir, "mupdf", NULL))
+		return 1; /* OK! */
+#else
+	tmpdir = getenv("XDG_CACHE_HOME");
+	if (tmpdir && create_accel_path(ctx, outname, len, create, absname, tmpdir, "mupdf", NULL))
+		return 1; /* OK! */
+	tmpdir = getenv("HOME");
+	if (tmpdir && create_accel_path(ctx, outname, len, create, absname, tmpdir, ".cache", "mupdf", NULL))
+		return 1; /* OK! */
+#endif
+	return 0; /* Fail */
 }
 
-static int get_accelerator_filename(fz_context *ctx, char outname[], size_t len, const char *filename)
+static int get_accelerator_filename(fz_context *ctx, char outname[], size_t len, const char *filename, int create)
 {
 	char absname[PATH_MAX];
 	if (!fz_realpath(filename, absname))
 		return 0;
-	if (!convert_to_accel_path(ctx, outname, absname, len))
+	if (!convert_to_accel_path(ctx, outname, absname, len, create))
 		return 0;
 	return 1;
 }
@@ -86,7 +142,7 @@ static void save_accelerator(fz_context *ctx, fz_document *doc, const char *file
 		return;
 	if (!fz_document_supports_accelerator(ctx, doc))
 		return;
-	if (!get_accelerator_filename(ctx, absname, sizeof(absname), filename))
+	if (!get_accelerator_filename(ctx, absname, sizeof(absname), filename, 1))
 		return;
 
 	fz_save_accelerator(ctx, doc, absname);
@@ -162,7 +218,7 @@ char *pdfapp_version(pdfapp_t *app)
 {
 	return
 		"MuPDF " FZ_VERSION "\n"
-		"Copyright 2006-2020 Artifex Software, Inc.\n";
+		"Copyright 2006-2022 Artifex Software, Inc.\n";
 }
 
 char *pdfapp_usage(pdfapp_t *app)
@@ -412,13 +468,13 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 		}
 		else if (kbps > 0)
 		{
-			fz_stream *stream = fz_open_file_progressive(ctx, filename, kbps, pdfapp_more_data, app);
+			app->stream = fz_open_file_progressive(ctx, filename, kbps, pdfapp_more_data, app);
 			while (1)
 			{
 				fz_try(ctx)
 				{
-					fz_seek(ctx, stream, 0, SEEK_SET);
-					app->doc = fz_open_document_with_stream(ctx, filename, stream);
+					fz_seek(ctx, app->stream, 0, SEEK_SET);
+					app->doc = fz_open_document_with_stream(ctx, filename, app->stream);
 				}
 				fz_catch(ctx)
 				{
@@ -441,12 +497,12 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 			time_t dtime;
 
 			/* If there was an accelerator to load, what would it be called? */
-			if (get_accelerator_filename(ctx, accelpath, sizeof(accelpath), filename))
+			if (get_accelerator_filename(ctx, accelpath, sizeof(accelpath), filename, 0))
 			{
 				/* Check whether that file exists, and isn't older than
 				 * the document. */
-				atime = stat_mtime(accelpath);
-				dtime = stat_mtime(filename);
+				atime = fz_stat_mtime(accelpath);
+				dtime = fz_stat_mtime(filename);
 				if (atime == 0)
 				{
 					/* No accelerator */
@@ -456,7 +512,11 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 				else
 				{
 					/* Accelerator data is out of date */
-					unlink(accelpath);
+#ifdef _WIN32
+					fz_remove_utf8(accelpath);
+#else
+					remove(accelpath);
+#endif
 					accel = NULL; /* In case we have jumped up from below */
 				}
 			}
@@ -502,12 +562,7 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 		}
 
 		app->docpath = fz_strdup(ctx, filename);
-		app->doctitle = filename;
-		if (strrchr(app->doctitle, '\\'))
-			app->doctitle = strrchr(app->doctitle, '\\') + 1;
-		if (strrchr(app->doctitle, '/'))
-			app->doctitle = strrchr(app->doctitle, '/') + 1;
-		app->doctitle = fz_strdup(ctx, app->doctitle);
+		app->doctitle = fz_strdup(ctx, fz_basename(filename));
 
 		fz_layout_document(app->ctx, app->doc, app->layout_w, app->layout_h, app->layout_em);
 
@@ -1120,6 +1175,12 @@ static void pdfapp_gotouri(pdfapp_t *app, char *uri)
 		}
 	}
 
+	if (strncmp(uri, "file://", 7) && strncmp(uri, "http://", 7) && strncmp(uri, "https://", 8) && strncmp(uri, "mailto:", 7))
+	{
+		fz_warn(app->ctx, "refusing to open unknown link (%s)", uri);
+		return;
+	}
+
 	winopenuri(app, uri);
 }
 
@@ -1192,7 +1253,7 @@ static void pdfapp_search_in_direction(pdfapp_t *app, enum panning *panto, int d
 			pdfapp_showpage(app, 1, 0, 0, 0, 1);
 		}
 
-		app->hit_count = fz_search_stext_page(app->ctx, app->page_text, app->search, app->hit_bbox, nelem(app->hit_bbox));
+		app->hit_count = fz_search_stext_page(app->ctx, app->page_text, app->search, NULL, app->hit_bbox, nelem(app->hit_bbox));
 		if (app->hit_count > 0)
 		{
 			*panto = dir == 1 ? PAN_TO_TOP : PAN_TO_BOTTOM;
