@@ -2,17 +2,23 @@
 Support for accessing parse tree for MuPDF headers.
 '''
 
+import os
+import sys
 import time
 
 import jlib
 
+try:
+    import clang
+except ImportError as e:
+    jlib.log( 'Warning, could not import clang: {e}')
+    clang = None
+
 from . import classes
 from . import cpp
 from . import state
 from . import util
 
-clang = state.clang
-
 
 def get_extras(tu, type_):
     '''
@@ -27,6 +33,14 @@ def get_extras(tu, type_):
     extras = classes.classextras.get( tu, base_typename)
     return base_type_cursor, base_typename, extras
 
+def fileline( cursor):
+    '''
+    Returns <file>:<line> from cursor.location.
+    '''
+    f = cursor.location.file
+    filename = os.path.relpath( f.name) if f else ''
+    return f'{filename}:{cursor.location.line}'
+
 
 def prefix( name):
     if name.startswith( 'fz_'):
@@ -46,14 +60,108 @@ def get_fz_extras( tu, fzname):
     ce = classes.classextras.get( tu, fzname)
     return ce
 
+def get_children(cursor):
+    '''
+    Like cursor.get_children() but recurses into cursors with
+    clang.cindex.CursorKind.UNEXPOSED_DECL, which picks up top-level items
+    marked with `extern "C"`.
+    '''
+    verbose = 0
+    for cursor in cursor.get_children():
+        #verbose = state.state_.show_details( cursor.spelling)
+        #verbose = 1
+        if cursor.kind == clang.cindex.CursorKind.UNEXPOSED_DECL:
+            # Things tagged with `extern "C" appear to be within this
+            # cursor.
+            for cursor2 in cursor.get_children():
+                if verbose and cursor.spelling:
+                    jlib.log( '{cursor.spelling=}')
+                yield cursor2
+        elif 1:
+            if verbose and cursor.spelling:
+                jlib.log( '{cursor.spelling=}')
+            yield cursor
+
+def get_members( type_or_cursor, include_empty=False):
+    '''
+    Yields cursor for each member. Uses whichever of
+    clang.cindex.Cursor.get_children() or clang.cindex.Type.get_fields() works.
+
+    Args:
+        type_or_cursor:
+            .
+        include_empty:
+            If false (the default), we first try
+            clang.cindex.Cursor.get_children(), but ignore items for which
+            .spelling==''. If resulting list is empty, we instead use
+            clang.cindex.Type.get_fields().
+
+            Otherwise, we return list of items from
+            clang.cindex.Cursor.get_children(), regardless of whether they
+            have .spelling==''. This allows finding of non-typedef enums, for
+            example.
+    '''
+    if isinstance( type_or_cursor, clang.cindex.Type):
+        cursor = type_or_cursor.get_declaration()
+    elif isinstance( type_or_cursor, clang.cindex.Cursor):
+        cursor = type_or_cursor
+    else:
+        assert 0
+    if cursor.type.kind in (state.clang.cindex.TypeKind.TYPEDEF, state.clang.cindex.TypeKind.ELABORATED):
+        cursor2 = cursor.underlying_typedef_type.get_declaration()
+    else:
+        cursor2 = cursor
+
+    if 0:
+        # Diagnostics to show the difference between
+        # clang.cindex.Cursor.get_children() and
+        # clang.cindex.Type.get_fields().
+        #
+        # For example it looks like clang.cindex.Cursor.get_children() can
+        # return an extra item with .spelling=='' for 'union {...} u;'.
+        #
+        ret_cursor = list()
+        ret_cursor_no_empty = list()
+        ret_type = list()
+        for cursor3 in cursor2.get_children():
+            item = (cursor3.spelling, cursor3.location.file.name, cursor3.location.line)
+            ret_cursor.append( item)
+            if cursor3.spelling:
+                ret_cursor_no_empty.append( item)
+        for cursor3 in cursor.type.get_canonical().get_fields():
+            ret_type.append( (cursor3.spelling, cursor3.location.file.name, cursor3.location.line))
+        ret_cursor.sort()
+        ret_type.sort()
+        ret_cursor_no_empty.sort()
+        if (not ret_cursor_no_empty) and ret_type:
+            jlib.log( 'ret_type and not ret_cursor_no_empty:')
+            for i in ret_type:
+                jlib.log( '    ret_type: {i}')
+        if 0 and ret_cursor != ret_type:
+            jlib.log('get_children() != get_fields():')
+            for i in ret_cursor:
+                jlib.log( '    ret_cursor: {i}')
+            for i in ret_type:
+                jlib.log( '    ret_type: {i}')
+
+    ret = list()
+    for cursor3 in cursor2.get_children():
+        if include_empty or cursor3.spelling:
+            ret.append(cursor3)
+    if not ret:
+        type_ = cursor.type.get_canonical()
+        for cursor3 in type_.get_fields():
+            ret.append( cursor3)
+    for i in ret:
+        yield i
+
 def get_field0( type_):
     '''
     Returns cursor for first field in <type_> or None if <type_> has no fields.
     '''
-    assert isinstance( type_, clang.cindex.Type)
-    type_ = type_.get_canonical()
-    for field in type_.get_fields():
-        return field
+    verbose = state.state_.show_details( type_.spelling)
+    for cursor in get_members(type_):
+        return cursor
 
 get_base_type_cache = dict()
 def get_base_type( type_):
@@ -65,7 +173,7 @@ def get_base_type( type_):
     ret = get_base_type_cache.get( key)
     if ret is None:
         while 1:
-            type_ = type_.get_canonical()
+            type_ = state.get_name_canonical( type_)
             if type_.kind != clang.cindex.TypeKind.POINTER:
                 break
             type_ = type_.get_pointee()
@@ -89,9 +197,9 @@ def is_double_pointer( type_):
     '''
     Returns true if <type_> is double pointer.
     '''
-    type_ = type_.get_canonical()
+    type_ = state.get_name_canonical( type_)
     if type_.kind == clang.cindex.TypeKind.POINTER:
-        type_ = type_.get_pointee().get_canonical()
+        type_ = state.get_name_canonical( type_.get_pointee())
         if type_.kind == clang.cindex.TypeKind.POINTER:
             return True
 
@@ -111,42 +219,62 @@ def has_refs( tu, type_):
 
     key = type_.spelling
     key = util.clip(key, 'struct ')
+    verbose = state.state_.show_details( key)
     ret = has_refs_cache.get( key, None)
     if ret is None:
         ret = False
-        #jlib.log( 'Analysing {type0.spelling=} {type_.spelling=} {key=}')
+        if verbose:
+            jlib.log( 'Analysing {type0.spelling=} {type_.spelling=} {key=}')
 
         for prefix in (
                 'fz_',
                 'pdf_',
                 ):
-            #jlib.log( '{type_.spelling=} {prefix=}')
+            if verbose:
+                jlib.log( '{type_.spelling=} {prefix=}')
             if key.startswith( prefix):
-                #jlib.log( 'Type is a fz_ or pdf_ struct: {key=}')
+                if verbose:
+                    jlib.log( 'Type is a fz_ or pdf_ struct: {key=}')
                 keep_name = f'{prefix}keep_{key[len(prefix):]}'
                 keep_fn_cursor = state.state_.find_function( tu, keep_name, method=False)
-                #jlib.log( '{keep_name=} {keep_fn_cursor=}')
+                if verbose:
+                    jlib.log( '{keep_name=} {keep_fn_cursor=}')
                 if keep_fn_cursor:
-                    #jlib.log( 'There is a keep() fn for this type so it uses reference counting: {keep_name=}')
+                    if verbose:
+                        jlib.log( 'There is a keep() fn for this type so it uses reference counting: {keep_name=}')
                     base_type_cursor = get_base_type( type_).get_declaration()
                     if base_type_cursor.is_definition():
-                        #jlib.log( 'Type definition is available so we look for .refs member: {key=}')
-                        for cursor in type_.get_fields():
+                        if verbose:
+                            jlib.log( 'Type definition is available so we look for .refs member: {key=} {type_.spelling=} {fileline(base_type_cursor)=}')
+                        if verbose:
+                            jlib.log('type_.get_fields()')
+                            for cursor in get_members(type_):
+                                jlib.log('    {cursor.spelling=}')
+                            jlib.log('base_type_cursor.get_children()')
+                            for cursor in base_type_cursor.get_children():
+                                jlib.log('    {cursor.spelling=}')
+                            jlib.log('.')
+                        for cursor in get_members(type_):
                             name = cursor.spelling
-                            type2 = cursor.type.get_canonical()
-                            #jlib.log( '{name=} {type2.spelling=}')
+                            type2 = state.get_name_canonical( cursor.type)
+                            if verbose:
+                                jlib.log( '{name=} {type2.spelling=}')
                             if name == 'refs' and type2.spelling == 'int':
                                 ret = 'refs', 32
                                 break
-                            if name == 'storable' and type2.spelling == 'struct fz_storable':
+                            if name == 'storable' and type2.spelling in ('struct fz_storable', 'fz_storable'):
                                 ret = 'storable.refs', 32
                                 break
                     else:
-                        #jlib.log('Definition is not available for {key=}')
-                        pass
+                        if 0:
+                            jlib.log('Definition is not available for {key=}'
+                                    ' because {base_type_cursor.spelling=} .is_definition()'
+                                    ' returns false.'
+                                    ' base_type_cursor.location={fileline(base_type_cursor)}'
+                                    )
 
                     if not ret:
-                        if 0:
+                        if verbose:
                             jlib.log(
                                     'Cannot find .refs member or we only have forward'
                                     ' declaration, so have to hard-code the size and offset'
@@ -199,7 +327,12 @@ def has_refs( tu, type_):
                                     ' and we have no hard-coded info about size and offset of .regs.'
                                     ' {type0.spelling=} {type_.spelling=} {base_type_cursor.spelling}'
                                     )
-                    assert ret, f'{key} has {keep_name}() but have not found size/location of .refs member.'
+                    assert ret, (
+                            f'{key} has {keep_name}() but have not found size/location of .refs member.'
+                            f' {type_.spelling=}'
+                            f' {base_type_cursor.spelling=}'
+                            f': {fileline(base_type_cursor)}'
+                            )
 
         if type_.spelling in (
                 'struct fz_document',
@@ -250,21 +383,59 @@ def get_text( item, prefix, sep, *names):
         ret.append( f'{name}={value}')
     return prefix + sep.join( ret)
 
-class Clang6FnArgsBug( Exception):
-    def __init__( self, text):
-        Exception.__init__( self, f'clang-6 unable to walk args for fn type. {text}')
 
-def dump_ast( cursor, depth=0):
-    indent = depth*4*' '
-    for cursor2 in cursor.get_children():
-        jlib.log( indent * ' ' + '{cursor2.kind=} {cursor2.mangled_name=} {cursor2.displayname=} {cursor2.spelling=}')
-        dump_ast( cursor2, depth+1)
+def dump_ast( cursor, out=None, depth=0):
+    cleanup = lambda: None
+    if out is None:
+        out = sys.stdout
+    if isinstance(out, str):
+        out = open(out, 'w')
+        cleanup = lambda : out.close()
+    try:
+        indent = depth*4*' '
+        for cursor2 in cursor.get_children():
 
+            def or_none(f):
+                try:
+                    return f()
+                except Exception:
+                    return
+            result = or_none( cursor2.type.get_result)
+            type_ = cursor2.type
+            type_canonical = or_none( cursor2.type.get_canonical)
 
-def show_ast( filename):
+            text = indent
+            text += jlib.log_text(
+                    '{cursor2.kind=}'
+                    ' {cursor2.displayname=}'
+                    ' {cursor2.spelling=}'
+                    ' {cursor2.linkage=}'
+                    ' {cursor2.is_definition()=}'
+                    )
+            if result:
+                text += jlib.log_text(' {result.spelling=}')
+            if type_:
+                text += jlib.log_text(' {type_.spelling=}')
+            if type_canonical:
+                text += jlib.log_text(' {type_canonical.spelling=}')
+            text += '\n'
+            if callable(out):
+                out( text)
+            else:
+                out.write(text)
+
+            dump_ast( cursor2, out, depth+1)
+    finally:
+        cleanup()
+
+def show_ast( filename, includes):
+    jlib.log('Parsing {filename=}')
     index = clang.cindex.Index.create()
+    args = []
+    for include in includes:
+        args += ['-I', include]
     tu = index.parse( filename,
-            args=( '-I', clang_info().include_path),
+            args = args,
             )
     dump_ast( tu.cursor)
 
@@ -327,23 +498,28 @@ def get_args( tu, cursor, include_fz_context=False, skip_first_alt=False, verbos
     # are slow, so we cache the returned items. E.g. this reduces total time of
     # --build 0 from 3.5s to 2.1s.
     #
+    if verbose:
+        jlib.log( '## Looking at args of {cursor.spelling=}')
     key = tu, cursor.location.file, cursor.location.line, include_fz_context, skip_first_alt
     ret = get_args_cache.get( key)
     if not verbose and state.state_.show_details(cursor.spelling):
         verbose = True
-        #jlib.log('Verbose because {cursor.spelling=}')
     if ret is None:
+        if verbose:
+            jlib.log( '## Looking at args of {cursor.spelling=}')
         ret = []
         i = 0
         i_alt = 0
         separator = ''
         for arg_cursor in cursor.get_arguments():
+            if verbose:
+                jlib.log('{arg_cursor.kind=} {arg_cursor.spelling=}')
             assert arg_cursor.kind == clang.cindex.CursorKind.PARM_DECL
             if not include_fz_context and is_pointer_to( arg_cursor.type, 'fz_context'):
                 # Omit this arg because our generated mupdf_*() wrapping functions
                 # use internalContextGet() to get a context.
                 continue
-            name = arg_cursor.mangled_name or f'arg_{i}'
+            name = arg_cursor.spelling or f'arg_{i}'
             if 0 and name == 'stmofsp':
                 verbose = True
             alt = None
@@ -366,6 +542,7 @@ def get_args( tu, cursor, include_fz_context=False, skip_first_alt=False, verbos
                     alt = base_type_cursor
             if verbose:
                 jlib.log( '{arg_cursor.type.spelling=} {base_typename=} {arg_cursor.type.kind=} {get_base_typename(arg_cursor.type)=}')
+                jlib.log( '{get_base_type(arg_cursor.type).kind=}')
             if alt:
                 if is_double_pointer( arg_cursor.type):
                     out_param = True
@@ -382,14 +559,23 @@ def get_args( tu, cursor, include_fz_context=False, skip_first_alt=False, verbos
                         if verbose:
                             jlib.log('setting out_param to true')
                         out_param = True
-            elif base_typename.startswith( ('fz_', 'pdf_')):
+            elif ( base_typename.startswith( ('fz_', 'pdf_'))
+                    and get_base_type(arg_cursor.type).kind != clang.cindex.TypeKind.ENUM
+                    ):
                 # Pointer to fz_ struct is not usually an out-param.
-                if verbose: jlib.log( 'not out-param because arg is: {arg_cursor.displayname=} {base_type.spelling=} {extras}')
+                if verbose:
+                    jlib.log(
+                            'not out-param because pointer to struct:'
+                            ' arg is: {arg_cursor.displayname=}'
+                            ' {base_typename.spelling=}'
+                            ' {extras}'
+                            ' {arg_cursor.type.kind=}'
+                            )
             elif arg_cursor.type.kind == clang.cindex.TypeKind.POINTER:
                 pointee = arg_cursor.type.get_pointee()
                 if verbose:
                     jlib.log( 'clang.cindex.TypeKind.POINTER')
-                if pointee.get_canonical().kind == clang.cindex.TypeKind.FUNCTIONPROTO:
+                if state.get_name_canonical( pointee).kind == clang.cindex.TypeKind.FUNCTIONPROTO:
                     # Don't mark function-pointer args as out-params.
                     if verbose:
                         jlib.log( 'clang.cindex.TypeKind.FUNCTIONPROTO')
@@ -441,6 +627,20 @@ def get_first_arg( tu, cursor):
     return ret, n
 
 
+is_cache = dict()
+
+def is_( type_, type2):
+    key = type_.spelling, type2
+    ret = is_cache.get( key)
+    if ret is None:
+        d = cpp.declaration_text( type_, '', top_level='')
+        d = util.clip( d, 'const ')
+        d = util.clip( d, 'struct ')
+        d = d.strip()
+        ret = (d == type2)
+        is_cache[ key] = ret
+    return ret
+
 is_pointer_to_cache = dict()
 
 def is_pointer_to( type_, destination, verbose=False):
@@ -464,28 +664,31 @@ def is_pointer_to( type_, destination, verbose=False):
     #
     key = type_.spelling, destination
     ret = is_pointer_to_cache.get( key)
-    if ret is None:
+    if verbose or ret is None:
         assert isinstance( type_, clang.cindex.Type)
+        if verbose: jlib.log( '{type_.spelling=}')
         ret = None
         destination = util.clip( destination, 'struct ')
-        if verbose:
-            jlib.log('{type_.kind=}')
         if type_.kind == clang.cindex.TypeKind.POINTER:
-            pointee = type_.get_pointee().get_canonical()
-            d = cpp.declaration_text( pointee, '', top_level='')
+            pointee = type_.get_pointee()
+            if verbose: jlib.log('{pointee.spelling=}')
+            d = cpp.declaration_text( pointee, '', top_level='', verbose=verbose)
             d = util.clip( d, 'const ')
             d = util.clip( d, 'struct ')
             if verbose:
-                jlib.log( '{destination!r=} {d!r=}')
-            ret = d == f'{destination} ' or d == f'const {destination} '
+                jlib.log( '{destination=} {type_.get_pointee().kind=} {type_.get_pointee().spelling=} {state.get_name_canonical( type_.get_pointee()).spelling=}')
+            ret = d.strip() == destination or d.strip() == f'const {destination}'
         is_pointer_to_cache[ key] = ret
 
     return ret
 
-def is_pointer_to_pointer_to( type_, destination):
+def is_pointer_to_pointer_to( type_, destination, verbose=False):
+    if verbose:
+        jlib.log( '{type_.spelling=}')
     if type_.kind != clang.cindex.TypeKind.POINTER:
         return False
-    return is_pointer_to( type_.get_pointee().get_canonical(), destination)
+    pointee = type_.get_pointee()
+    return is_pointer_to( pointee, destination, verbose=verbose)
 
 
 class MethodExcludeReason_VARIADIC:
@@ -553,31 +756,37 @@ def find_wrappable_function_with_arg0_type_cache_populate( tu):
 
         # Look at resulttype.
         #
-        result_type = cursor.type.get_result().get_canonical()
+        result_type = cursor.type.get_result()
         if result_type.kind == clang.cindex.TypeKind.POINTER:
-            result_type = result_type.get_pointee().get_canonical()
-        result_type = util.clip( result_type.spelling, 'struct ')
-        if result_type.startswith( ('fz_', 'pdf_')):
-            result_type_extras = get_fz_extras( tu, result_type)
-            if not result_type_extras:
-                exclude_reasons.append(
-                        (
-                        MethodExcludeReason_NO_EXTRAS,
-                        f'no extras defined for result_type={result_type}',
-                        ))
-            else:
-                if not result_type_extras.constructor_raw:
+            result_type = result_type.get_pointee()
+        result_type_name = state.get_name_canonical( result_type)
+        result_type_name = util.clip( result_type.spelling, 'struct ')
+        if result_type_name.startswith( ('fz_', 'pdf_')):
+            if result_type.kind == clang.cindex.TypeKind.TYPEDEF:
+                result_cursor = result_type.get_declaration()
+                result_type = result_cursor.underlying_typedef_type
+
+            if result_type.kind == state.clang.cindex.TypeKind.ELABORATED:
+                result_type_extras = get_fz_extras( tu, result_type_name)
+                if not result_type_extras:
                     exclude_reasons.append(
                             (
-                            MethodExcludeReason_NO_RAW_CONSTRUCTOR,
-                            f'wrapper for result_type={result_type} does not have raw constructor.',
-                            ))
-                if not result_type_extras.copyable:
-                    exclude_reasons.append(
-                            (
-                            MethodExcludeReason_NOT_COPYABLE,
-                                f'wrapper for result_type={result_type} is not copyable.',
+                            MethodExcludeReason_NO_EXTRAS,
+                            f'no extras defined for result_type={result_type_name}.'
                             ))
+                else:
+                    if not result_type_extras.constructor_raw:
+                        exclude_reasons.append(
+                                (
+                                MethodExcludeReason_NO_RAW_CONSTRUCTOR,
+                                f'wrapper for result_type={result_type_name} does not have raw constructor.',
+                                ))
+                    if not result_type_extras.copyable:
+                        exclude_reasons.append(
+                                (
+                                MethodExcludeReason_NOT_COPYABLE,
+                                f'wrapper for result_type={result_type_name} is not copyable.',
+                                ))
 
         # Look at args
         #
@@ -587,7 +796,8 @@ def find_wrappable_function_with_arg0_type_cache_populate( tu):
 
             base_typename = get_base_typename( arg.cursor.type)
             if not arg.alt and base_typename.startswith( ('fz_', 'pdf_')):
-                if arg.cursor.type.get_canonical().kind == clang.cindex.TypeKind.ENUM:
+                t_canonical = state.get_name_canonical( arg.cursor.type)
+                if t_canonical.kind == clang.cindex.TypeKind.ENUM:
                     # We don't (yet) wrap fz_* enums, but for now at least we
                     # still wrap functions that take fz_* enum parameters -
                     # callers will have to use the fz_* type.
@@ -595,12 +805,22 @@ def find_wrappable_function_with_arg0_type_cache_populate( tu):
                     # For example this is required by mutool_draw.py because
                     # mudraw.c calls fz_set_separation_behavior().
                     #
-                    jlib.logx( 'not excluding {fnname=} with enum fz_ param : {arg.cursor.spelling=} {arg.cursor.type.kind} {arg.cursor.type.get_canonical().kind=}')
+                    jlib.logx(
+                            'not excluding {fnname=} with enum fz_ param:'
+                            ' {arg.cursor.spelling=}'
+                            ' {arg.cursor.type.kind}'
+                            ' {state.get_name_canonical(arg.cursor.type).kind=}'
+                            )
+                elif t_canonical.kind == clang.cindex.TypeKind.POINTER:
+                    pass
                 else:
                     exclude_reasons.append(
                             (
                             MethodExcludeReason_NO_WRAPPER_CLASS,
-                            f'no wrapper class for arg i={i}: {arg.cursor.type.get_canonical().spelling} {arg.cursor.type.get_canonical().kind}',
+                            f'no wrapper class for arg i={i}:'
+                                f' {state.get_name_canonical( arg.cursor.type).spelling}'
+                                f' {state.get_name_canonical(arg.cursor.type).kind}'
+                                ,
                             ))
             if i == 0:
                 if arg.alt:
@@ -623,7 +843,7 @@ def find_wrappable_function_with_arg0_type_cache_populate( tu):
         else:
             if i > 0:
                 # <fnname> is ok to wrap.
-                arg0 = arg0_cursor.type.get_canonical().spelling
+                arg0 = state.get_name_canonical( arg0_cursor.type).spelling
                 arg0 = util.clip( arg0, 'struct ')
 
                 #jlib.log( '=== Adding to {arg0=}: {fnname=}. {len(fnname_to_method_structname)=}')
@@ -672,6 +892,9 @@ def find_class_for_wrappable_function( fn_name):
 def find_struct( tu, structname, require_definition=True):
     '''
     Finds definition of struct.
+
+    fixme: actually finds definition of anything, doesn't have to be a struct.
+
     Args:
         tu:
             Translation unit.
@@ -683,6 +906,7 @@ def find_struct( tu, structname, require_definition=True):
     Returns cursor for definition or None.
     '''
     verbose = state.state_.show_details( structname)
+    verbose = False
     if verbose:
         jlib.log( '{=structname}')
     structname = util.clip( structname, ('const ', 'struct '))   # Remove any 'struct ' prefix.
@@ -691,7 +915,7 @@ def find_struct( tu, structname, require_definition=True):
     global find_struct_cache
     if find_struct_cache is None:
         find_struct_cache = dict()
-        for cursor in tu.cursor.get_children():
+        for cursor in get_children( tu.cursor):
             already = find_struct_cache.get( cursor.spelling)
             if already is None:
                 find_struct_cache[ cursor.spelling] = cursor
@@ -720,12 +944,13 @@ def find_name( cursor, name, nest=0):
     cursor:
         Item to search.
     '''
+    assert cursor.spelling != ''
     if cursor.spelling == '':
         # Anonymous item; this seems to occur for (non-anonymous) unions.
         #
         # We recurse into children directly.
         #
-        for c in cursor.get_children():
+        for c in get_members(cursor):
             ret = find_name_internal( c, name, nest+1)
             if ret:
                 return ret
@@ -740,7 +965,7 @@ def find_name( cursor, name, nest=0):
         ret = find_name( c, tail, nest+2)
         return ret
 
-    for c in cursor.type.get_canonical().get_fields():
+    for c in get_members(cursor):
         if c.spelling == '':
             ret = find_name( c, name, nest+1)
             if ret:
