@@ -1,4 +1,26 @@
-#include "fitz-imp.h"
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
+#include "mupdf/fitz.h"
 
 #include <string.h>
 #include <stdarg.h>
@@ -14,7 +36,7 @@ fz_new_buffer(fz_context *ctx, size_t size)
 	b->refs = 1;
 	fz_try(ctx)
 	{
-		b->data = fz_malloc(ctx, size);
+		b->data = Memento_label(fz_malloc(ctx, size), "fz_buffer_data");
 	}
 	fz_catch(ctx)
 	{
@@ -28,19 +50,6 @@ fz_new_buffer(fz_context *ctx, size_t size)
 	return b;
 }
 
-/*
-	Create a new buffer with existing data.
-
-	data: Pointer to existing data.
-	size: Size of existing data.
-
-	Takes ownership of data. Does not make a copy. Calls fz_free on the
-	data when the buffer is deallocated. Do not use 'data' after passing
-	to this function.
-
-	Returns pointer to new buffer. Throws exception on allocation
-	failure.
-*/
 fz_buffer *
 fz_new_buffer_from_data(fz_context *ctx, unsigned char *data, size_t size)
 {
@@ -64,9 +73,6 @@ fz_new_buffer_from_data(fz_context *ctx, unsigned char *data, size_t size)
 	return b;
 }
 
-/*
-	Like fz_new_buffer, but does not take ownership.
-*/
 fz_buffer *
 fz_new_buffer_from_shared_data(fz_context *ctx, const unsigned char *data, size_t size)
 {
@@ -83,9 +89,6 @@ fz_new_buffer_from_shared_data(fz_context *ctx, const unsigned char *data, size_
 	return b;
 }
 
-/*
-	Create a new buffer containing a copy of the passed data.
-*/
 fz_buffer *
 fz_new_buffer_from_copied_data(fz_context *ctx, const unsigned char *data, size_t size)
 {
@@ -95,38 +98,96 @@ fz_new_buffer_from_copied_data(fz_context *ctx, const unsigned char *data, size_
 	return b;
 }
 
-/*
-	Create a new buffer with data decoded from a base64 input string.
-*/
+fz_buffer *fz_clone_buffer(fz_context *ctx, fz_buffer *buf)
+{
+	return fz_new_buffer_from_copied_data(ctx, buf ? buf->data : NULL, buf ? buf->len : 0);
+}
+
+static inline int iswhite(int a)
+{
+	switch (a) {
+	case '\n': case '\r': case '\t': case ' ':
+	case '\f':
+		return 1;
+	}
+	return 0;
+}
+
 fz_buffer *
 fz_new_buffer_from_base64(fz_context *ctx, const char *data, size_t size)
 {
-	fz_buffer *buf = fz_new_buffer(ctx, size);
+	fz_buffer *out = fz_new_buffer(ctx, size > 0 ? size : strlen(data));
 	const char *end = data + (size > 0 ? size : strlen(data));
 	const char *s = data;
+	uint32_t buf = 0;
+	int bits = 0;
+
+	/* This is https://infra.spec.whatwg.org/#forgiving-base64-decode
+	 * but even more relaxed. We allow any number of trailing '=' code
+	 * points and instead of returning failure on invalid characters, we
+	 * warn and truncate.
+	 */
+
+	while (s < end && iswhite(*s))
+		s++;
+	while (s < end && iswhite(end[-1]))
+		end--;
+	while (s < end && end[-1] == '=')
+		end--;
+
 	fz_try(ctx)
 	{
 		while (s < end)
 		{
 			int c = *s++;
+
 			if (c >= 'A' && c <= 'Z')
-				fz_append_bits(ctx, buf, c - 'A', 6);
+				c = c - 'A';
 			else if (c >= 'a' && c <= 'z')
-				fz_append_bits(ctx, buf, c - 'a' + 26, 6);
+				c = c - 'a' + 26;
 			else if (c >= '0' && c <= '9')
-				fz_append_bits(ctx, buf, c - '0' + 52, 6);
+				c = c - '0' + 52;
 			else if (c == '+')
-				fz_append_bits(ctx, buf, 62, 6);
+				c = 62;
 			else if (c == '/')
-				fz_append_bits(ctx, buf, 63, 6);
+				c = 63;
+			else if (iswhite(c))
+				continue;
+			else
+			{
+				fz_warn(ctx, "invalid character in base64");
+				break;
+			}
+
+			buf <<= 6;
+			buf |= c & 0x3f;
+			bits += 6;
+
+			if (bits == 24)
+			{
+				fz_append_byte(ctx, out, buf >> 16);
+				fz_append_byte(ctx, out, buf >> 8);
+				fz_append_byte(ctx, out, buf >> 0);
+				bits = 0;
+			}
+		}
+
+		if (bits == 18)
+		{
+			fz_append_byte(ctx, out, buf >> 10);
+			fz_append_byte(ctx, out, buf >> 2);
+		}
+		else if (bits == 12)
+		{
+			fz_append_byte(ctx, out, buf >> 4);
 		}
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_buffer(ctx, buf);
+		fz_drop_buffer(ctx, out);
 		fz_rethrow(ctx);
 	}
-	return buf;
+	return out;
 }
 
 fz_buffer *
@@ -146,13 +207,6 @@ fz_drop_buffer(fz_context *ctx, fz_buffer *buf)
 	}
 }
 
-/*
-	Ensure that a buffer has a given capacity,
-	truncating data if required.
-
-	capacity: The desired capacity for the buffer. If the current size
-	of the buffer contents is smaller than capacity, it is truncated.
-*/
 void
 fz_resize_buffer(fz_context *ctx, fz_buffer *buf, size_t size)
 {
@@ -164,10 +218,6 @@ fz_resize_buffer(fz_context *ctx, fz_buffer *buf, size_t size)
 		buf->len = buf->cap;
 }
 
-/*
-	Make some space within a buffer (i.e. ensure that
-	capacity > size).
-*/
 void
 fz_grow_buffer(fz_context *ctx, fz_buffer *buf)
 {
@@ -190,9 +240,6 @@ fz_ensure_buffer(fz_context *ctx, fz_buffer *buf, size_t min)
 	fz_resize_buffer(ctx, buf, newsize);
 }
 
-/*
-	Trim wasted capacity from a buffer by resizing internal memory.
-*/
 void
 fz_trim_buffer(fz_context *ctx, fz_buffer *buf)
 {
@@ -206,12 +253,6 @@ fz_clear_buffer(fz_context *ctx, fz_buffer *buf)
 	buf->len = 0;
 }
 
-/*
-	Zero-terminate buffer in order to use as a C string.
-
-	This byte is invisible and does not affect the length of the buffer as returned by fz_buffer_storage.
-	The zero byte is written *after* the data, and subsequent writes will overwrite the terminating byte.
-*/
 void
 fz_terminate_buffer(fz_context *ctx, fz_buffer *buf)
 {
@@ -221,13 +262,6 @@ fz_terminate_buffer(fz_context *ctx, fz_buffer *buf)
 	buf->data[buf->len] = 0;
 }
 
-/*
-	Retrieve internal memory of buffer.
-
-	datap: Output parameter that will be pointed to the data.
-
-	Returns the current size of the data in bytes.
-*/
 size_t
 fz_buffer_storage(fz_context *ctx, fz_buffer *buf, unsigned char **datap)
 {
@@ -236,10 +270,6 @@ fz_buffer_storage(fz_context *ctx, fz_buffer *buf, unsigned char **datap)
 	return (buf ? buf->len : 0);
 }
 
-/*
-	Ensure that a buffer's data ends in a
-	0 byte, and return a pointer to it.
-*/
 const char *
 fz_string_from_buffer(fz_context *ctx, fz_buffer *buf)
 {
@@ -249,19 +279,6 @@ fz_string_from_buffer(fz_context *ctx, fz_buffer *buf)
 	return (const char *)buf->data;
 }
 
-/*
-	Take ownership of buffer contents.
-	Performs the same task as fz_buffer_storage, but ownership of
-	the data buffer returns with this call. The buffer is left
-	empty.
-
-	Note: Bad things may happen if this is called on a buffer with
-	multiple references that is being used from multiple threads.
-
-	data: Pointer to place to retrieve data pointer.
-
-	Returns length of stream.
-*/
 size_t
 fz_buffer_extract(fz_context *ctx, fz_buffer *buf, unsigned char **datap)
 {
@@ -276,6 +293,27 @@ fz_buffer_extract(fz_context *ctx, fz_buffer *buf, unsigned char **datap)
 	return len;
 }
 
+fz_buffer *
+fz_slice_buffer(fz_context *ctx, fz_buffer *buf, int64_t start, int64_t end)
+{
+	unsigned char *src = NULL;
+	size_t size = fz_buffer_storage(ctx, buf, &src);
+	size_t s, e;
+
+	if (start < 0)
+		start += size;
+	if (end < 0)
+		end += size;
+
+	s = fz_clamp64(start, 0, size);
+	e = fz_clamp64(end, 0, size);
+
+	if (s == size || e <= s)
+		return fz_new_buffer(ctx, 0);
+
+	return fz_new_buffer_from_copied_data(ctx, &src[s], e - s);
+}
+
 void
 fz_append_buffer(fz_context *ctx, fz_buffer *buf, fz_buffer *extra)
 {
@@ -289,12 +327,6 @@ fz_append_buffer(fz_context *ctx, fz_buffer *buf, fz_buffer *extra)
 	buf->len += extra->len;
 }
 
-/*
-	fz_append_*: Append data to a buffer.
-	fz_append_printf: Format and append data to buffer using printf-like formatting (see fz_vsnprintf).
-	fz_append_pdf_string: Append a string with PDF syntax quotes and escapes.
-	The buffer will automatically grow as required.
-*/
 void
 fz_append_data(fz_context *ctx, fz_buffer *buf, const void *data, size_t len)
 {
