@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -60,6 +60,8 @@ fz_disable_device(fz_context *ctx, fz_device *dev)
 	dev->set_default_colorspaces = NULL;
 	dev->begin_layer = NULL;
 	dev->end_layer = NULL;
+	dev->begin_structure = NULL;
+	dev->end_structure = NULL;
 }
 
 void
@@ -345,6 +347,8 @@ fz_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix ctm, f
 void
 fz_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
+	if (image->colorspace == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "argument to fill image must be a color image");
 	if (dev->fill_image)
 	{
 		fz_try(ctx)
@@ -566,6 +570,62 @@ void fz_end_layer(fz_context *ctx, fz_device *dev)
 	}
 }
 
+void fz_begin_structure(fz_context *ctx, fz_device *dev, fz_structure str, const char *raw, int uid)
+{
+	if (dev->begin_structure)
+	{
+		fz_try(ctx)
+			dev->begin_structure(ctx, dev, str, raw, uid);
+		fz_catch(ctx)
+		{
+			fz_disable_device(ctx, dev);
+			fz_rethrow(ctx);
+		}
+	}
+}
+
+void fz_end_structure(fz_context *ctx, fz_device *dev)
+{
+	if (dev->end_structure)
+	{
+		fz_try(ctx)
+			dev->end_structure(ctx, dev);
+		fz_catch(ctx)
+		{
+			fz_disable_device(ctx, dev);
+			fz_rethrow(ctx);
+		}
+	}
+}
+
+void fz_begin_metatext(fz_context *ctx, fz_device *dev, fz_metatext meta, const char *meta_text)
+{
+	if (dev->begin_metatext)
+	{
+		fz_try(ctx)
+			dev->begin_metatext(ctx, dev, meta, meta_text);
+		fz_catch(ctx)
+		{
+			fz_disable_device(ctx, dev);
+			fz_rethrow(ctx);
+		}
+	}
+}
+
+void fz_end_metatext(fz_context *ctx, fz_device *dev)
+{
+	if (dev->end_metatext)
+	{
+		fz_try(ctx)
+			dev->end_metatext(ctx, dev);
+		fz_catch(ctx)
+		{
+			fz_disable_device(ctx, dev);
+			fz_rethrow(ctx);
+		}
+	}
+}
+
 fz_rect
 fz_device_current_scissor(fz_context *ctx, fz_device *dev)
 {
@@ -573,3 +633,205 @@ fz_device_current_scissor(fz_context *ctx, fz_device *dev)
 		return dev->container[dev->container_len-1].scissor;
 	return fz_infinite_rect;
 }
+
+const char *
+fz_structure_to_string(fz_structure type)
+{
+	switch (type)
+	{
+	default:
+		return "Invalid";
+	case FZ_STRUCTURE_DOCUMENT:
+		return "Document";
+	case FZ_STRUCTURE_PART:
+		return "Part";
+	case FZ_STRUCTURE_ART:
+		return "Art";
+	case FZ_STRUCTURE_SECT:
+		return "Sect";
+	case FZ_STRUCTURE_DIV:
+		return "Div";
+	case FZ_STRUCTURE_BLOCKQUOTE:
+		return "BlockQuote";
+	case FZ_STRUCTURE_CAPTION:
+		return "Caption";
+	case FZ_STRUCTURE_TOC:
+		return "TOC";
+	case FZ_STRUCTURE_TOCI:
+		return "TOCI";
+	case FZ_STRUCTURE_INDEX:
+		return "Index";
+	case FZ_STRUCTURE_NONSTRUCT:
+		return "NonDtruct";
+	case FZ_STRUCTURE_PRIVATE:
+		return "Private";
+	/* Grouping elements (PDF 2.0 - Table 364) */
+	case FZ_STRUCTURE_DOCUMENTFRAGMENT:
+		return "DocumentFragment";
+	/* Grouping elements (PDF 2.0 - Table 365) */
+	case FZ_STRUCTURE_ASIDE:
+		return "Aside";
+	/* Grouping elements (PDF 2.0 - Table 366) */
+	case FZ_STRUCTURE_TITLE:
+		return "Title";
+	case FZ_STRUCTURE_FENOTE:
+		return "FENote";
+	/* Grouping elements (PDF 2.0 - Table 367) */
+	case FZ_STRUCTURE_SUB:
+		return "Sub";
+
+	/* Paragraphlike elements (PDF 1.7 - Table 10.21) */
+	case FZ_STRUCTURE_P:
+		return "P";
+	case FZ_STRUCTURE_H:
+		return "H";
+	case FZ_STRUCTURE_H1:
+		return "H1";
+	case FZ_STRUCTURE_H2:
+		return "H2";
+	case FZ_STRUCTURE_H3:
+		return "H3";
+	case FZ_STRUCTURE_H4:
+		return "H4";
+	case FZ_STRUCTURE_H5:
+		return "H5";
+	case FZ_STRUCTURE_H6:
+		return "H6";
+
+	/* List elements (PDF 1.7 - Table 10.23) */
+	case FZ_STRUCTURE_LIST:
+		return "List";
+	case FZ_STRUCTURE_LISTITEM:
+		return "LI";
+	case FZ_STRUCTURE_LABEL:
+		return "Lbl";
+	case FZ_STRUCTURE_LISTBODY:
+		return "LBody";
+
+	/* Table elements (PDF 1.7 - Table 10.24) */
+	case FZ_STRUCTURE_TABLE:
+		return "Table";
+	case FZ_STRUCTURE_TR:
+		return "TR";
+	case FZ_STRUCTURE_TH:
+		return "TH";
+	case FZ_STRUCTURE_TD:
+		return "TD";
+	case FZ_STRUCTURE_THEAD:
+		return "THead";
+	case FZ_STRUCTURE_TBODY:
+		return "TBody";
+	case FZ_STRUCTURE_TFOOT:
+		return "TFoot";
+
+	/* Inline elements (PDF 1.7 - Table 10.25) */
+	case FZ_STRUCTURE_SPAN:
+		return "Span";
+	case FZ_STRUCTURE_QUOTE:
+		return "Quote";
+	case FZ_STRUCTURE_NOTE:
+		return "Note";
+	case FZ_STRUCTURE_REFERENCE:
+		return "Reference";
+	case FZ_STRUCTURE_BIBENTRY:
+		return "BibEntry";
+	case FZ_STRUCTURE_CODE:
+		return "Code";
+	case FZ_STRUCTURE_LINK:
+		return "Link";
+	case FZ_STRUCTURE_ANNOT:
+		return "Annot";
+	/* Inline elements (PDF 2.0 - Table 368) */
+	case FZ_STRUCTURE_EM:
+		return "Em";
+	case FZ_STRUCTURE_STRONG:
+		return "Strong";
+
+	/* Ruby inline element (PDF 1.7 - Table 10.26) */
+	case FZ_STRUCTURE_RUBY:
+		return "Ruby";
+	case FZ_STRUCTURE_RB:
+		return "RB";
+	case FZ_STRUCTURE_RT:
+		return "RT";
+	case FZ_STRUCTURE_RP:
+		return "RP";
+
+	/* Warichu inline element (PDF 1.7 - Table 10.26) */
+	case FZ_STRUCTURE_WARICHU:
+		return "Warichu";
+	case FZ_STRUCTURE_WT:
+		return "WT";
+	case FZ_STRUCTURE_WP:
+		return "WP";
+
+	/* Illustration elements (PDF 1.7 - Table 10.27) */
+	case FZ_STRUCTURE_FIGURE:
+		return "Figure";
+	case FZ_STRUCTURE_FORMULA:
+		return "Formula";
+	case FZ_STRUCTURE_FORM:
+		return "Form";
+
+	/* Artifact structure type (PDF 2.0 - Table 375) */
+	case FZ_STRUCTURE_ARTIFACT:
+		return "Artifact";
+	}
+
+	return NULL;
+}
+
+fz_structure
+fz_structure_from_string(const char *str)
+{
+	if (!strcmp(str, "Document")) return FZ_STRUCTURE_DOCUMENT;
+	if (!strcmp(str, "Part")) return FZ_STRUCTURE_PART;
+	if (!strcmp(str, "Art")) return FZ_STRUCTURE_ART;
+	if (!strcmp(str, "Sect")) return FZ_STRUCTURE_SECT;
+	if (!strcmp(str, "Div")) return FZ_STRUCTURE_DIV;
+	if (!strcmp(str, "BlockQuote")) return FZ_STRUCTURE_BLOCKQUOTE;
+	if (!strcmp(str, "Caption")) return FZ_STRUCTURE_CAPTION;
+	if (!strcmp(str, "TOC")) return FZ_STRUCTURE_TOC;
+	if (!strcmp(str, "TOCI")) return FZ_STRUCTURE_TOCI;
+	if (!strcmp(str, "Index")) return FZ_STRUCTURE_INDEX;
+	if (!strcmp(str, "NonStruct")) return FZ_STRUCTURE_NONSTRUCT;
+	if (!strcmp(str, "Private")) return FZ_STRUCTURE_PRIVATE;
+	if (!strcmp(str, "P")) return FZ_STRUCTURE_P;
+	if (!strcmp(str, "H")) return FZ_STRUCTURE_H;
+	if (!strcmp(str, "H1")) return FZ_STRUCTURE_H1;
+	if (!strcmp(str, "H2")) return FZ_STRUCTURE_H2;
+	if (!strcmp(str, "H3")) return FZ_STRUCTURE_H3;
+	if (!strcmp(str, "H4")) return FZ_STRUCTURE_H4;
+	if (!strcmp(str, "H5")) return FZ_STRUCTURE_H5;
+	if (!strcmp(str, "H6")) return FZ_STRUCTURE_H6;
+	if (!strcmp(str, "L")) return FZ_STRUCTURE_LIST;
+	if (!strcmp(str, "LI")) return FZ_STRUCTURE_LISTITEM;
+	if (!strcmp(str, "Lbl")) return FZ_STRUCTURE_LABEL;
+	if (!strcmp(str, "LBody")) return FZ_STRUCTURE_LISTBODY;
+	if (!strcmp(str, "Table")) return FZ_STRUCTURE_TABLE;
+	if (!strcmp(str, "TR")) return FZ_STRUCTURE_TR;
+	if (!strcmp(str, "TH")) return FZ_STRUCTURE_TH;
+	if (!strcmp(str, "TD")) return FZ_STRUCTURE_TD;
+	if (!strcmp(str, "THead")) return FZ_STRUCTURE_THEAD;
+	if (!strcmp(str, "TBody")) return FZ_STRUCTURE_TBODY;
+	if (!strcmp(str, "TFoot")) return FZ_STRUCTURE_TFOOT;
+	if (!strcmp(str, "Span")) return FZ_STRUCTURE_SPAN;
+	if (!strcmp(str, "Quote")) return FZ_STRUCTURE_QUOTE;
+	if (!strcmp(str, "Note")) return FZ_STRUCTURE_NOTE;
+	if (!strcmp(str, "Reference")) return FZ_STRUCTURE_REFERENCE;
+	if (!strcmp(str, "BibEntry")) return FZ_STRUCTURE_BIBENTRY;
+	if (!strcmp(str, "Code")) return FZ_STRUCTURE_CODE;
+	if (!strcmp(str, "Link")) return FZ_STRUCTURE_LINK;
+	if (!strcmp(str, "Annot")) return FZ_STRUCTURE_ANNOT;
+	if (!strcmp(str, "Ruby")) return FZ_STRUCTURE_RUBY;
+	if (!strcmp(str, "RB")) return FZ_STRUCTURE_RB;
+	if (!strcmp(str, "RT")) return FZ_STRUCTURE_RT;
+	if (!strcmp(str, "RP")) return FZ_STRUCTURE_RP;
+	if (!strcmp(str, "Warichu")) return FZ_STRUCTURE_WARICHU;
+	if (!strcmp(str, "WT")) return FZ_STRUCTURE_WT;
+	if (!strcmp(str, "WP")) return FZ_STRUCTURE_WP;
+	if (!strcmp(str, "Figure")) return FZ_STRUCTURE_FIGURE;
+	if (!strcmp(str, "Formula")) return FZ_STRUCTURE_FORMULA;
+	if (!strcmp(str, "Form")) return FZ_STRUCTURE_FORM;
+	return FZ_STRUCTURE_INVALID;
+}
