@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -26,7 +26,7 @@
 
 enum
 {
-	FZ_DOCUMENT_HANDLER_MAX = 10
+	FZ_DOCUMENT_HANDLER_MAX = 32
 };
 
 #define DEFW (450)
@@ -88,11 +88,11 @@ void fz_register_document_handler(fz_context *ctx, const fz_document_handler *ha
 }
 
 const fz_document_handler *
-fz_recognize_document(fz_context *ctx, const char *magic)
+fz_recognize_document_stream_content(fz_context *ctx, fz_stream *stream, const char *magic)
 {
 	fz_document_handler_context *dc;
 	int i, best_score, best_i;
-	const char *ext, *needle;
+	const char *ext;
 
 	dc = ctx->handler;
 	if (dc->count == 0)
@@ -100,42 +100,64 @@ fz_recognize_document(fz_context *ctx, const char *magic)
 
 	ext = strrchr(magic, '.');
 	if (ext)
-		needle = ext + 1;
+		ext = ext + 1;
 	else
-		needle = magic;
+		ext = magic;
 
 	best_score = 0;
 	best_i = -1;
 
-	for (i = 0; i < dc->count; i++)
+	if (stream && stream->seek != NULL)
 	{
-		int score = 0;
-		const char **entry;
-
-		if (dc->handler[i]->recognize)
-			score = dc->handler[i]->recognize(ctx, magic);
-
-		if (!ext)
+		for (i = 0; i < dc->count; i++)
 		{
+			int score = 0;
+
+			fz_seek(ctx, stream, 0, SEEK_SET);
+
+			if (dc->handler[i]->recognize_content)
+				score = dc->handler[i]->recognize_content(ctx, stream);
+
+			if (best_score < score)
+			{
+				best_score = score;
+				best_i = i;
+			}
+		}
+	}
+
+	if (best_score < 100)
+	{
+		for (i = 0; i < dc->count; i++)
+		{
+			int score = 0;
+			const char **entry;
+
+			if (dc->handler[i]->recognize)
+				score = dc->handler[i]->recognize(ctx, magic);
+
 			for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
-				if (!fz_strcasecmp(needle, *entry) && score < 100)
+				if (!fz_strcasecmp(magic, *entry) && score < 100)
 				{
 					score = 100;
 					break;
 				}
-		}
 
-		for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
-			if (!fz_strcasecmp(needle, *entry) && score < 100)
+			if (ext)
 			{
-				score = 100;
-				break;
+				for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
+					if (!fz_strcasecmp(ext, *entry) && score < 100)
+					{
+						score = 100;
+						break;
+					}
 			}
 
-		if (best_score < score)
-		{
-			best_score = score;
-			best_i = i;
+			if (best_score < score)
+			{
+				best_score = score;
+				best_i = i;
+			}
 		}
 	}
 
@@ -145,6 +167,27 @@ fz_recognize_document(fz_context *ctx, const char *magic)
 	return dc->handler[best_i];
 }
 
+const fz_document_handler *fz_recognize_document_content(fz_context *ctx, const char *filename)
+{
+	fz_stream *stream  = fz_open_file(ctx, filename);
+	const fz_document_handler *handler = NULL;
+
+	fz_try(ctx)
+		handler = fz_recognize_document_stream_content(ctx, stream, filename);
+	fz_always(ctx)
+		fz_drop_stream(ctx, stream);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return handler;
+}
+
+const fz_document_handler *
+fz_recognize_document(fz_context *ctx, const char *magic)
+{
+	return fz_recognize_document_stream_content(ctx, NULL, magic);
+}
+
 #if FZ_ENABLE_PDF
 extern fz_document_handler pdf_document_handler;
 #endif
@@ -154,15 +197,17 @@ fz_open_accelerated_document_with_stream(fz_context *ctx, const char *magic, fz_
 {
 	const fz_document_handler *handler;
 
-	if (magic == NULL || stream == NULL)
+	if (stream == NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "no document to open");
+	if (magic == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "missing file type");
 
-	handler = fz_recognize_document(ctx, magic);
+	handler = fz_recognize_document_stream_content(ctx, stream, magic);
 	if (!handler)
 #if FZ_ENABLE_PDF
 		handler = &pdf_document_handler;
 #else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find document handler for file type: %s", magic);
+		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find document handler for file type: '%s'", magic);
 #endif
 	if (handler->open_accel_with_stream)
 		if (accel || handler->open_with_stream == NULL)
@@ -183,6 +228,20 @@ fz_open_document_with_stream(fz_context *ctx, const char *magic, fz_stream *stre
 	return fz_open_accelerated_document_with_stream(ctx, magic, stream, NULL);
 }
 
+fz_document *
+fz_open_document_with_buffer(fz_context *ctx, const char *magic, fz_buffer *buffer)
+{
+	fz_document *doc;
+	fz_stream *stream = fz_open_buffer(ctx, buffer);
+	fz_try(ctx)
+		doc = fz_open_document_with_stream(ctx, magic, stream);
+	fz_always(ctx)
+		fz_drop_stream(ctx, stream);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+	return doc;
+}
+
 fz_document *
 fz_open_accelerated_document(fz_context *ctx, const char *filename, const char *accel)
 {
@@ -196,7 +255,7 @@ fz_open_accelerated_document(fz_context *ctx, const char *filename, const char *
 	if (filename == NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "no document to open");
 
-	handler = fz_recognize_document(ctx, filename);
+	handler = fz_recognize_document_content(ctx, filename);
 	if (!handler)
 #if FZ_ENABLE_PDF
 		handler = &pdf_document_handler;
@@ -361,21 +420,52 @@ fz_has_permission(fz_context *ctx, fz_document *doc, fz_permission p)
 fz_outline *
 fz_load_outline(fz_context *ctx, fz_document *doc)
 {
+	if (doc == NULL)
+		return NULL;
 	fz_ensure_layout(ctx, doc);
-	if (doc && doc->load_outline)
+	if (doc->load_outline)
 		return doc->load_outline(ctx, doc);
-	return NULL;
+	if (doc->outline_iterator == NULL)
+		return NULL;
+	return fz_load_outline_from_iterator(ctx, doc->outline_iterator(ctx, doc));
+}
+
+fz_outline_iterator *
+fz_new_outline_iterator(fz_context *ctx, fz_document *doc)
+{
+	if (doc == NULL)
+		return NULL;
+	if (doc->outline_iterator)
+		return doc->outline_iterator(ctx, doc);
+	if (doc->load_outline == NULL)
+		return NULL;
+	return fz_outline_iterator_from_outline(ctx, fz_load_outline(ctx, doc));
+}
+
+fz_link_dest
+fz_resolve_link_dest(fz_context *ctx, fz_document *doc, const char *uri)
+{
+	fz_ensure_layout(ctx, doc);
+	if (doc && doc->resolve_link_dest)
+		return doc->resolve_link_dest(ctx, doc, uri);
+	return fz_make_link_dest_none();
+}
+
+char *
+fz_format_link_uri(fz_context *ctx, fz_document *doc, fz_link_dest dest)
+{
+	if (doc && doc->format_link_uri)
+		return doc->format_link_uri(ctx, doc, dest);
+	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot create internal links for this document type");
 }
 
 fz_location
 fz_resolve_link(fz_context *ctx, fz_document *doc, const char *uri, float *xp, float *yp)
 {
-	fz_ensure_layout(ctx, doc);
-	if (xp) *xp = 0;
-	if (yp) *yp = 0;
-	if (doc && doc->resolve_link)
-		return doc->resolve_link(ctx, doc, uri, xp, yp);
-	return fz_make_location(-1, -1);
+	fz_link_dest dest = fz_resolve_link_dest(ctx, doc, uri);
+	if (xp) *xp = dest.x;
+	if (yp) *yp = dest.y;
+	return dest.loc;
 }
 
 void
@@ -524,6 +614,13 @@ fz_lookup_metadata(fz_context *ctx, fz_document *doc, const char *key, char *buf
 	return -1;
 }
 
+void
+fz_set_metadata(fz_context *ctx, fz_document *doc, const char *key, const char *value)
+{
+	if (doc && doc->set_metadata)
+		doc->set_metadata(ctx, doc, key, value);
+}
+
 fz_colorspace *
 fz_document_output_intent(fz_context *ctx, fz_document *doc)
 {
@@ -588,7 +685,15 @@ fz_rect
 fz_bound_page(fz_context *ctx, fz_page *page)
 {
 	if (page && page->bound_page)
-		return page->bound_page(ctx, page);
+		return page->bound_page(ctx, page, FZ_CROP_BOX);
+	return fz_empty_rect;
+}
+
+fz_rect
+fz_bound_page_box(fz_context *ctx, fz_page *page, fz_box_type box)
+{
+	if (page && page->bound_page)
+		return page->bound_page(ctx, page, box);
 	return fz_empty_rect;
 }
 
@@ -727,15 +832,40 @@ int fz_page_uses_overprint(fz_context *ctx, fz_page *page)
 
 fz_link *fz_create_link(fz_context *ctx, fz_page *page, fz_rect bbox, const char *uri)
 {
-	if (page == NULL)
+	if (page == NULL || uri == NULL)
 		return NULL;
 	if (page->create_link == NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support creating links");
-	if (uri && !fz_is_external_link(ctx, uri))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "URI should be NULL, or an external link");
 	return page->create_link(ctx, page, bbox, uri);
 }
 
+void fz_delete_link(fz_context *ctx, fz_page *page, fz_link *link)
+{
+	if (page == NULL || link == NULL)
+		return;
+	if (page->delete_link == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support deleting links");
+	page->delete_link(ctx, page, link);
+}
+
+void fz_set_link_rect(fz_context *ctx, fz_link *link, fz_rect rect)
+{
+	if (link == NULL)
+		return;
+	if (link->set_rect_fn == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support updating link bounds");
+	link->set_rect_fn(ctx, link, rect);
+}
+
+void fz_set_link_uri(fz_context *ctx, fz_link *link, const char *uri)
+{
+	if (link == NULL)
+		return;
+	if (link->set_uri_fn == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support updating link uri");
+	link->set_uri_fn(ctx, link, uri);
+}
+
 void *
 fz_process_opened_pages(fz_context *ctx, fz_document *doc, fz_process_opened_page_fn *process_opened_page, void *state)
 {
@@ -786,3 +916,45 @@ fz_process_opened_pages(fz_context *ctx, fz_document *doc, fz_process_opened_pag
 
 	return ret;
 }
+
+const char *
+fz_page_label(fz_context *ctx, fz_page *page, char *buf, int size)
+{
+	fz_document *doc = page->doc;
+	if (doc->page_label)
+		doc->page_label(ctx, page->doc, page->chapter, page->number, buf, size);
+	else if (fz_count_chapters(ctx, page->doc) > 1)
+		fz_snprintf(buf, size, "%d/%d", page->chapter + 1, page->number + 1);
+	else
+		fz_snprintf(buf, size, "%d", page->number + 1);
+	return buf;
+}
+
+
+fz_box_type fz_box_type_from_string(const char *name)
+{
+	if (!fz_strcasecmp(name, "MediaBox"))
+		return FZ_MEDIA_BOX;
+	if (!fz_strcasecmp(name, "CropBox"))
+		return FZ_CROP_BOX;
+	if (!fz_strcasecmp(name, "BleedBox"))
+		return FZ_BLEED_BOX;
+	if (!fz_strcasecmp(name, "TrimBox"))
+		return FZ_TRIM_BOX;
+	if (!fz_strcasecmp(name, "ArtBox"))
+		return FZ_ART_BOX;
+	return FZ_UNKNOWN_BOX;
+}
+
+const char *fz_string_from_box_type(fz_box_type box)
+{
+	switch (box)
+	{
+	case FZ_MEDIA_BOX: return "MediaBox";
+	case FZ_CROP_BOX: return "CropBox";
+	case FZ_BLEED_BOX: return "BleedBox";
+	case FZ_TRIM_BOX: return "TrimBox";
+	case FZ_ART_BOX: return "ArtBox";
+	default: return "UnknownBox";
+	}
+}
