@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -61,6 +61,7 @@ typedef struct {
 	fz_matrix ctm;
 	float xstep, ystep;
 	fz_irect area;
+	int flags;
 } fz_draw_state;
 
 typedef struct fz_draw_device
@@ -72,6 +73,7 @@ typedef struct fz_draw_device
 	fz_colorspace *proof_cs;
 	int flags;
 	int resolve_spots;
+	int overprint_possible;
 	int top;
 	fz_scale_cache *cache_x;
 	fz_scale_cache *cache_y;
@@ -409,8 +411,23 @@ set_op_from_spaces(fz_context *ctx, fz_overprint *op, const fz_pixmap *dest, fz_
 	if (!op)
 		return NULL;
 
-	if (!fz_colorspace_is_subtractive(ctx, src) || !fz_colorspace_is_subtractive(ctx, dest->colorspace))
+	if (fz_colorspace_is_indexed(ctx, src))
+		src = fz_base_colorspace(ctx, src);
+
+	if (!fz_colorspace_is_subtractive(ctx, dest->colorspace))
 		return NULL;
+	if (fz_colorspace_is_gray(ctx, src))
+	{
+		/* gray counts as a CMYK with CMY = 0 */
+	}
+	else if (!fz_colorspace_is_subtractive(ctx, src))
+	{
+		/* The source pixmap was not in a subtractive space, so we can't
+		 * base overprint decisions on that. But we know we've converted
+		 * to a subtractive destination, so we can base overprint
+		 * decisions on what we ended up with. */
+		src = dest->colorspace;
+	}
 
 	sn = fz_colorspace_n(ctx, src);
 	dn = dest->n - dest->alpha;
@@ -505,7 +522,8 @@ resolve_color(fz_context *ctx,
 	float alpha,
 	fz_color_params color_params,
 	unsigned char *colorbv,
-	fz_pixmap *dest)
+	fz_pixmap *dest,
+	int overprint_possible)
 {
 	float colorfv[FZ_MAX_COLORS];
 	int i;
@@ -522,7 +540,7 @@ resolve_color(fz_context *ctx,
 	devgray = fz_colorspace_is_device_gray(ctx, colorspace);
 
 	/* We can only overprint when enabled, and when we are in a subtractive colorspace */
-	if (color_params.op == 0 || !fz_colorspace_is_subtractive(ctx, dest->colorspace))
+	if (color_params.op == 0 || !fz_colorspace_is_subtractive(ctx, dest->colorspace) || !overprint_possible)
 		op = NULL;
 
 	else if (devgray)
@@ -650,10 +668,13 @@ fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 	if (fz_flatten_fill_path(ctx, rast, path, ctm, flatness, bbox, &bbox))
 		return;
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (alpha == 0)
+		return;
+
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		state = fz_knockout_begin(ctx, dev);
 
-	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest);
+	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest, dev->overprint_possible);
 
 	fz_convert_rasterizer(ctx, rast, even_odd, state->dest, colorbv, eop);
 	if (state->shape)
@@ -673,7 +694,7 @@ fz_draw_fill_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 		fz_convert_rasterizer(ctx, rast, even_odd, state->group_alpha, colorbv, 0);
 	}
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		fz_knockout_end(ctx, dev);
 }
 
@@ -713,10 +734,13 @@ fz_draw_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path, const
 	if (fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, bbox, &bbox))
 		return;
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (alpha == 0)
+		return;
+
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		state = fz_knockout_begin(ctx, dev);
 
-	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest);
+	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest, dev->overprint_possible);
 
 #ifdef DUMP_GROUP_BLENDS
 	dump_spaces(dev->top, "");
@@ -755,7 +779,7 @@ fz_draw_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path, const
 	printf("\n");
 #endif
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		fz_knockout_end(ctx, dev);
 }
 
@@ -785,16 +809,9 @@ fz_draw_clip_path(fz_context *ctx, fz_device *devp, const fz_path *path, int eve
 
 	model = state->dest->colorspace;
 
+	bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
 	if (!fz_is_infinite_rect(scissor))
-	{
-		bbox = fz_irect_from_rect(fz_transform_rect(scissor, dev->transform));
-		bbox = fz_intersect_irect(bbox, fz_pixmap_bbox(ctx, state->dest));
-		bbox = fz_intersect_irect(bbox, state->scissor);
-	}
-	else
-	{
-		bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
-	}
+		bbox = fz_intersect_irect(bbox, fz_irect_from_rect(fz_transform_rect(scissor, dev->transform)));
 
 	if (fz_flatten_fill_path(ctx, rast, path, ctm, flatness, bbox, &bbox) || fz_is_rect_rasterizer(ctx, rast))
 	{
@@ -863,16 +880,9 @@ fz_draw_clip_stroke_path(fz_context *ctx, fz_device *devp, const fz_path *path,
 
 	model = state->dest->colorspace;
 
+	bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
 	if (!fz_is_infinite_rect(scissor))
-	{
-		bbox = fz_irect_from_rect(fz_transform_rect(scissor, dev->transform));
-		bbox = fz_intersect_irect(bbox, fz_pixmap_bbox(ctx, state->dest));
-		bbox = fz_intersect_irect(bbox, state->scissor);
-	}
-	else
-	{
-		bbox = fz_intersect_irect(fz_pixmap_bbox(ctx, state->dest), state->scissor);
-	}
+		bbox = fz_intersect_irect(bbox, fz_irect_from_rect(fz_transform_rect(scissor, dev->transform)));
 
 	if (fz_flatten_stroke_path(ctx, rast, path, stroke, ctm, flatness, linewidth, bbox, &bbox))
 	{
@@ -958,7 +968,6 @@ draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_glyph *glyph,
 			fz_span_color_painter_t *fn;
 
 			fn = fz_get_span_color_painter(dst->n, da, colorbv, eop);
-			assert(fn);
 			if (fn == NULL)
 				return;
 			while (h--)
@@ -971,14 +980,14 @@ draw_glyph(unsigned char *colorbv, fz_pixmap *dst, fz_glyph *glyph,
 		else
 		{
 			fz_span_painter_t *fn;
+			int col = colorbv ? colorbv[0] : 255;
 
-			fn = fz_get_span_painter(da, 1, 0, 255, eop);
-			assert(fn);
+			fn = fz_get_span_painter(da, 1, 0, col, eop);
 			if (fn == NULL)
 				return;
 			while (h--)
 			{
-				(*fn)(dp, da, mp, 1, 0, w, 255, eop);
+				(*fn)(dp, da, mp, 1, 0, w, col, eop);
 				dp += dst->stride;
 				mp += msk->stride;
 			}
@@ -1012,10 +1021,13 @@ fz_draw_fill_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 	if (colorspace == NULL && model != NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "color destination requires source color");
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (alpha == 0)
+		return;
+
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		state = fz_knockout_begin(ctx, dev);
 
-	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest);
+	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest, dev->overprint_possible);
 	shapebv = 255;
 	shapebva = 255 * alpha;
 
@@ -1077,7 +1089,7 @@ fz_draw_fill_text(fz_context *ctx, fz_device *devp, const fz_text *text, fz_matr
 		}
 	}
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		fz_knockout_end(ctx, dev);
 }
 
@@ -1088,6 +1100,7 @@ fz_draw_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text, const
 	fz_draw_device *dev = (fz_draw_device*)devp;
 	fz_matrix ctm = fz_concat(in_ctm, dev->transform);
 	fz_draw_state *state = &dev->stack[dev->top];
+	fz_colorspace *model = state->dest->colorspace;
 	unsigned char colorbv[FZ_MAX_COLORS + 1];
 	unsigned char solid = 255;
 	unsigned char alpha_byte = alpha * 255;
@@ -1104,10 +1117,13 @@ fz_draw_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text, const
 	if (colorspace_in)
 		colorspace = fz_default_colorspace(ctx, dev->default_cs, colorspace_in);
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (alpha == 0)
+		return;
+
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		state = fz_knockout_begin(ctx, dev);
 
-	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest);
+	eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest, dev->overprint_possible);
 
 	for (span = text->head; span; span = span->next)
 	{
@@ -1127,7 +1143,7 @@ fz_draw_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text, const
 			tm.f = span->items[i].y;
 			trm = fz_concat(tm, ctm);
 
-			glyph = fz_render_stroked_glyph(ctx, span->font, gid, &trm, ctm, stroke, &state->scissor, aa);
+			glyph = fz_render_stroked_glyph(ctx, span->font, gid, &trm, ctm, model, stroke, &state->scissor, aa);
 			if (glyph)
 			{
 				fz_pixmap *pixmap = glyph->pixmap;
@@ -1154,20 +1170,22 @@ fz_draw_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text, const
 			else
 			{
 				fz_path *path = fz_outline_glyph(ctx, span->font, gid, tm);
-				if (path)
-				{
-					fz_draw_stroke_path(ctx, devp, path, stroke, in_ctm, colorspace, color, alpha, color_params);
-					fz_drop_path(ctx, path);
-				}
+				if (!path)
+					fz_warn(ctx, "cannot render glyph");
 				else
 				{
-					fz_warn(ctx, "cannot render glyph");
+					fz_try(ctx)
+						fz_draw_stroke_path(ctx, devp, path, stroke, in_ctm, colorspace, color, alpha, color_params);
+					fz_always(ctx)
+						fz_drop_path(ctx, path);
+					fz_catch(ctx)
+						fz_rethrow(ctx);
 				}
 			}
 		}
 	}
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		fz_knockout_end(ctx, dev);
 }
 
@@ -1374,7 +1392,7 @@ fz_draw_clip_stroke_text(fz_context *ctx, fz_device *devp, const fz_text *text,
 				tm.f = span->items[i].y;
 				trm = fz_concat(tm, ctm);
 
-				glyph = fz_render_stroked_glyph(ctx, span->font, gid, &trm, ctm, stroke, &state->scissor, aa);
+				glyph = fz_render_stroked_glyph(ctx, span->font, gid, &trm, ctm, model, stroke, &state->scissor, aa);
 				if (glyph)
 				{
 					int x = (int)trm.e;
@@ -1452,7 +1470,10 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 	if (fz_is_empty_irect(bbox))
 		return;
 
-	if (state->blendmode & FZ_BLEND_KNOCKOUT)
+	if (alpha == 0)
+		return;
+
+	if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 		state = fz_knockout_begin(ctx, dev);
 
 	fz_var(dest);
@@ -1492,7 +1513,7 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 			/* Disable OPM */
 			color_params.opm = 0;
 
-			eop = resolve_color(ctx, &op, shade->background, colorspace, alpha, color_params, colorbv, state->dest);
+			eop = resolve_color(ctx, &op, shade->background, colorspace, alpha, color_params, colorbv, state->dest, dev->overprint_possible);
 
 			n = dest->n;
 			if (fz_overprint_required(eop))
@@ -1587,7 +1608,7 @@ fz_draw_fill_shade(fz_context *ctx, fz_device *devp, fz_shade *shade, fz_matrix
 			}
 		}
 
-		if (state->blendmode & FZ_BLEND_KNOCKOUT)
+		if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 			fz_knockout_end(ctx, dev);
 	}
 	fz_catch(ctx)
@@ -1671,7 +1692,7 @@ convert_pixmap_for_painting(fz_context *ctx, fz_pixmap *pixmap, fz_colorspace *m
 {
 	fz_pixmap *converted;
 
-	if (fz_colorspace_is_device_n(ctx, src_cs) && dest->seps)
+	if ((fz_colorspace_is_device_n(ctx, src_cs) && dest->seps) || fz_compare_separations(ctx, pixmap->seps, dest->seps))
 	{
 		converted = fz_clone_pixmap_area_with_different_seps(ctx, pixmap, NULL, model, dest->seps, color_params, dev->default_cs);
 		*eop = set_op_from_spaces(ctx, *eop, dest, src_cs, 0);
@@ -1777,7 +1798,8 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 	if (color_params.op == 0)
 		eop = NULL;
 
-	local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
+	if (!(dev->flags & FZ_DRAWDEV_FLAGS_TYPE3))
+		local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
 
 	src_area = find_src_area_required(local_ctm, image, clip);
 	if (fz_is_empty_irect(src_area))
@@ -1794,9 +1816,9 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 
 	fz_try(ctx)
 	{
-		int conversion_required = (src_cs != model || state->dest->seps);
+		int conversion_required = (src_cs != model || fz_compare_separations(ctx, state->dest->seps, pixmap->seps));
 
-		if (state->blendmode & FZ_BLEND_KNOCKOUT)
+		if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 			state = fz_knockout_begin(ctx, dev);
 
 		switch (fz_colorspace_type(ctx, src_cs))
@@ -1853,7 +1875,7 @@ fz_draw_fill_image(fz_context *ctx, fz_device *devp, fz_image *image, fz_matrix
 
 		fz_paint_image(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, local_ctm, alpha * 255, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), eop);
 
-		if (state->blendmode & FZ_BLEND_KNOCKOUT)
+		if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 			fz_knockout_end(ctx, dev);
 	}
 	fz_always(ctx)
@@ -1894,7 +1916,8 @@ fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 	if (image->w == 0 || image->h == 0)
 		return;
 
-	local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
+	if (!(dev->flags & FZ_DRAWDEV_FLAGS_TYPE3))
+		local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
 
 	src_area = find_src_area_required(local_ctm, image, clip);
 	if (fz_is_empty_irect(src_area))
@@ -1906,7 +1929,7 @@ fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 
 	fz_try(ctx)
 	{
-		if (state->blendmode & FZ_BLEND_KNOCKOUT)
+		if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 			state = fz_knockout_begin(ctx, dev);
 
 		if (!(devp->hints & FZ_DONT_INTERPOLATE_IMAGES) && ctx->tuning->image_scale(ctx->tuning->image_scale_arg, dx, dy, pixmap->w, pixmap->h))
@@ -1928,11 +1951,11 @@ fz_draw_fill_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 			}
 		}
 
-		eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest);
+		eop = resolve_color(ctx, &op, color, colorspace, alpha, color_params, colorbv, state->dest, dev->overprint_possible);
 
 		fz_paint_image_with_color(ctx, state->dest, &state->scissor, state->shape, state->group_alpha, pixmap, local_ctm, colorbv, !(devp->hints & FZ_DONT_INTERPOLATE_IMAGES), eop);
 
-		if (state->blendmode & FZ_BLEND_KNOCKOUT)
+		if (state->blendmode & FZ_BLEND_KNOCKOUT && alpha != 1)
 			fz_knockout_end(ctx, dev);
 	}
 	fz_always(ctx)
@@ -1973,7 +1996,8 @@ fz_draw_clip_image_mask(fz_context *ctx, fz_device *devp, fz_image *image, fz_ma
 		return;
 	}
 
-	local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
+	if (!(dev->flags & FZ_DRAWDEV_FLAGS_TYPE3))
+		local_ctm = fz_gridfit_matrix(devp->flags & FZ_DEVFLAG_GRIDFIT_AS_TILED, local_ctm);
 
 	src_area = find_src_area_required(local_ctm, image, clip);
 	if (fz_is_empty_irect(src_area))
@@ -2110,13 +2134,13 @@ fz_draw_pop_clip(fz_context *ctx, fz_device *devp)
 		fz_paint_pixmap_with_mask(state[0].dest, state[1].dest, state[1].mask);
 		if (state[0].shape != state[1].shape)
 		{
-			fz_paint_pixmap_with_mask(state[0].shape, state[1].shape, state[1].mask);
+			fz_paint_over_pixmap_with_mask(state[0].shape, state[1].shape, state[1].mask);
 			fz_drop_pixmap(ctx, state[1].shape);
 			state[1].shape = NULL;
 		}
 		if (state[0].group_alpha != state[1].group_alpha)
 		{
-			fz_paint_pixmap_with_mask(state[0].group_alpha, state[1].group_alpha, state[1].mask);
+			fz_paint_over_pixmap_with_mask(state[0].group_alpha, state[1].group_alpha, state[1].mask);
 			fz_drop_pixmap(ctx, state[1].group_alpha);
 			state[1].group_alpha = NULL;
 		}
@@ -2595,13 +2619,20 @@ fz_draw_begin_tile(fz_context *ctx, fz_device *devp, fz_rect area, fz_rect view,
 	if (dev->top == 0 && dev->resolve_spots)
 		state = push_group_for_separations(ctx, dev, fz_default_color_params /* FIXME */, dev->default_cs);
 
-	/* area, view, xstep, ystep are in pattern space */
+	/* area, view, xstep, ystep are in pattern space
+	 * area = the extent that we need to tile the pattern into. (i.e this is
+	 * the area to be filled with the pattern mapped back into pattern space).
+	 * view = the pattern bbox.
+	 * xstep and ystep are the repeats for the file.
+	 */
 	/* ctm maps from pattern space to device space */
 
 	if (state->blendmode & FZ_BLEND_KNOCKOUT)
 		fz_knockout_begin(ctx, dev);
 
 	state = push_stack(ctx, dev, "tile");
+	state[1].flags = dev->flags;
+	dev->flags &= ~FZ_DRAWDEV_FLAGS_TYPE3;
 
 	local_view = fz_transform_rect(view, ctm);
 	bbox = fz_irect_from_rect(local_view);
@@ -2613,6 +2644,12 @@ fz_draw_begin_tile(fz_context *ctx, fz_device *devp, fz_rect area, fz_rect view,
 	 *	bbox.y0 > state->dest->y || bbox.y1 < state->dest->y + state->dest->h);
 	 */
 
+	/* A BBox of zero height or width should still paint one pixel! */
+	if (bbox.x1 == bbox.x0)
+		bbox.x1 = bbox.x0 + 1;
+	if (bbox.y1 == bbox.y0)
+		bbox.y1 = bbox.y0 + 1;
+
 	/* Check to see if we have one cached */
 	if (id)
 	{
@@ -2701,6 +2738,7 @@ fz_draw_end_tile(fz_context *ctx, fz_device *devp)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "unexpected end tile");
 
 	state = pop_stack(ctx, dev, "tile");
+	dev->flags = state[1].flags;
 
 	xstep = state[1].xstep;
 	ystep = state[1].ystep;
@@ -3015,6 +3053,7 @@ new_draw_device(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, const fz_
 	dev->stack[0].scissor.y0 = dest->y;
 	dev->stack[0].scissor.x1 = dest->x + dest->w;
 	dev->stack[0].scissor.y1 = dest->y + dest->h;
+	dev->stack[0].flags = dev->flags;
 
 	if (clip)
 	{
@@ -3052,6 +3091,8 @@ new_draw_device(fz_context *ctx, fz_matrix transform, fz_pixmap *dest, const fz_
 		fz_throw(ctx, FZ_ERROR_GENERIC, "Spot rendering (and overprint/overprint simulation) not available in this build");
 #endif
 
+	dev->overprint_possible = (dest->seps != NULL);
+
 	fz_try(ctx)
 	{
 		dev->rast = fz_new_rasterizer(ctx, aa);
