@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2022 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -29,7 +29,7 @@
 struct null_filter
 {
 	fz_stream *chain;
-	size_t remain;
+	uint64_t remain;
 	int64_t offset;
 	unsigned char buffer[4096];
 };
@@ -38,7 +38,7 @@ static int
 next_null(fz_context *ctx, fz_stream *stm, size_t max)
 {
 	struct null_filter *state = stm->state;
-	size_t n;
+	uint64_t n;
 
 	if (state->remain == 0)
 		return EOF;
@@ -71,7 +71,7 @@ close_null(fz_context *ctx, void *state_)
 }
 
 fz_stream *
-fz_open_null_filter(fz_context *ctx, fz_stream *chain, int len, int64_t offset)
+fz_open_null_filter(fz_context *ctx, fz_stream *chain, uint64_t len, int64_t offset)
 {
 	struct null_filter *state = fz_malloc_struct(ctx, struct null_filter);
 	state->chain = fz_keep_stream(ctx, chain);
@@ -179,7 +179,8 @@ fz_open_range_filter(fz_context *ctx, fz_stream *chain, fz_range *ranges, int nr
 struct endstream_filter
 {
 	fz_stream *chain;
-	size_t remain, extras, size;
+	uint64_t remain;
+	size_t extras, size;
 	int64_t offset;
 	int warned;
 	unsigned char buffer[4096];
@@ -292,13 +293,10 @@ close_endstream(fz_context *ctx, void *state_)
 }
 
 fz_stream *
-fz_open_endstream_filter(fz_context *ctx, fz_stream *chain, int len, int64_t offset)
+fz_open_endstream_filter(fz_context *ctx, fz_stream *chain, uint64_t len, int64_t offset)
 {
 	struct endstream_filter *state;
 
-	if (len < 0)
-		len = 0;
-
 	state = fz_malloc_struct(ctx, struct endstream_filter);
 	state->chain = fz_keep_stream(ctx, chain);
 	state->remain = len;
@@ -733,12 +731,23 @@ close_rld(fz_context *ctx, void *state_)
 fz_stream *
 fz_open_rld(fz_context *ctx, fz_stream *chain)
 {
+	fz_stream *stm;
 	fz_rld *state = fz_malloc_struct(ctx, fz_rld);
 	state->chain = fz_keep_stream(ctx, chain);
 	state->run = 0;
 	state->n = 0;
 	state->c = 0;
-	return fz_new_stream(ctx, state, next_rld, close_rld);
+
+	stm = fz_new_stream(ctx, state, next_rld, close_rld);
+
+	/* Don't explode RLE compression bombs. */
+	if (chain->next == next_rld)
+	{
+		fz_warn(ctx, "RLE bomb defused");
+		stm->eof = 1;
+	}
+
+	return stm;
 }
 
 /* RC4 Filter */
