@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <stdio.h>
@@ -15,6 +37,7 @@ typedef struct
 	fz_stream *curr_stm;
 	fz_context *ctx;
 	int color_transform;
+	int invert_cmyk; /* has inverted CMYK polarity */
 	int init;
 	int stride;
 	int l2factor;
@@ -95,6 +118,11 @@ static void error_exit_dct(j_common_ptr cinfo)
 	fz_throw(ctx, FZ_ERROR_GENERIC, "jpeg error: %s", msg);
 }
 
+static void output_message_dct(j_common_ptr cinfo)
+{
+	/* swallow message */
+}
+
 static void init_source_dct(j_decompress_ptr cinfo)
 {
 	/* nothing to do */
@@ -149,6 +177,13 @@ static void skip_input_data_dct(j_decompress_ptr cinfo, long num_bytes)
 	}
 }
 
+static void invert_cmyk(unsigned char *p, int n)
+{
+	int i;
+	for (i = 0; i < n; ++i)
+		p[i] = 255 - p[i];
+}
+
 static int
 next_dctd(fz_context *ctx, fz_stream *stm, size_t max)
 {
@@ -197,33 +232,32 @@ next_dctd(fz_context *ctx, fz_stream *stm, size_t max)
 
 			jpeg_read_header(cinfo, 1);
 
-			/* default value if ColorTransform is not set */
-			if (state->color_transform == -1)
-			{
-				if (state->cinfo.num_components == 3)
-					state->color_transform = 1;
-				else
-					state->color_transform = 0;
-			}
+			/* Invert CMYK polarity if:
+			 *    It is a standalone JPEG file (i.e. not embedded in PDF; color_transform is set to -1).
+			 *       In PDF, the polarity inversion is usually done with the image Decode array if necessary.
+			 *       We set color_transform to -2 or a positive value in this cases.
+			 *    It has an Adobe marker setting the color transform to YCCK to CMYK.
+			 *       Experimentation has shown that if the color transform is set to 0 the polarity is
+			 *       usually not inverted.
+			 */
+			if (cinfo->out_color_space == JCS_CMYK && cinfo->Adobe_transform == 2 && state->color_transform == -1)
+				state->invert_cmyk = 1;
 
+			/* Adobe APP marker overrides ColorTransform from PDF */
 			if (cinfo->saw_Adobe_marker)
 				state->color_transform = cinfo->Adobe_transform;
 
-			/* Guess the input colorspace, and set output colorspace accordingly */
-			switch (cinfo->num_components)
+			/* Disable JPEG color transformations if ColorTransform is 0.
+			 * This is usually handled by libjpeg, but since PDF can override
+			 * the default behavior if the Adobe APP marker is missing
+			 * we must do it here as well.
+			 */
+			if (state->color_transform == 0)
 			{
-			case 3:
-				if (state->color_transform)
-					cinfo->jpeg_color_space = JCS_YCbCr;
-				else
+				if (cinfo->num_components == 3)
 					cinfo->jpeg_color_space = JCS_RGB;
-				break;
-			case 4:
-				if (state->color_transform)
-					cinfo->jpeg_color_space = JCS_YCCK;
-				else
+				if (cinfo->num_components == 4)
 					cinfo->jpeg_color_space = JCS_CMYK;
-				break;
 			}
 
 			cinfo->scale_num = 8/(1<<state->l2factor);
@@ -248,11 +282,15 @@ next_dctd(fz_context *ctx, fz_stream *stm, size_t max)
 			if (p + state->stride <= ep)
 			{
 				jpeg_read_scanlines(cinfo, &p, 1);
+				if (state->invert_cmyk)
+					invert_cmyk(p, state->stride);
 				p += state->stride;
 			}
 			else
 			{
 				jpeg_read_scanlines(cinfo, &state->scanline, 1);
+				if (state->invert_cmyk)
+					invert_cmyk(state->scanline, state->stride);
 				state->rp = state->scanline;
 				state->wp = state->scanline + state->stride;
 			}
@@ -336,6 +374,7 @@ fz_open_dctd(fz_context *ctx, fz_stream *chain, int color_transform, int l2facto
 	cinfo->src = NULL;
 	cinfo->err = &state->errmgr;
 	jpeg_std_error(cinfo->err);
+	cinfo->err->output_message = output_message_dct;
 	cinfo->err->error_exit = error_exit_dct;
 
 	return fz_new_stream(ctx, state, next_dctd, close_dctd);
