@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <assert.h>
@@ -30,7 +52,11 @@ typedef enum
 	FZ_CMD_RENDER_FLAGS,
 	FZ_CMD_DEFAULT_COLORSPACES,
 	FZ_CMD_BEGIN_LAYER,
-	FZ_CMD_END_LAYER
+	FZ_CMD_END_LAYER,
+	FZ_CMD_BEGIN_STRUCTURE,
+	FZ_CMD_END_STRUCTURE,
+	FZ_CMD_BEGIN_METATEXT,
+	FZ_CMD_END_METATEXT
 } fz_display_command;
 
 /* The display list is a list of nodes.
@@ -42,9 +68,15 @@ typedef enum
  *	cmd:	What type of node this is.
  *
  *	size:	The number of sizeof(fz_display_node) bytes that this node's
- *		data occupies. (i.e. &node[node->size] = the next node in the
+ *		data occupies. (i.e. &node[size] = the next node in the
  *		chain; 0 for end of list).
  *
+ *		At 9 bits for this field, and 4 bytes for a node, this means
+ *		the largest node can span 511*4 bytes. We therefore reserve
+ *		511 as a special value to mean that the actual size for the
+ *		node is given in an (unaligned!) size_t following the node
+ *		before the rest of the data.
+ *
  *	rect:	0 for unchanged, 1 for present.
  *
  *	path:	0 for unchanged, 1 for present.
@@ -109,7 +141,7 @@ enum {
 	CTM_CHANGE_BC = 2,
 	CTM_CHANGE_EF = 4,
 
-	MAX_NODE_SIZE = (1<<9)-sizeof(fz_display_node)
+	INDIRECT_NODE_THRESHOLD = (1<<9)-1
 };
 
 struct fz_display_list
@@ -150,7 +182,57 @@ enum { OPM = 1, OP = 2, BP = 3, RI = 4};
 #define SIZE_IN_NODES(t) \
 	((t + sizeof(fz_display_node) - 1) / sizeof(fz_display_node))
 
-static void
+/* The display list node are 32bit aligned. For some architectures we
+ * need to pad to 64bit for pointers. We allow for that here. */
+static void pad_size_for_pointer(const fz_display_list *list, size_t *size)
+{
+	/* list->len and size are both counting in nodes, not bytes.
+	 * Nodes are consistently 32bit things, hence we are looking
+	 * for even/odd for "8 byte aligned or not". */
+	if (FZ_POINTER_ALIGN_MOD <= 4)
+		return;
+	/* Otherwise, ensure we're on an even boundary. */
+	if (FZ_POINTER_ALIGN_MOD == 8)
+	{
+		if ((list->len + (*size)) & 1)
+			(*size)++;
+	} else
+		(*size) = ((list->len + (*size) + (FZ_POINTER_ALIGN_MOD>>2) - 1) & ~((FZ_POINTER_ALIGN_MOD>>2)-1)) - list->len;
+}
+
+static void align_node_for_pointer(fz_display_node **node)
+{
+	intptr_t ptr;
+
+	if (FZ_POINTER_ALIGN_MOD <= 4)
+		return;
+
+	ptr = (intptr_t)*node;
+	if (FZ_POINTER_ALIGN_MOD == 8)
+	{
+		if (ptr & 4)
+			(*node) = (fz_display_node *)(ptr+4);
+	}
+	else
+		(*node) = (fz_display_node *)((ptr + FZ_POINTER_ALIGN_MOD - 1) & ~(FZ_POINTER_ALIGN_MOD-1));
+}
+
+static int
+cmd_needs_alignment(fz_display_command cmd)
+{
+	return (cmd == FZ_CMD_FILL_TEXT ||
+		cmd == FZ_CMD_STROKE_TEXT ||
+		cmd == FZ_CMD_CLIP_TEXT ||
+		cmd == FZ_CMD_CLIP_STROKE_TEXT ||
+		cmd == FZ_CMD_IGNORE_TEXT ||
+		cmd == FZ_CMD_FILL_SHADE ||
+		cmd == FZ_CMD_FILL_IMAGE ||
+		cmd == FZ_CMD_FILL_IMAGE_MASK ||
+		cmd == FZ_CMD_CLIP_IMAGE_MASK ||
+		cmd == FZ_CMD_DEFAULT_COLORSPACES);
+}
+
+static unsigned char *
 fz_append_display_node(
 	fz_context *ctx,
 	fz_device *dev,
@@ -184,6 +266,7 @@ fz_append_display_node(
 	fz_stroke_state *my_stroke = NULL;
 	fz_rect local_rect;
 	size_t path_size = 0;
+	unsigned char *out_private = NULL;
 
 	switch (cmd)
 	{
@@ -262,11 +345,15 @@ fz_append_display_node(
 		rect_off = size;
 		size += SIZE_IN_NODES(sizeof(fz_rect));
 	}
-	if (color || colorspace)
+	if (color == NULL)
+	{
+		if (colorspace)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "Colorspace cannot be specified without color.");
+	}
+	else
 	{
 		if (colorspace != writer->colorspace)
 		{
-			assert(color);
 			if (colorspace == fz_device_gray(ctx))
 			{
 				if (color[0] == 0.0f)
@@ -309,6 +396,7 @@ fz_append_display_node(
 				int i;
 				int n = fz_colorspace_n(ctx, colorspace);
 
+				pad_size_for_pointer(list, &size);
 				colorspace_off = size;
 				size += SIZE_IN_NODES(sizeof(fz_colorspace *));
 				node.cs = CS_OTHER_0;
@@ -385,6 +473,7 @@ fz_append_display_node(
 				if (i == n)
 				{
 					node.cs = CS_OTHER_0;
+					pad_size_for_pointer(list, &size);
 					colorspace_off = size;
 					size += SIZE_IN_NODES(sizeof(fz_colorspace *));
 					color = NULL;
@@ -444,28 +533,34 @@ fz_append_display_node(
 	}
 	if (stroke && (writer->stroke == NULL || stroke != writer->stroke))
 	{
+		pad_size_for_pointer(list, &size);
 		stroke_off = size;
 		size += SIZE_IN_NODES(sizeof(fz_stroke_state *));
 		node.stroke = 1;
 	}
 	if (path && (writer->path == NULL || path != writer->path))
 	{
-		size_t max = SIZE_IN_NODES(MAX_NODE_SIZE) - size - SIZE_IN_NODES(private_data_len);
-		path_size = SIZE_IN_NODES(fz_pack_path(ctx, NULL, max, path));
+		pad_size_for_pointer(list, &size);
+		path_size = SIZE_IN_NODES(fz_pack_path(ctx, NULL, path));
 		node.path = 1;
 		path_off = size;
 
 		size += path_size;
 	}
-	if (private_data != NULL)
+	if (private_data_len)
 	{
-		size_t max = SIZE_IN_NODES(MAX_NODE_SIZE) - size;
-		if (SIZE_IN_NODES(private_data_len) > max)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Private data too large to pack into display list node");
+		if (cmd_needs_alignment(cmd))
+			pad_size_for_pointer(list, &size);
 		private_off = size;
 		size += SIZE_IN_NODES(private_data_len);
 	}
 
+	/* If the size is more than 511, then we can't signal that in 9 bits,
+	 * so we'll send it as 511, and then put an extra size_t with the
+	 * size in. */
+	if (size >= INDIRECT_NODE_THRESHOLD)
+		size += SIZE_IN_NODES(sizeof(size_t));
+
 	while (list->len + size > list->max)
 	{
 		size_t newsize = list->max * 2;
@@ -488,22 +583,28 @@ fz_append_display_node(
 			writer->path = (fz_path *)(((char *)writer->path) + diff);
 	}
 
-	if ((unsigned int)size != size)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Display list node too large");
-
 	/* Write the node to the list */
-	node.size = (unsigned int)size;
+	if (size >= INDIRECT_NODE_THRESHOLD)
+		node.size = INDIRECT_NODE_THRESHOLD;
+	else
+		node.size = (unsigned int)size;
 	node.flags = flags;
-	assert(size < (1<<9));
 	node_ptr = &list->list[list->len];
 	*node_ptr = node;
 
+	/* Insert the explicit size (unaligned) if required. */
+	if (size >= INDIRECT_NODE_THRESHOLD)
+	{
+		memcpy(&node_ptr[1], &size, sizeof(size));
+		node_ptr += SIZE_IN_NODES(sizeof(size_t));
+	}
+
 	/* Path is the most frequent one, so try to avoid the try/catch in
 	 * this case */
 	if (path_off)
 	{
 		my_path = (void *)(&node_ptr[path_off]);
-		(void)fz_pack_path(ctx, (void *)my_path, path_size * sizeof(fz_display_node), path);
+		(void)fz_pack_path(ctx, (void *)my_path, path);
 	}
 
 	if (stroke_off)
@@ -628,12 +729,15 @@ fz_append_display_node(
 		/* Can never fail as my_stroke was kept above */
 		writer->stroke = fz_keep_stroke_state(ctx, my_stroke);
 	}
-	if (private_off)
+	if (private_data_len)
 	{
-		char *out_private = (char *)(void *)(&node_ptr[private_off]);
-		memcpy(out_private, private_data, private_data_len);
+		out_private = (unsigned char *)(void *)(&node_ptr[private_off]);
+		if (private_data)
+			memcpy(out_private, private_data, private_data_len);
 	}
 	list->len += size;
+
+	return out_private;
 }
 
 /* Pack ri, op, opm, bp into flags upper bits, even/odd in lower bit */
@@ -1071,8 +1175,7 @@ static void
 fz_list_begin_group(fz_context *ctx, fz_device *dev, fz_rect rect, fz_colorspace *colorspace, int isolated, int knockout, int blendmode, float alpha)
 {
 	int flags;
-
-	colorspace = fz_keep_colorspace(ctx, colorspace);
+	static const float color[FZ_MAX_COLORS] = { 0 };
 
 	flags = (blendmode<<2);
 	if (isolated)
@@ -1080,28 +1183,20 @@ fz_list_begin_group(fz_context *ctx, fz_device *dev, fz_rect rect, fz_colorspace
 	if (knockout)
 		flags |= KNOCKOUT;
 
-	fz_try(ctx)
-	{
-		fz_append_display_node(
-			ctx,
-			dev,
-			FZ_CMD_BEGIN_GROUP,
-			flags,
-			&rect,
-			NULL, /* path */
-			NULL, /* color */
-			NULL, /* colorspace */
-			&alpha, /* alpha */
-			NULL, /* ctm */
-			NULL, /* stroke */
-			&colorspace, /* private_data */
-			sizeof(colorspace)); /* private_data_len */
-	}
-	fz_catch(ctx)
-	{
-		fz_drop_colorspace(ctx, colorspace);
-		fz_rethrow(ctx);
-	}
+	fz_append_display_node(
+		ctx,
+		dev,
+		FZ_CMD_BEGIN_GROUP,
+		flags,
+		&rect,
+		NULL, /* path */
+		color, /* color */
+		colorspace, /* colorspace */
+		&alpha, /* alpha */
+		NULL, /* ctm */
+		NULL, /* stroke */
+		NULL, /* private_data */
+		0); /* private_data_len */
 }
 
 static void
@@ -1275,6 +1370,99 @@ fz_list_end_layer(fz_context *ctx, fz_device *dev)
 		0); /* private_data_len */
 }
 
+static void
+fz_list_begin_structure(fz_context *ctx, fz_device *dev, fz_structure standard, const char *raw, int uid)
+{
+	unsigned char *data;
+	size_t len = (raw ? strlen(raw) : 0);
+
+	data = fz_append_display_node(
+		ctx,
+		dev,
+		FZ_CMD_BEGIN_STRUCTURE,
+		0, /* flags */
+		NULL,
+		NULL, /* path */
+		NULL, /* color */
+		NULL, /* colorspace */
+		NULL, /* alpha */
+		NULL,
+		NULL, /* stroke */
+		NULL, /* private_data */
+		len+2+sizeof(uid)); /* private_data_len */
+	data[0] = (char)standard;
+	memcpy(data+1, &uid, sizeof(uid));
+	if (len)
+		memcpy(data+1+sizeof(uid), raw, len+1);
+	else
+		data[1+sizeof(uid)] = 0;
+}
+
+static void
+fz_list_end_structure(fz_context *ctx, fz_device *dev)
+{
+	fz_append_display_node(
+		ctx,
+		dev,
+		FZ_CMD_END_STRUCTURE,
+		0, /* flags */
+		NULL,
+		NULL, /* path */
+		NULL, /* color */
+		NULL, /* colorspace */
+		NULL, /* alpha */
+		NULL, /* ctm */
+		NULL, /* stroke */
+		NULL, /* private_data */
+		0); /* private_data_len */
+}
+
+static void
+fz_list_begin_metatext(fz_context *ctx, fz_device *dev, fz_metatext meta, const char *text)
+{
+	unsigned char *data;
+	size_t len = (text ? strlen(text) : 0);
+
+	data = fz_append_display_node(
+		ctx,
+		dev,
+		FZ_CMD_BEGIN_METATEXT,
+		0, /* flags */
+		NULL,
+		NULL, /* path */
+		NULL, /* color */
+		NULL, /* colorspace */
+		NULL, /* alpha */
+		NULL,
+		NULL, /* stroke */
+		NULL, /* private_data */
+		len + 2); /* private_data_len */
+	data[0] = (char)meta;
+	if (len)
+		memcpy(data+1, text, len+1);
+	else
+		data[1] = 0;
+}
+
+static void
+fz_list_end_metatext(fz_context *ctx, fz_device *dev)
+{
+	fz_append_display_node(
+		ctx,
+		dev,
+		FZ_CMD_END_METATEXT,
+		0, /* flags */
+		NULL,
+		NULL, /* path */
+		NULL, /* color */
+		NULL, /* colorspace */
+		NULL, /* alpha */
+		NULL, /* ctm */
+		NULL, /* stroke */
+		NULL, /* private_data */
+		0); /* private_data_len */
+}
+
 static void
 fz_list_drop_device(fz_context *ctx, fz_device *dev)
 {
@@ -1283,6 +1471,7 @@ fz_list_drop_device(fz_context *ctx, fz_device *dev)
 	fz_drop_colorspace(ctx, writer->colorspace);
 	fz_drop_stroke_state(ctx, writer->stroke);
 	fz_drop_path(ctx, writer->path);
+	fz_drop_display_list(ctx, writer->list);
 }
 
 fz_device *
@@ -1324,9 +1513,15 @@ fz_new_list_device(fz_context *ctx, fz_display_list *list)
 	dev->super.begin_layer = fz_list_begin_layer;
 	dev->super.end_layer = fz_list_end_layer;
 
+	dev->super.begin_structure = fz_list_begin_structure;
+	dev->super.end_structure = fz_list_end_structure;
+
+	dev->super.begin_metatext = fz_list_begin_metatext;
+	dev->super.end_metatext = fz_list_end_metatext;
+
 	dev->super.drop_device = fz_list_drop_device;
 
-	dev->list = list;
+	dev->list = fz_keep_display_list(ctx, list);
 	dev->path = NULL;
 	dev->alpha = 1.0f;
 	dev->ctm = fz_identity;
@@ -1351,7 +1546,17 @@ fz_drop_display_list_imp(fz_context *ctx, fz_storable *list_)
 	while (node != node_end)
 	{
 		fz_display_node n = *node;
-		fz_display_node *next = node + n.size;
+		size_t size = n.size;
+		fz_display_node *next;
+
+		if (size == INDIRECT_NODE_THRESHOLD)
+		{
+			memcpy(&size, &node[1], sizeof(size));
+			node += SIZE_IN_NODES(sizeof(size_t));
+			size -= SIZE_IN_NODES(sizeof(size_t));
+		}
+
+		next = node + size;
 
 		node++;
 		if (n.rect)
@@ -1376,6 +1581,7 @@ fz_drop_display_list_imp(fz_context *ctx, fz_storable *list_)
 			cs_n = 4;
 			break;
 		case CS_OTHER_0:
+			align_node_for_pointer(&node);
 			cs = *(fz_colorspace **)node;
 			cs_n = fz_colorspace_n(ctx, cs);
 			fz_drop_colorspace(ctx, cs);
@@ -1398,12 +1604,15 @@ fz_drop_display_list_imp(fz_context *ctx, fz_storable *list_)
 			node += SIZE_IN_NODES(2*sizeof(float));
 		if (n.stroke)
 		{
+			align_node_for_pointer(&node);
 			fz_drop_stroke_state(ctx, *(fz_stroke_state **)node);
 			node += SIZE_IN_NODES(sizeof(fz_stroke_state *));
 		}
 		if (n.path)
 		{
-			int path_size = fz_packed_path_size((fz_path *)node);
+			int path_size;
+			align_node_for_pointer(&node);
+			path_size = fz_packed_path_size((fz_path *)node);
 			fz_drop_path(ctx, (fz_path *)node);
 			node += SIZE_IN_NODES(path_size);
 		}
@@ -1414,20 +1623,21 @@ fz_drop_display_list_imp(fz_context *ctx, fz_storable *list_)
 		case FZ_CMD_CLIP_TEXT:
 		case FZ_CMD_CLIP_STROKE_TEXT:
 		case FZ_CMD_IGNORE_TEXT:
+			align_node_for_pointer(&node);
 			fz_drop_text(ctx, *(fz_text **)node);
 			break;
 		case FZ_CMD_FILL_SHADE:
+			align_node_for_pointer(&node);
 			fz_drop_shade(ctx, *(fz_shade **)node);
 			break;
 		case FZ_CMD_FILL_IMAGE:
 		case FZ_CMD_FILL_IMAGE_MASK:
 		case FZ_CMD_CLIP_IMAGE_MASK:
+			align_node_for_pointer(&node);
 			fz_drop_image(ctx, *(fz_image **)node);
 			break;
-		case FZ_CMD_BEGIN_GROUP:
-			fz_drop_colorspace(ctx, *(fz_colorspace **)node);
-			break;
 		case FZ_CMD_DEFAULT_COLORSPACES:
+			align_node_for_pointer(&node);
 			fz_drop_default_colorspaces(ctx, *(fz_default_colorspaces **)node);
 			break;
 		}
@@ -1513,8 +1723,16 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 	{
 		int empty;
 		fz_display_node n = *node;
+		size_t size = n.size;
 
-		next_node = node + n.size;
+		if (size == INDIRECT_NODE_THRESHOLD)
+		{
+			memcpy(&size, &node[1], sizeof(size_t));
+			node += SIZE_IN_NODES(sizeof(size_t));
+			size -= SIZE_IN_NODES(sizeof(size_t));
+		}
+
+		next_node = node + size;
 
 		/* Check the cookie for aborting */
 		if (cookie)
@@ -1522,7 +1740,7 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 			if (cookie->abort)
 				break;
 			cookie->progress = progress;
-			progress += n.size;
+			progress += (int)size;
 		}
 
 		node++;
@@ -1574,6 +1792,7 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 				color[3] = 1.0f;
 				break;
 			case CS_OTHER_0:
+				align_node_for_pointer(&node);
 				colorspace = fz_keep_colorspace(ctx, *(fz_colorspace **)(node));
 				node += SIZE_IN_NODES(sizeof(fz_colorspace *));
 				en = fz_colorspace_n(ctx, colorspace);
@@ -1629,12 +1848,14 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 		}
 		if (n.stroke)
 		{
+			align_node_for_pointer(&node);
 			fz_drop_stroke_state(ctx, stroke);
 			stroke = fz_keep_stroke_state(ctx, *(fz_stroke_state **)node);
 			node += SIZE_IN_NODES(sizeof(fz_stroke_state *));
 		}
 		if (n.path)
 		{
+			align_node_for_pointer(&node);
 			fz_drop_path(ctx, path);
 			path = fz_keep_path(ctx, (fz_path *)node);
 			node += SIZE_IN_NODES(fz_packed_path_size(path));
@@ -1657,10 +1878,26 @@ fz_run_display_list(fz_context *ctx, fz_display_list *list, fz_device *dev, fz_m
 		if (tiled ||
 			n.cmd == FZ_CMD_BEGIN_TILE || n.cmd == FZ_CMD_END_TILE ||
 			n.cmd == FZ_CMD_RENDER_FLAGS || n.cmd == FZ_CMD_DEFAULT_COLORSPACES ||
-			n.cmd == FZ_CMD_BEGIN_LAYER || n.cmd == FZ_CMD_END_LAYER)
+			n.cmd == FZ_CMD_BEGIN_LAYER || n.cmd == FZ_CMD_END_LAYER ||
+			n.cmd == FZ_CMD_BEGIN_STRUCTURE || n.cmd == FZ_CMD_END_STRUCTURE ||
+			n.cmd == FZ_CMD_BEGIN_METATEXT || n.cmd == FZ_CMD_END_METATEXT
+			)
 		{
 			empty = 0;
 		}
+		else if (n.cmd == FZ_CMD_FILL_PATH || n.cmd == FZ_CMD_STROKE_PATH)
+		{
+			/* Zero area paths are suitable for stroking. */
+			empty = !fz_is_valid_rect(fz_intersect_rect(trans_rect, scissor));
+		}
+		else if (n.cmd == FZ_CMD_FILL_TEXT || n.cmd == FZ_CMD_STROKE_TEXT ||
+			n.cmd == FZ_CMD_CLIP_TEXT || n.cmd == FZ_CMD_CLIP_STROKE_TEXT)
+		{
+			/* Zero area text (such as spaces) should be passed
+			 * through. Text that is completely outside the scissor
+			 * can be elided. */
+			empty = !fz_is_valid_rect(fz_intersect_rect(trans_rect, scissor));
+		}
 		else
 		{
 			empty = fz_is_empty_rect(fz_intersect_rect(trans_rect, scissor));
@@ -1717,34 +1954,43 @@ visible:
 				break;
 			case FZ_CMD_FILL_TEXT:
 				fz_unpack_color_params(&color_params, n.flags);
+				align_node_for_pointer(&node);
 				fz_fill_text(ctx, dev, *(fz_text **)node, trans_ctm, colorspace, color, alpha, color_params);
 				break;
 			case FZ_CMD_STROKE_TEXT:
 				fz_unpack_color_params(&color_params, n.flags);
+				align_node_for_pointer(&node);
 				fz_stroke_text(ctx, dev, *(fz_text **)node, stroke, trans_ctm, colorspace, color, alpha, color_params);
 				break;
 			case FZ_CMD_CLIP_TEXT:
+				align_node_for_pointer(&node);
 				fz_clip_text(ctx, dev, *(fz_text **)node, trans_ctm, trans_rect);
 				break;
 			case FZ_CMD_CLIP_STROKE_TEXT:
+				align_node_for_pointer(&node);
 				fz_clip_stroke_text(ctx, dev, *(fz_text **)node, stroke, trans_ctm, trans_rect);
 				break;
 			case FZ_CMD_IGNORE_TEXT:
+				align_node_for_pointer(&node);
 				fz_ignore_text(ctx, dev, *(fz_text **)node, trans_ctm);
 				break;
 			case FZ_CMD_FILL_SHADE:
 				fz_unpack_color_params(&color_params, n.flags);
+				align_node_for_pointer(&node);
 				fz_fill_shade(ctx, dev, *(fz_shade **)node, trans_ctm, alpha, color_params);
 				break;
 			case FZ_CMD_FILL_IMAGE:
 				fz_unpack_color_params(&color_params, n.flags);
+				align_node_for_pointer(&node);
 				fz_fill_image(ctx, dev, *(fz_image **)node, trans_ctm, alpha, color_params);
 				break;
 			case FZ_CMD_FILL_IMAGE_MASK:
 				fz_unpack_color_params(&color_params, n.flags);
+				align_node_for_pointer(&node);
 				fz_fill_image_mask(ctx, dev, *(fz_image **)node, trans_ctm, colorspace, color, alpha, color_params);
 				break;
 			case FZ_CMD_CLIP_IMAGE_MASK:
+				align_node_for_pointer(&node);
 				fz_clip_image_mask(ctx, dev, *(fz_image **)node, trans_ctm, trans_rect);
 				break;
 			case FZ_CMD_POP_CLIP:
@@ -1758,7 +2004,7 @@ visible:
 				fz_end_mask(ctx, dev);
 				break;
 			case FZ_CMD_BEGIN_GROUP:
-				fz_begin_group(ctx, dev, trans_rect, *(fz_colorspace **)node, (n.flags & ISOLATED) != 0, (n.flags & KNOCKOUT) != 0, (n.flags>>2), alpha);
+				fz_begin_group(ctx, dev, trans_rect, colorspace, (n.flags & ISOLATED) != 0, (n.flags & KNOCKOUT) != 0, (n.flags>>2), alpha);
 				break;
 			case FZ_CMD_END_GROUP:
 				fz_end_group(ctx, dev);
@@ -1766,8 +2012,10 @@ visible:
 			case FZ_CMD_BEGIN_TILE:
 			{
 				int cached;
-				fz_list_tile_data *data = (fz_list_tile_data *)node;
+				fz_list_tile_data *data;
 				fz_rect tile_rect;
+				align_node_for_pointer(&node);
+				data = (fz_list_tile_data *)node;
 				tiled++;
 				tile_rect = data->view;
 				cached = fz_begin_tile_id(ctx, dev, rect, tile_rect, data->xstep, data->ystep, trans_ctm, data->id);
@@ -1786,14 +2034,41 @@ visible:
 					fz_render_flags(ctx, dev, FZ_DEVFLAG_GRIDFIT_AS_TILED, 0);
 				break;
 			case FZ_CMD_DEFAULT_COLORSPACES:
+				align_node_for_pointer(&node);
 				fz_set_default_colorspaces(ctx, dev, *(fz_default_colorspaces **)node);
 				break;
 			case FZ_CMD_BEGIN_LAYER:
+				align_node_for_pointer(&node);
 				fz_begin_layer(ctx, dev, (const char *)node);
 				break;
 			case FZ_CMD_END_LAYER:
 				fz_end_layer(ctx, dev);
 				break;
+			case FZ_CMD_BEGIN_STRUCTURE:
+			{
+				const unsigned char *data;
+				int uid;
+				align_node_for_pointer(&node);
+				data = (const unsigned char *)node;
+				memcpy(&uid, data+1, sizeof(uid));
+				fz_begin_structure(ctx, dev, (fz_structure)data[0], (const char *)(data[1+sizeof(uid)] == 0 ? NULL : &data[1+sizeof(uid)]), uid);
+				break;
+			}
+			case FZ_CMD_END_STRUCTURE:
+				fz_end_structure(ctx, dev);
+				break;
+			case FZ_CMD_BEGIN_METATEXT:
+			{
+				const unsigned char *data;
+				const char *text;
+				data = (const unsigned char *)node;
+				text = (const char *)&data[1];
+				fz_begin_metatext(ctx, dev, (fz_metatext)data[0], (text[0] == 0 ? NULL : text));
+				break;
+			}
+			case FZ_CMD_END_METATEXT:
+				fz_end_metatext(ctx, dev);
+				break;
 			}
 		}
 		fz_catch(ctx)
