@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
@@ -392,18 +392,26 @@ copy_jpx_to_pixmap(fz_context *ctx, fz_pixmap *img, opj_image_t *jpx)
 		int32_t oy = safe_mul32(ctx, comp->y0, cdy) - jpx->y0;
 		int32_t ox = safe_mul32(ctx, comp->x0, cdx) - jpx->x0;
 		unsigned char *dst0 = dst + oy * stride;
+		int prec = comp->prec;
+		int sgnd = comp->sgnd;
 
 		if (comp->data == NULL)
 			fz_throw(ctx, FZ_ERROR_GENERIC, "No data for JP2 image component %d", k);
 
+		if (fz_colorspace_is_indexed(ctx, img->colorspace))
+		{
+			prec = 8; /* Don't do any scaling! */
+			sgnd = 0;
+		}
+
 		/* Check that none of the following will overflow. */
 		(void)safe_mla32(ctx, ch, cdy, oy);
 		(void)safe_mla32(ctx, cw, cdx, ox);
 
 		if (cdx == 1 && cdy == 1)
-			template_copy_comp(dst0, w, h, stride, comp->data, ox, oy, 1 /*cdx*/, 1 /*cdy*/, cw, ch, comp->sgnd, comp->prec, comps);
+			template_copy_comp(dst0, w, h, stride, comp->data, ox, oy, 1 /*cdx*/, 1 /*cdy*/, cw, ch, sgnd, prec, comps);
 		else
-			template_copy_comp(dst0, w, h, stride, comp->data, ox, oy, cdx, cdy, cw, ch, comp->sgnd, comp->prec, comps);
+			template_copy_comp(dst0, w, h, stride, comp->data, ox, oy, cdx, cdy, cw, ch, sgnd, prec, comps);
 		dst++;
 	}
 }
@@ -522,7 +530,7 @@ jpx_read_image(fz_context *ctx, fz_jpxd *state, const unsigned char *data, size_
 	}
 
 #if FZ_ENABLE_ICC
-	if (!state->cs && jpx->icc_profile_buf)
+	if (!state->cs && jpx->icc_profile_buf && jpx->icc_profile_len > 0)
 	{
 		fz_buffer *cbuf = NULL;
 		fz_var(cbuf);
