@@ -1,5 +1,29 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 
+#include "pixmap-imp.h"
+
 #ifdef HAVE_JPEGXR
 
 #include <math.h>
@@ -282,7 +306,8 @@ jxr_read_image(fz_context *ctx, const unsigned char *data, int size, struct info
 	jxr_container_t container;
 	jxr_image_t image = NULL;
 	jxr_image_t alpha = NULL;
-	int rc, i;
+	size_t i;
+	int rc;
 
 	fz_var(image);
 	fz_var(alpha);
@@ -300,10 +325,10 @@ jxr_read_image(fz_context *ctx, const unsigned char *data, int size, struct info
 		info->width = jxrc_image_width(container, 0);
 		info->height = jxrc_image_height(container, 0);
 
-		info->format = jxrc_image_pixelformat(container, 0);
+		info->format = (int) jxrc_image_pixelformat(container, 0);
 
 		for (i = 0; i < nelem(pixelformats); i++)
-			if (pixelformats[i].format == info->format)
+			if ((int) pixelformats[i].format == info->format)
 			{
 				info->comps = pixelformats[i].comps;
 				break;
