@@ -17,11 +17,12 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
+#include <assert.h>
 #include <string.h>
 #include <limits.h>
 
@@ -272,6 +273,7 @@ pdfocr_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 
 	if (sh == 0)
 		sh = h;
+	assert(sh != 0 && "pdfocr_write_header() should not be given zero height input.");
 	strips = (h + sh-1)/sh;
 
 	if (a != 0)
