@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/ucdn.h"
@@ -86,6 +86,7 @@ void fz_add_layout_char(fz_context *ctx, fz_layout_block *block, float x, float
 #define PARAGRAPH_DIST 1.5f
 #define SPACE_DIST 0.15f
 #define SPACE_MAX_DIST 0.8f
+#define BASE_MAX_DIST 0.8f
 
 typedef struct
 {
@@ -109,6 +110,7 @@ const char *fz_stext_options_usage =
 	"\tpreserve-whitespace: do not convert all whitespace into space characters\n"
 	"\tpreserve-spans: do not merge spans on the same line\n"
 	"\tdehyphenate: attempt to join up hyphenated words\n"
+	"\tuse-cid-for-unknown-unicode: guess unicode from cid if normal mapping fails\n"
 	"\tmediabox-clip=no: include characters outside mediabox\n"
 	"\n";
 
@@ -237,9 +239,8 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, fz_m
 	}
 	else
 	{
-		fz_rect bbox = fz_font_bbox(ctx, font);
-		a.x = bbox.x1;
-		d.x = bbox.x0;
+		a.x = 1;
+		d.x = 0;
 		a.y = 0;
 		d.y = 0;
 	}
@@ -268,7 +269,9 @@ remove_last_char(fz_context *ctx, fz_stext_line *line)
 		}
 		if (prev)
 		{
-			/* the characters are pool allocated, so we don't actually leak the removed node */
+			/* The characters are pool allocated, so we don't actually leak the removed node. */
+			/* We do need to drop the char's font reference though. */
+			fz_drop_font(ctx, prev->next->font);
 			line->last_char = prev;
 			line->last_char->next = NULL;
 		}
@@ -320,6 +323,75 @@ vec_dot(const fz_point *a, const fz_point *b)
 	return a->x * b->x + a->y * b->y;
 }
 
+static void
+prepend_line_if_possible(fz_context *ctx, fz_stext_page *page, fz_stext_block *cur_block, fz_matrix trm, fz_font *font, float size, int c, fz_point *pen, int color, int no_space)
+{
+	fz_stext_line *cur_line;
+	fz_stext_line *line;
+	fz_point ndir;
+	float cur_size;
+	fz_point p;
+	fz_point delta;
+	float spacing, perp;
+
+	if (cur_block == NULL || cur_block->type != FZ_STEXT_BLOCK_TEXT)
+		return;
+
+	cur_line = cur_block->u.t.last_line;
+	if (cur_line == NULL)
+		return;
+
+	line = cur_line->prev;
+	if (line == NULL)
+		return;
+
+	if (line->wmode != cur_line->wmode)
+		return;
+
+	ndir = cur_line->dir;
+	cur_size = cur_line->last_char->size;
+	p = line->first_char->origin;
+	delta.x = p.x - pen->x;
+	delta.y = p.y - pen->y;
+
+	spacing = ndir.x * delta.x + ndir.y * delta.y;
+	perp = ndir.x * delta.y - ndir.y * delta.x;
+
+	/* If cur_line overlaps line by more than a small amount, can't prepend it. */
+	if (spacing < -size * SPACE_DIST)
+		return;
+	/* If cur_line is a long way behind line, can't prepend it. */
+	if (spacing >= size * SPACE_MAX_DIST)
+		return;
+	/* If cur_line is not pretty much in line with line, can't prepend it. */
+	if (fabsf(perp) >= size * BASE_MAX_DIST)
+		return;
+
+	/* So we can prepend. Do we need to add a space? Match the sizing logic
+	 * that would happend with normal character addition. */
+	if (spacing >= cur_size * SPACE_DIST && cur_line->last_char->c != ' ' && cur_line->wmode == 0 && !no_space)
+	{
+		/* We need to add a space onto the end of the prepended line. */
+		add_char_to_line(ctx, page, cur_line, trm, font, size, ' ', pen, &p, color);
+	}
+
+	/* cur_line plausibly finishes at the start of line. */
+	/* Move all the chars from cur_line onto the start of line */
+	cur_line->last_char->next = line->first_char;
+	line->first_char = cur_line->first_char;
+	cur_line->first_char = NULL;
+	cur_line->last_char = NULL;
+
+	/* Merge the bboxes */
+	line->bbox = fz_union_rect(line->bbox, cur_line->bbox);
+
+	/* Unlink cur_line from the block. */
+	cur_block->u.t.last_line = cur_block->u.t.last_line->prev;
+	cur_block->u.t.last_line->next = NULL;
+
+	/* Can't bin the line storage as it's from a pool. */
+}
+
 static void
 fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix trm, float adv, int wmode, int force_new_line)
 {
@@ -335,7 +407,6 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 	fz_point delta;
 	float spacing = 0;
 	float base_offset = 0;
-	int rtl = 0;
 
 	dev->curdir = direction_from_bidi_class(ucdn_get_bidi_class(c), dev->curdir);
 
@@ -426,7 +497,7 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		base_offset = -ndir.y * delta.x + ndir.x * delta.y;
 
 		/* Only a small amount off the baseline - we'll take this */
-		if (fabsf(base_offset) < size * 0.8f)
+		if (fabsf(base_offset) < size * BASE_MAX_DIST)
 		{
 			/* LTR or neutral character */
 			if (dev->curdir >= 0)
@@ -459,10 +530,12 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 			else
 			{
 				new_line = 0;
+#if 0 /* TODO: handle RTL visual/logical ordering */
 				if (spacing > size * SPACE_DIST || spacing < 0)
 					rtl = 0; /* backward (or big jump to 'right' side) means logical order */
 				else
 					rtl = 1; /* visual order, we need to reverse in a post process pass */
+#endif
 			}
 		}
 
@@ -497,6 +570,14 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		new_line = 0;
 	}
 
+	if (new_line)
+	{
+		/* We are about to start a new line. This means we've finished with this
+		 * one. Can this be prepended to a previous line in this block? */
+		/* dev->pen records the previous stopping point - so where cur_line ends. */
+		prepend_line_if_possible(ctx, page, cur_block, trm, font, size, ' ', &dev->pen, dev->color, (dev->flags & FZ_STEXT_INHIBIT_SPACES));
+	}
+
 	/* Start a new line */
 	if (new_line || !cur_line || force_new_line)
 	{
@@ -516,6 +597,18 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 	dev->trm = trm;
 }
 
+static void
+flush_text(fz_context *ctx, fz_stext_device *dev)
+{
+	fz_stext_page *page = dev->page;
+
+	float size = fz_matrix_expansion(dev->trm);
+
+	/* Find current position to enter new text. */
+	if (dev->lasttext && dev->lasttext->tail)
+		prepend_line_if_possible(ctx, page, page->last_block, dev->trm, dev->lasttext->tail->font, size, ' ', &dev->pen, dev->color, (dev->flags & FZ_STEXT_INHIBIT_SPACES));
+}
+
 static void
 fz_add_stext_char(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix trm, float adv, int wmode, int force_new_line)
 {
@@ -594,6 +687,7 @@ fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, fz_m
 	fz_matrix tm = span->trm;
 	fz_matrix trm;
 	float adv;
+	int unicode;
 	int i;
 
 	if (span->len == 0)
@@ -616,8 +710,12 @@ fz_stext_extract(fz_context *ctx, fz_stext_device *dev, fz_text_span *span, fz_m
 		else
 			adv = 0;
 
+		unicode = span->items[i].ucs;
+		if (unicode == FZ_REPLACEMENT_CHARACTER && (dev->flags & FZ_STEXT_USE_CID_FOR_UNKNOWN_UNICODE))
+			unicode = span->items[i].cid;
+
 		fz_add_stext_char(ctx, dev, font,
-			span->items[i].ucs,
+			unicode,
 			span->items[i].gid,
 			trm,
 			adv,
@@ -794,6 +892,8 @@ fz_stext_close_device(fz_context *ctx, fz_device *dev)
 	fz_stext_line *line;
 	fz_stext_char *ch;
 
+	flush_text(ctx, tdev);
+
 	for (block = page->first_block; block; block = block->next)
 	{
 		if (block->type != FZ_STEXT_BLOCK_TEXT)
@@ -842,11 +942,17 @@ fz_parse_stext_options(fz_context *ctx, fz_stext_options *opts, const char *stri
 		opts->flags |= FZ_STEXT_DEHYPHENATE;
 	if (fz_has_option(ctx, string, "preserve-spans", &val) && fz_option_eq(val, "yes"))
 		opts->flags |= FZ_STEXT_PRESERVE_SPANS;
+	if (fz_has_option(ctx, string, "use-cid-for-unknown-unicode", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_USE_CID_FOR_UNKNOWN_UNICODE;
 
 	opts->flags |= FZ_STEXT_MEDIABOX_CLIP;
 	if (fz_has_option(ctx, string, "mediabox-clip", &val) && fz_option_eq(val, "no"))
 		opts->flags ^= FZ_STEXT_MEDIABOX_CLIP;
 
+	opts->scale = 1;
+	if (fz_has_option(ctx, string, "resolution", &val))
+		opts->scale = fz_atof(val) / 96.0f; /* HTML base resolution is 96ppi */
+
 	return opts;
 }
 
