@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "html-imp.h"
 
@@ -6,29 +28,6 @@
 #include <stdio.h>
 #include <assert.h>
 
-static const char *inherit_list[] = {
-	"color",
-	"direction",
-	"font-family",
-	"font-style",
-	"font-variant",
-	"font-weight",
-	"letter-spacing",
-	"line-height",
-	"list-style-image",
-	"list-style-position",
-	"list-style-type",
-	"orphans",
-	"quotes",
-	"text-align",
-	"text-indent",
-	"text-transform",
-	"visibility",
-	"white-space",
-	"widows",
-	"word-spacing",
-};
-
 static const char *border_width_kw[] = {
 	"medium",
 	"thick",
@@ -243,15 +242,15 @@ match_att_has_condition(fz_xml *node, const char *att, const char *needle)
 	const char *ss;
 	size_t n;
 	if (haystack) {
-		/* Try matching whole property first. */
-		if (!strcmp(haystack, needle))
-			return 1;
-
-		/* Look for matching words. */
-		n = strlen(needle);
 		ss = strstr(haystack, needle);
-		if (ss && (ss[n] == ' ' || ss[n] == 0) && (ss == haystack || ss[-1] == ' '))
-			return 1;
+		if (ss)
+		{
+			n = strlen(needle);
+
+			/* Look for exact matches or matching words. */
+			if ((ss[n] == ' ' || ss[n] == 0) && (ss == haystack || ss[-1] == ' '))
+				return 1;
+		}
 	}
 	return 0;
 }
@@ -288,11 +287,13 @@ match_selector(fz_css_selector *sel, fz_xml *node)
 		if (sel->combine == ' ')
 		{
 			fz_xml *parent = fz_xml_up(node);
+			if (!parent || !match_selector(sel->right, node))
+				return 0;
+
 			while (parent)
 			{
 				if (match_selector(sel->left, parent))
-					if (match_selector(sel->right, node))
-						return 1;
+					return 1;
 				parent = fz_xml_up(parent);
 			}
 			return 0;
@@ -358,11 +359,11 @@ count_values(fz_css_value *value)
 	return n;
 }
 
-static void add_property(fz_css_match *match, const char *name, fz_css_value *value, int spec);
+static void add_property(fz_css_match *match, int name, fz_css_value *value, int spec);
 
 static void
 add_shorthand_trbl(fz_css_match *match, fz_css_value *value, int spec,
-	const char *name_t, const char *name_r, const char *name_b, const char *name_l)
+	int name_t, int name_r, int name_b, int name_l)
 {
 	int n = count_values(value);
 
@@ -415,35 +416,50 @@ static void
 add_shorthand_margin(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"margin-top", "margin-right", "margin-bottom", "margin-left");
+		PRO_MARGIN_TOP,
+		PRO_MARGIN_RIGHT,
+		PRO_MARGIN_BOTTOM,
+		PRO_MARGIN_LEFT);
 }
 
 static void
 add_shorthand_padding(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"padding-top", "padding-right", "padding-bottom", "padding-left");
+		PRO_PADDING_TOP,
+		PRO_PADDING_RIGHT,
+		PRO_PADDING_BOTTOM,
+		PRO_PADDING_LEFT);
 }
 
 static void
 add_shorthand_border_width(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"border-top-width", "border-right-width", "border-bottom-width", "border-left-width");
+		PRO_BORDER_TOP_WIDTH,
+		PRO_BORDER_RIGHT_WIDTH,
+		PRO_BORDER_BOTTOM_WIDTH,
+		PRO_BORDER_LEFT_WIDTH);
 }
 
 static void
 add_shorthand_border_color(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"border-top-color", "border-right-color", "border-bottom-color", "border-left-color");
+		PRO_BORDER_TOP_COLOR,
+		PRO_BORDER_RIGHT_COLOR,
+		PRO_BORDER_BOTTOM_COLOR,
+		PRO_BORDER_LEFT_COLOR);
 }
 
 static void
 add_shorthand_border_style(fz_css_match *match, fz_css_value *value, int spec)
 {
 	add_shorthand_trbl(match, value, spec,
-		"border-top-style", "border-right-style", "border-bottom-style", "border-left-style");
+		PRO_BORDER_TOP_STYLE,
+		PRO_BORDER_RIGHT_STYLE,
+		PRO_BORDER_BOTTOM_STYLE,
+		PRO_BORDER_LEFT_STYLE);
 }
 
 static void
@@ -453,41 +469,41 @@ add_shorthand_border(fz_css_match *match, fz_css_value *value, int spec, int T,
 	{
 		if (value->type == CSS_HASH)
 		{
-			if (T) add_property(match, "border-top-color", value, spec);
-			if (R) add_property(match, "border-right-color", value, spec);
-			if (B) add_property(match, "border-bottom-color", value, spec);
-			if (L) add_property(match, "border-left-color", value, spec);
+			if (T) add_property(match, PRO_BORDER_TOP_COLOR, value, spec);
+			if (R) add_property(match, PRO_BORDER_RIGHT_COLOR, value, spec);
+			if (B) add_property(match, PRO_BORDER_BOTTOM_COLOR, value, spec);
+			if (L) add_property(match, PRO_BORDER_LEFT_COLOR, value, spec);
 		}
 		else if (value->type == CSS_KEYWORD)
 		{
 			if (keyword_in_list(value->data, border_width_kw, nelem(border_width_kw)))
 			{
-				if (T) add_property(match, "border-top-width", value, spec);
-				if (R) add_property(match, "border-right-width", value, spec);
-				if (B) add_property(match, "border-bottom-width", value, spec);
-				if (L) add_property(match, "border-left-width", value, spec);
+				if (T) add_property(match, PRO_BORDER_TOP_WIDTH, value, spec);
+				if (R) add_property(match, PRO_BORDER_RIGHT_WIDTH, value, spec);
+				if (B) add_property(match, PRO_BORDER_BOTTOM_WIDTH, value, spec);
+				if (L) add_property(match, PRO_BORDER_LEFT_WIDTH, value, spec);
 			}
 			else if (keyword_in_list(value->data, border_style_kw, nelem(border_style_kw)))
 			{
-				if (T) add_property(match, "border-top-style", value, spec);
-				if (R) add_property(match, "border-right-style", value, spec);
-				if (B) add_property(match, "border-bottom-style", value, spec);
-				if (L) add_property(match, "border-left-style", value, spec);
+				if (T) add_property(match, PRO_BORDER_TOP_STYLE, value, spec);
+				if (R) add_property(match, PRO_BORDER_RIGHT_STYLE, value, spec);
+				if (B) add_property(match, PRO_BORDER_BOTTOM_STYLE, value, spec);
+				if (L) add_property(match, PRO_BORDER_LEFT_STYLE, value, spec);
 			}
 			else if (keyword_in_list(value->data, color_kw, nelem(color_kw)))
 			{
-				if (T) add_property(match, "border-top-color", value, spec);
-				if (R) add_property(match, "border-right-color", value, spec);
-				if (B) add_property(match, "border-bottom-color", value, spec);
-				if (L) add_property(match, "border-left-color", value, spec);
+				if (T) add_property(match, PRO_BORDER_TOP_COLOR, value, spec);
+				if (R) add_property(match, PRO_BORDER_RIGHT_COLOR, value, spec);
+				if (B) add_property(match, PRO_BORDER_BOTTOM_COLOR, value, spec);
+				if (L) add_property(match, PRO_BORDER_LEFT_COLOR, value, spec);
 			}
 		}
 		else
 		{
-			if (T) add_property(match, "border-top-width", value, spec);
-			if (R) add_property(match, "border-right-width", value, spec);
-			if (B) add_property(match, "border-bottom-width", value, spec);
-			if (L) add_property(match, "border-left-width", value, spec);
+			if (T) add_property(match, PRO_BORDER_TOP_WIDTH, value, spec);
+			if (R) add_property(match, PRO_BORDER_RIGHT_WIDTH, value, spec);
+			if (B) add_property(match, PRO_BORDER_BOTTOM_WIDTH, value, spec);
+			if (L) add_property(match, PRO_BORDER_LEFT_WIDTH, value, spec);
 		}
 		value = value->next;
 	}
@@ -502,11 +518,11 @@ add_shorthand_list_style(fz_css_match *match, fz_css_value *value, int spec)
 		{
 			if (keyword_in_list(value->data, list_style_type_kw, nelem(list_style_type_kw)))
 			{
-				add_property(match, "list-style-type", value, spec);
+				add_property(match, PRO_LIST_STYLE_TYPE, value, spec);
 			}
 			else if (keyword_in_list(value->data, list_style_position_kw, nelem(list_style_position_kw)))
 			{
-				add_property(match, "list-style-position", value, spec);
+				add_property(match, PRO_LIST_STYLE_POSITION, value, spec);
 			}
 		}
 		value = value->next;
@@ -514,126 +530,70 @@ add_shorthand_list_style(fz_css_match *match, fz_css_value *value, int spec)
 }
 
 static void
-add_property(fz_css_match *match, const char *name, fz_css_value *value, int spec)
+add_property(fz_css_match *match, int name, fz_css_value *value, int spec)
 {
-	int i;
-
-	if (!strcmp(name, "margin"))
+	/* shorthand expansions: */
+	switch (name)
 	{
+	case PRO_MARGIN:
 		add_shorthand_margin(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "padding"))
-	{
+	case PRO_PADDING:
 		add_shorthand_padding(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border-width"))
-	{
+	case PRO_BORDER_WIDTH:
 		add_shorthand_border_width(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border-color"))
-	{
+	case PRO_BORDER_COLOR:
 		add_shorthand_border_color(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border-style"))
-	{
+	case PRO_BORDER_STYLE:
 		add_shorthand_border_style(match, value, spec);
 		return;
-	}
-	if (!strcmp(name, "border"))
-	{
+	case PRO_BORDER:
 		add_shorthand_border(match, value, spec, 1, 1, 1, 1);
 		return;
-	}
-	if (!strcmp(name, "border-top"))
-	{
+	case PRO_BORDER_TOP:
 		add_shorthand_border(match, value, spec, 1, 0, 0, 0);
 		return;
-	}
-	if (!strcmp(name, "border-right"))
-	{
+	case PRO_BORDER_RIGHT:
 		add_shorthand_border(match, value, spec, 0, 1, 0, 0);
 		return;
-	}
-	if (!strcmp(name, "border-bottom"))
-	{
+	case PRO_BORDER_BOTTOM:
 		add_shorthand_border(match, value, spec, 0, 0, 1, 0);
 		return;
-	}
-	if (!strcmp(name, "border-left"))
-	{
+	case PRO_BORDER_LEFT:
 		add_shorthand_border(match, value, spec, 0, 0, 0, 1);
 		return;
-	}
-	if (!strcmp(name, "list-style"))
-	{
+	case PRO_LIST_STYLE:
 		add_shorthand_list_style(match, value, spec);
 		return;
-	}
-
-	/* shorthand expansions: */
-	/* TODO: border-color */
-	/* TODO: border-style */
 	/* TODO: font */
-	/* TODO: list-style */
 	/* TODO: background */
-
-	for (i = 0; i < match->count; ++i)
-	{
-		if (!strcmp(match->prop[i].name, name))
-		{
-			if (match->prop[i].spec <= spec)
-			{
-				match->prop[i].value = value;
-				match->prop[i].spec = spec;
-			}
-			return;
-		}
 	}
 
-	if (match->count + 1 >= (int)nelem(match->prop))
+	if (name < NUM_PROPERTIES && match->spec[name] <= spec)
 	{
-		// fz_warn(ctx, "too many css properties");
-		return;
-	}
-
-	match->prop[match->count].name = name;
-	match->prop[match->count].value = value;
-	match->prop[match->count].spec = spec;
-	++match->count;
-}
-
-static void
-sort_properties(fz_css_match *match)
-{
-	int count = match->count;
-	fz_css_match_prop *prop = match->prop;
-	int i, k;
-
-	/* Insertion sort. */
-	for (i = 1; i < count; ++i)
-	{
-		k = i;
-		while (k > 0 && strcmp(prop[k-1].name, prop[k].name) > 0)
-		{
-			fz_css_match_prop save = prop[k-1];
-			prop[k-1] = prop[k];
-			prop[k] = save;
-			--k;
-		}
+		match->value[name] = value;
+		match->spec[name] = spec;
 	}
 }
 
 void
-fz_match_css(fz_context *ctx, fz_css_match *match, fz_css *css, fz_xml *node)
+fz_match_css(fz_context *ctx, fz_css_match *match, fz_css_match *up, fz_css *css, fz_xml *node)
 {
 	fz_css_rule *rule;
 	fz_css_selector *sel;
 	fz_css_property *prop;
 	const char *s;
+	int i;
+
+	match->up = up;
+	for (i = 0; i < NUM_PROPERTIES; ++i)
+	{
+		match->spec[i] = -1;
+		match->value[i] = NULL;
+	}
 
 	for (rule = css->rule; rule; rule = rule->next)
 	{
@@ -671,8 +631,6 @@ fz_match_css(fz_context *ctx, fz_css_match *match, fz_css *css, fz_xml *node)
 			}
 		}
 	}
-
-	sort_properties(match); /* speed up subsequent value_from_raw_property lookups */
 }
 
 void
@@ -681,6 +639,14 @@ fz_match_css_at_page(fz_context *ctx, fz_css_match *match, fz_css *css)
 	fz_css_rule *rule;
 	fz_css_selector *sel;
 	fz_css_property *prop;
+	int i;
+
+	match->up = NULL;
+	for (i = 0; i < NUM_PROPERTIES; ++i)
+	{
+		match->spec[i] = -1;
+		match->value[i] = NULL;
+	}
 
 	for (rule = css->rule; rule; rule = rule->next)
 	{
@@ -696,8 +662,6 @@ fz_match_css_at_page(fz_context *ctx, fz_css_match *match, fz_css *css)
 			sel = sel->next;
 		}
 	}
-
-	sort_properties(match); /* speed up subsequent value_from_raw_property lookups */
 }
 
 void
@@ -718,11 +682,11 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 
 	for (prop = declaration; prop; prop = prop->next)
 	{
-		if (!strcmp(prop->name, "font-family")) family = prop->value->data;
-		if (!strcmp(prop->name, "font-weight")) weight = prop->value->data;
-		if (!strcmp(prop->name, "font-style")) style = prop->value->data;
-		if (!strcmp(prop->name, "font-variant")) variant = prop->value->data;
-		if (!strcmp(prop->name, "src")) src = prop->value->data;
+		if (prop->name == PRO_FONT_FAMILY) family = prop->value->data;
+		if (prop->name == PRO_FONT_WEIGHT) weight = prop->value->data;
+		if (prop->name == PRO_FONT_STYLE) style = prop->value->data;
+		if (prop->name == PRO_FONT_VARIANT) variant = prop->value->data;
+		if (prop->name == PRO_SRC) src = prop->value->data;
 	}
 
 	if (!src)
@@ -736,7 +700,6 @@ fz_add_css_font_face(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, co
 	fz_strlcat(path, "/", sizeof path);
 	fz_strlcat(path, src, sizeof path);
 	fz_urldecode(path);
-	fz_cleanname(path);
 
 	for (custom = set->custom; custom; custom = custom->next)
 		if (!strcmp(custom->src, path) && !strcmp(custom->family, family) &&
@@ -794,48 +757,71 @@ fz_add_css_font_faces(fz_context *ctx, fz_html_font_set *set, fz_archive *zip, c
 	}
 }
 
-static fz_css_value *
-value_from_raw_property(fz_css_match *match, const char *name)
+static int
+is_inheritable_property(int name)
 {
-	fz_css_match_prop *prop = match->prop;
-	int l = 0;
-	int r = match->count - 1;
-	while (l <= r)
-	{
-		int m = (l + r) >> 1;
-		int c = strcmp(name, prop[m].name);
-		if (c < 0)
-			r = m - 1;
-		else if (c > 0)
-			l = m + 1;
-		else
-			return prop[m].value;
-	}
-	return NULL;
+	return
+		name == PRO_COLOR ||
+		name == PRO_DIRECTION ||
+		name == PRO_FONT_FAMILY ||
+		name == PRO_FONT_STYLE ||
+		name == PRO_FONT_VARIANT ||
+		name == PRO_FONT_WEIGHT ||
+		name == PRO_LEADING ||
+		name == PRO_LETTER_SPACING ||
+		name == PRO_LINE_HEIGHT ||
+		name == PRO_LIST_STYLE_IMAGE ||
+		name == PRO_LIST_STYLE_POSITION ||
+		name == PRO_LIST_STYLE_TYPE ||
+		name == PRO_ORPHANS ||
+		name == PRO_OVERFLOW_WRAP ||
+		name == PRO_QUOTES ||
+		name == PRO_TEXT_ALIGN ||
+		name == PRO_TEXT_INDENT ||
+		name == PRO_TEXT_TRANSFORM ||
+		name == PRO_VISIBILITY ||
+		name == PRO_WHITE_SPACE ||
+		name == PRO_WIDOWS ||
+		name == PRO_WORD_SPACING ||
+		// Strictly speaking, text-decoration is not an inherited property,
+		// but since when drawing an underlined element, all children are also underlined,
+		// we may as well make it inherited.
+		name == PRO_TEXT_DECORATION;
 }
 
 static fz_css_value *
-value_from_property(fz_css_match *match, const char *name)
+value_from_inheritable_property(fz_css_match *match, int name)
 {
-	fz_css_value *value;
-
-	value = value_from_raw_property(match, name);
+	fz_css_value *value = match->value[name];
 	if (match->up)
 	{
 		if (value && !strcmp(value->data, "inherit"))
-			if (strcmp(name, "font-size") != 0) /* never inherit 'font-size' textually */
+			return value_from_inheritable_property(match->up, name);
+		if (!value)
+			return value_from_inheritable_property(match->up, name);
+	}
+	return value;
+}
+
+static fz_css_value *
+value_from_property(fz_css_match *match, int name)
+{
+	fz_css_value *value = match->value[name];
+	if (match->up)
+	{
+		if (value && !strcmp(value->data, "inherit"))
+			if (name != PRO_FONT_SIZE) /* never inherit 'font-size' textually */
 				return value_from_property(match->up, name);
-		if (!value && keyword_in_list(name, inherit_list, nelem(inherit_list)))
-			return value_from_property(match->up, name);
+		if (!value && is_inheritable_property(name))
+			return value_from_inheritable_property(match->up, name);
 	}
 	return value;
 }
 
 static const char *
-string_from_property(fz_css_match *match, const char *name, const char *initial)
+string_from_property(fz_css_match *match, int name, const char *initial)
 {
-	fz_css_value *value;
-	value = value_from_property(match, name);
+	fz_css_value *value = value_from_property(match, name);
 	if (!value)
 		return initial;
 	return value->data;
@@ -850,6 +836,15 @@ make_number(float v, int u)
 	return n;
 }
 
+static fz_css_number
+make_undefined_number(void)
+{
+	fz_css_number n;
+	n.value = 0;
+	n.unit = N_UNDEFINED;
+	return n;
+}
+
 /* Fast but inaccurate strtof. */
 static float
 fz_css_strtof(char *s, char **endptr)
@@ -947,13 +942,13 @@ number_from_value(fz_css_value *value, float initial, int initial_unit)
 }
 
 static fz_css_number
-number_from_property(fz_css_match *match, const char *property, float initial, int initial_unit)
+number_from_property(fz_css_match *match, int property, float initial, int initial_unit)
 {
 	return number_from_value(value_from_property(match, property), initial, initial_unit);
 }
 
 static fz_css_number
-border_width_from_property(fz_css_match *match, const char *property)
+border_width_from_property(fz_css_match *match, int property)
 {
 	fz_css_value *value = value_from_property(match, property);
 	if (value)
@@ -970,7 +965,7 @@ border_width_from_property(fz_css_match *match, const char *property)
 }
 
 static int
-border_style_from_property(fz_css_match *match, const char *property)
+border_style_from_property(fz_css_match *match, int property)
 {
 	fz_css_value *value = value_from_property(match, property);
 	if (value)
@@ -982,6 +977,11 @@ border_style_from_property(fz_css_match *match, const char *property)
 	return BS_NONE;
 }
 
+int fz_css_number_defined(fz_css_number number)
+{
+	return number.unit != N_UNDEFINED;
+}
+
 float
 fz_from_css_number(fz_css_number number, float em, float percent_value, float auto_value)
 {
@@ -1034,7 +1034,7 @@ color_from_value(fz_css_value *value, fz_css_color initial)
 
 	if (value->type == CSS_HASH)
 	{
-		int r, g, b;
+		int r, g, b, a;
 		size_t n;
 hex_color:
 		n = strlen(value->data);
@@ -1043,18 +1043,35 @@ hex_color:
 			r = tohex(value->data[0]) * 16 + tohex(value->data[0]);
 			g = tohex(value->data[1]) * 16 + tohex(value->data[1]);
 			b = tohex(value->data[2]) * 16 + tohex(value->data[2]);
+			a = 255;
+		}
+		else if (n == 4)
+		{
+			r = tohex(value->data[0]) * 16 + tohex(value->data[0]);
+			g = tohex(value->data[1]) * 16 + tohex(value->data[1]);
+			b = tohex(value->data[2]) * 16 + tohex(value->data[2]);
+			a = tohex(value->data[3]) * 16 + tohex(value->data[3]);
 		}
 		else if (n == 6)
 		{
 			r = tohex(value->data[0]) * 16 + tohex(value->data[1]);
 			g = tohex(value->data[2]) * 16 + tohex(value->data[3]);
 			b = tohex(value->data[4]) * 16 + tohex(value->data[5]);
+			a = 255;
+		}
+		else if (n == 8)
+		{
+			r = tohex(value->data[0]) * 16 + tohex(value->data[1]);
+			g = tohex(value->data[2]) * 16 + tohex(value->data[3]);
+			b = tohex(value->data[4]) * 16 + tohex(value->data[5]);
+			a = tohex(value->data[6]) * 16 + tohex(value->data[7]);
 		}
 		else
 		{
 			r = g = b = 0;
+			a = 255;
 		}
-		return make_color(r, g, b, 255);
+		return make_color(r, g, b, a);
 	}
 
 	if (value->type == '(' && !strcmp(value->data, "rgb"))
@@ -1070,6 +1087,21 @@ hex_color:
 		return make_color(r, g, b, 255);
 	}
 
+	if (value->type == '(' && !strcmp(value->data, "rgba"))
+	{
+		fz_css_value *vr, *vg, *vb, *va;
+		int r, g, b, a;
+		vr = value->args;
+		vg = vr && vr->next ? vr->next->next : NULL; /* skip the ',' nodes */
+		vb = vg && vg->next ? vg->next->next : NULL; /* skip the ',' nodes */
+		va = vb && vb->next ? vb->next->next : NULL; /* skip the ',' nodes */
+		r = fz_from_css_number(number_from_value(vr, 0, N_NUMBER), 255, 255, 0);
+		g = fz_from_css_number(number_from_value(vg, 0, N_NUMBER), 255, 255, 0);
+		b = fz_from_css_number(number_from_value(vb, 0, N_NUMBER), 255, 255, 0);
+		a = fz_from_css_number(number_from_value(va, 0, N_NUMBER), 255, 255, 255);
+		return make_color(r, g, b, a);
+	}
+
 	if (value->type == CSS_KEYWORD)
 	{
 		if (!strcmp(value->data, "transparent"))
@@ -1114,7 +1146,7 @@ hex_color:
 }
 
 static fz_css_color
-color_from_property(fz_css_match *match, const char *property, fz_css_color initial)
+color_from_property(fz_css_match *match, int property, fz_css_color initial)
 {
 	return color_from_value(value_from_property(match, property), initial);
 }
@@ -1122,7 +1154,7 @@ color_from_property(fz_css_match *match, const char *property, fz_css_color init
 int
 fz_get_css_match_display(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, "display");
+	fz_css_value *value = value_from_property(match, PRO_DISPLAY);
 	if (value)
 	{
 		if (!strcmp(value->data, "none"))
@@ -1141,6 +1173,16 @@ fz_get_css_match_display(fz_css_match *match)
 			return DIS_TABLE_ROW;
 		if (!strcmp(value->data, "table-cell"))
 			return DIS_TABLE_CELL;
+		if (!strcmp(value->data, "table-row-group"))
+			return DIS_TABLE_GROUP;
+		if (!strcmp(value->data, "table-header-group"))
+			return DIS_TABLE_GROUP;
+		if (!strcmp(value->data, "table-footer-group"))
+			return DIS_TABLE_GROUP;
+		if (!strcmp(value->data, "table-column-group"))
+			return DIS_NONE;
+		if (!strcmp(value->data, "table-column"))
+			return DIS_NONE;
 	}
 	return DIS_INLINE;
 }
@@ -1148,7 +1190,7 @@ fz_get_css_match_display(fz_css_match *match)
 static int
 white_space_from_property(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, "white-space");
+	fz_css_value *value = value_from_property(match, PRO_WHITE_SPACE);
 	if (value)
 	{
 		if (!strcmp(value->data, "normal")) return WS_NORMAL;
@@ -1160,10 +1202,22 @@ white_space_from_property(fz_css_match *match)
 	return WS_NORMAL;
 }
 
+static int
+text_decoration_from_property(fz_css_match *match)
+{
+	fz_css_value *value = value_from_property(match, PRO_TEXT_DECORATION);
+	if (value)
+	{
+		if (!strcmp(value->data, "underline")) return TD_UNDERLINE;
+		if (!strcmp(value->data, "line-through")) return TD_LINE_THROUGH;
+	}
+	return TD_NONE;
+}
+
 static int
 visibility_from_property(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, "visibility");
+	fz_css_value *value = value_from_property(match, PRO_VISIBILITY);
 	if (value)
 	{
 		if (!strcmp(value->data, "visible")) return V_VISIBLE;
@@ -1174,7 +1228,7 @@ visibility_from_property(fz_css_match *match)
 }
 
 static int
-page_break_from_property(fz_css_match *match, char *prop)
+page_break_from_property(fz_css_match *match, int prop)
 {
 	fz_css_value *value = value_from_property(match, prop);
 	if (value)
@@ -1200,6 +1254,7 @@ fz_default_css_style(fz_context *ctx, fz_css_style *style)
 	style->font_size = make_number(1, N_SCALE);
 	style->width = make_number(0, N_AUTO);
 	style->height = make_number(0, N_AUTO);
+	style->leading = make_undefined_number();
 }
 
 void
@@ -1214,10 +1269,11 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 
 	style->visibility = visibility_from_property(match);
 	style->white_space = white_space_from_property(match);
-	style->page_break_before = page_break_from_property(match, "page-break-before");
-	style->page_break_after = page_break_from_property(match, "page-break-after");
+	style->text_decoration = text_decoration_from_property(match);
+	style->page_break_before = page_break_from_property(match, PRO_PAGE_BREAK_BEFORE);
+	style->page_break_after = page_break_from_property(match, PRO_PAGE_BREAK_AFTER);
 
-	value = value_from_property(match, "text-align");
+	value = value_from_property(match, PRO_TEXT_ALIGN);
 	if (value)
 	{
 		if (!strcmp(value->data, "left")) style->text_align = TA_LEFT;
@@ -1226,7 +1282,7 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "justify")) style->text_align = TA_JUSTIFY;
 	}
 
-	value = value_from_property(match, "vertical-align");
+	value = value_from_property(match, PRO_VERTICAL_ALIGN);
 	if (value)
 	{
 		if (!strcmp(value->data, "baseline")) style->vertical_align = VA_BASELINE;
@@ -1238,7 +1294,7 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "text-bottom")) style->vertical_align = VA_TEXT_BOTTOM;
 	}
 
-	value = value_from_property(match, "font-size");
+	value = value_from_property(match, PRO_FONT_SIZE);
 	if (value)
 	{
 		if (!strcmp(value->data, "xx-large")) style->font_size = make_number(1.73f, N_SCALE);
@@ -1257,7 +1313,7 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		style->font_size = make_number(1, N_SCALE);
 	}
 
-	value = value_from_property(match, "list-style-type");
+	value = value_from_property(match, PRO_LIST_STYLE_TYPE);
 	if (value)
 	{
 		if (!strcmp(value->data, "none")) style->list_style_type = LST_NONE;
@@ -1278,49 +1334,59 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "georgian")) style->list_style_type = LST_GEORGIAN;
 	}
 
-	style->line_height = number_from_property(match, "line-height", 1.2f, N_SCALE);
+	value = value_from_property(match, PRO_OVERFLOW_WRAP);
+	if (value)
+	{
+		if (!strcmp(value->data, "break-word")) style->overflow_wrap = OVERFLOW_WRAP_BREAK_WORD;
+		else style->overflow_wrap = OVERFLOW_WRAP_NORMAL;
+	}
 
-	style->text_indent = number_from_property(match, "text-indent", 0, N_LENGTH);
+	style->line_height = number_from_property(match, PRO_LINE_HEIGHT, 1.2f, N_SCALE);
+	style->leading = number_from_property(match, PRO_LEADING, 0, N_UNDEFINED);
 
-	style->width = number_from_property(match, "width", 0, N_AUTO);
-	style->height = number_from_property(match, "height", 0, N_AUTO);
+	style->text_indent = number_from_property(match, PRO_TEXT_INDENT, 0, N_LENGTH);
 
-	style->margin[0] = number_from_property(match, "margin-top", 0, N_LENGTH);
-	style->margin[1] = number_from_property(match, "margin-right", 0, N_LENGTH);
-	style->margin[2] = number_from_property(match, "margin-bottom", 0, N_LENGTH);
-	style->margin[3] = number_from_property(match, "margin-left", 0, N_LENGTH);
+	style->width = number_from_property(match, PRO_WIDTH, 0, N_AUTO);
+	style->height = number_from_property(match, PRO_HEIGHT, 0, N_AUTO);
 
-	style->padding[0] = number_from_property(match, "padding-top", 0, N_LENGTH);
-	style->padding[1] = number_from_property(match, "padding-right", 0, N_LENGTH);
-	style->padding[2] = number_from_property(match, "padding-bottom", 0, N_LENGTH);
-	style->padding[3] = number_from_property(match, "padding-left", 0, N_LENGTH);
+	style->margin[0] = number_from_property(match, PRO_MARGIN_TOP, 0, N_LENGTH);
+	style->margin[1] = number_from_property(match, PRO_MARGIN_RIGHT, 0, N_LENGTH);
+	style->margin[2] = number_from_property(match, PRO_MARGIN_BOTTOM, 0, N_LENGTH);
+	style->margin[3] = number_from_property(match, PRO_MARGIN_LEFT, 0, N_LENGTH);
 
-	style->color = color_from_property(match, "color", black);
-	style->background_color = color_from_property(match, "background-color", transparent);
+	style->padding[0] = number_from_property(match, PRO_PADDING_TOP, 0, N_LENGTH);
+	style->padding[1] = number_from_property(match, PRO_PADDING_RIGHT, 0, N_LENGTH);
+	style->padding[2] = number_from_property(match, PRO_PADDING_BOTTOM, 0, N_LENGTH);
+	style->padding[3] = number_from_property(match, PRO_PADDING_LEFT, 0, N_LENGTH);
 
-	style->border_style_0 = border_style_from_property(match, "border-top-style");
-	style->border_style_1 = border_style_from_property(match, "border-right-style");
-	style->border_style_2 = border_style_from_property(match, "border-bottom-style");
-	style->border_style_3 = border_style_from_property(match, "border-left-style");
+	style->color = color_from_property(match, PRO_COLOR, black);
+	style->background_color = color_from_property(match, PRO_BACKGROUND_COLOR, transparent);
 
-	style->border_color[0] = color_from_property(match, "border-top-color", style->color);
-	style->border_color[1] = color_from_property(match, "border-right-color", style->color);
-	style->border_color[2] = color_from_property(match, "border-bottom-color", style->color);
-	style->border_color[3] = color_from_property(match, "border-left-color", style->color);
+	style->border_spacing = number_from_property(match, PRO_BORDER_SPACING, 0, N_LENGTH);
 
-	style->border_width[0] = border_width_from_property(match, "border-top-width");
-	style->border_width[1] = border_width_from_property(match, "border-right-width");
-	style->border_width[2] = border_width_from_property(match, "border-bottom-width");
-	style->border_width[3] = border_width_from_property(match, "border-left-width");
+	style->border_style_0 = border_style_from_property(match, PRO_BORDER_TOP_STYLE);
+	style->border_style_1 = border_style_from_property(match, PRO_BORDER_RIGHT_STYLE);
+	style->border_style_2 = border_style_from_property(match, PRO_BORDER_BOTTOM_STYLE);
+	style->border_style_3 = border_style_from_property(match, PRO_BORDER_LEFT_STYLE);
+
+	style->border_color[0] = color_from_property(match, PRO_BORDER_TOP_COLOR, style->color);
+	style->border_color[1] = color_from_property(match, PRO_BORDER_RIGHT_COLOR, style->color);
+	style->border_color[2] = color_from_property(match, PRO_BORDER_BOTTOM_COLOR, style->color);
+	style->border_color[3] = color_from_property(match, PRO_BORDER_LEFT_COLOR, style->color);
+
+	style->border_width[0] = border_width_from_property(match, PRO_BORDER_TOP_WIDTH);
+	style->border_width[1] = border_width_from_property(match, PRO_BORDER_RIGHT_WIDTH);
+	style->border_width[2] = border_width_from_property(match, PRO_BORDER_BOTTOM_WIDTH);
+	style->border_width[3] = border_width_from_property(match, PRO_BORDER_LEFT_WIDTH);
 
 	{
-		const char *font_weight = string_from_property(match, "font-weight", "normal");
-		const char *font_style = string_from_property(match, "font-style", "normal");
-		const char *font_variant = string_from_property(match, "font-variant", "normal");
+		const char *font_weight = string_from_property(match, PRO_FONT_WEIGHT, "normal");
+		const char *font_style = string_from_property(match, PRO_FONT_STYLE, "normal");
+		const char *font_variant = string_from_property(match, PRO_FONT_VARIANT, "normal");
 		int is_bold = is_bold_from_font_weight(font_weight);
 		int is_italic = is_italic_from_font_style(font_style);
 		style->small_caps = !strcmp(font_variant, "small-caps");
-		value = value_from_property(match, "font-family");
+		value = value_from_property(match, PRO_FONT_FAMILY);
 		while (value)
 		{
 			if (strcmp(value->data, ",") != 0)
@@ -1541,7 +1607,7 @@ static void print_value(fz_css_value *val)
 
 static void print_property(fz_css_property *prop)
 {
-	printf("\t%s: ", prop->name);
+	printf("\t%s: ", fz_css_property_name(prop->name));
 	print_value(prop->value);
 	if (prop->important)
 		printf(" !important");
