@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2022 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "html-imp.h"
 
@@ -14,30 +36,6 @@ static int is_internal_uri(const char *uri)
 	return 1;
 }
 
-static const char *box_href(fz_html_box *box)
-{
-	while (box)
-	{
-		const char *href = box->href;
-		if (href)
-			return href;
-		box = box->up;
-	}
-	return NULL;
-}
-
-static int has_same_href(fz_html_box *box, const char *old_href)
-{
-	while (box)
-	{
-		const char *href = box->href;
-		if (href)
-			return !strcmp(old_href, href);
-		box = box->up;
-	}
-	return 0;
-}
-
 static fz_link *load_link_flow(fz_context *ctx, fz_html_flow *flow, fz_link *head, int page, float page_h, const char *dir, const char *file)
 {
 	fz_link *link;
@@ -48,60 +46,66 @@ static fz_link *load_link_flow(fz_context *ctx, fz_html_flow *flow, fz_link *hea
 	const char *href;
 	float end;
 
+	float page_y0 = page * page_h;
+	float page_y1 = (page + 1) * page_h;
+
 	while (flow)
 	{
-		href = box_href(flow->box);
 		next = flow->next;
-		if (href && (int)(flow->y / page_h) == page)
+		if (flow->y >= page_y0 && flow->y <= page_y1)
 		{
-			/* Coalesce contiguous flow boxes into one link node */
-			end = flow->x + flow->w;
-			while (next &&
+			href = flow->box->href;
+			if (href)
+			{
+				/* Coalesce contiguous flow boxes into one link node */
+				end = flow->x + flow->w;
+				while (next &&
 					next->y == flow->y &&
 					next->h == flow->h &&
-					has_same_href(next->box, href))
-			{
-				end = next->x + next->w;
-				next = next->next;
-			}
-
-			bbox.x0 = flow->x;
-			bbox.y0 = flow->y - page * page_h;
-			bbox.x1 = end;
-			bbox.y1 = bbox.y0 + flow->h;
-			if (flow->type != FLOW_IMAGE)
-			{
-				/* flow->y is the baseline, adjust bbox appropriately */
-				bbox.y0 -= 0.8f * flow->h;
-				bbox.y1 -= 0.8f * flow->h;
-			}
-
-			if (is_internal_uri(href))
-			{
-				if (href[0] == '#')
+					next->box->href == href)
 				{
-					fz_strlcpy(path, file, sizeof path);
-					fz_strlcat(path, href, sizeof path);
+					end = next->x + next->w;
+					next = next->next;
+				}
+
+				bbox.x0 = flow->x;
+				bbox.y0 = flow->y - page * page_h;
+				bbox.x1 = end;
+				bbox.y1 = bbox.y0 + flow->h;
+				if (flow->type != FLOW_IMAGE)
+				{
+					/* flow->y is the baseline, adjust bbox appropriately */
+					bbox.y0 -= 0.8f * flow->h;
+					bbox.y1 -= 0.8f * flow->h;
+				}
+
+				if (is_internal_uri(href))
+				{
+					if (href[0] == '#')
+					{
+						fz_strlcpy(path, file, sizeof path);
+						fz_strlcat(path, href, sizeof path);
+					}
+					else
+					{
+						fz_strlcpy(path, dir, sizeof path);
+						fz_strlcat(path, "/", sizeof path);
+						fz_strlcat(path, href, sizeof path);
+					}
+					fz_urldecode(path);
+					fz_cleanname(path);
+
+					dest = path;
 				}
 				else
 				{
-					fz_strlcpy(path, dir, sizeof path);
-					fz_strlcat(path, "/", sizeof path);
-					fz_strlcat(path, href, sizeof path);
+					dest = href;
 				}
-				fz_urldecode(path);
-				fz_cleanname(path);
 
-				dest = path;
+				link = fz_new_derived_link(ctx, fz_link, bbox, dest);
+				link->next = head;
+				head = link;
 			}
-			else
-			{
-				dest = href;
-			}
-
-			link = fz_new_link(ctx, bbox, NULL, dest);
-			link->next = head;
-			head = link;
 		}
 		flow = next;
 	}
@@ -112,8 +116,8 @@ static fz_link *load_link_box(fz_context *ctx, fz_html_box *box, fz_link *head,
 {
 	while (box)
 	{
-		if (box->flow_head)
-			head = load_link_flow(ctx, box->flow_head, head, page, page_h, dir, file);
+		if (box->type == BOX_FLOW)
+			head = load_link_flow(ctx, box->u.flow.head, head, page, page_h, dir, file);
 		if (box->down)
 			head = load_link_box(ctx, box->down, head, page, page_h, dir, file);
 		box = box->next;
@@ -122,13 +126,13 @@ static fz_link *load_link_box(fz_context *ctx, fz_html_box *box, fz_link *head,
 }
 
 fz_link *
-fz_load_html_links(fz_context *ctx, fz_html *html, int page, const char *file, void *doc)
+fz_load_html_links(fz_context *ctx, fz_html *html, int page, const char *file)
 {
 	fz_link *link, *head;
 	char dir[2048];
 	fz_dirname(dir, file, sizeof dir);
 
-	head = load_link_box(ctx, html->root, NULL, page, html->page_h, dir, file);
+	head = load_link_box(ctx, html->tree.root, NULL, page, html->page_h, dir, file);
 
 	for (link = head; link; link = link->next)
 	{
@@ -137,9 +141,6 @@ fz_load_html_links(fz_context *ctx, fz_html *html, int page, const char *file, v
 		link->rect.x1 += html->page_margin[L];
 		link->rect.y0 += html->page_margin[T];
 		link->rect.y1 += html->page_margin[T];
-
-		/* Set document pointer */
-		link->doc = doc;
 	}
 
 	return head;
@@ -151,7 +152,7 @@ find_first_content(fz_html_box *box)
 	while (box)
 	{
 		if (box->type == BOX_FLOW)
-			return box->flow_head;
+			return box->u.flow.head;
 		box = box->down;
 	}
 	return NULL;
@@ -180,11 +181,11 @@ find_box_target(fz_html_box *box, const char *id)
 			fz_html_flow *flow = find_first_content(box);
 			if (flow)
 				return flow->y;
-			return box->y;
+			return box->s.layout.y;
 		}
 		if (box->type == BOX_FLOW)
 		{
-			y = find_flow_target(box->flow_head, id);
+			y = find_flow_target(box->u.flow.head, id);
 			if (y >= 0)
 				return y;
 		}
@@ -202,7 +203,7 @@ find_box_target(fz_html_box *box, const char *id)
 float
 fz_find_html_target(fz_context *ctx, fz_html *html, const char *id)
 {
-	return find_box_target(html->root, id);
+	return find_box_target(html->tree.root, id);
 }
 
 static fz_html_flow *
@@ -225,9 +226,9 @@ make_box_bookmark(fz_context *ctx, fz_html_box *box, float y)
 	{
 		if (box->type == BOX_FLOW)
 		{
-			if (box->y >= y)
+			if (box->s.layout.y >= y)
 			{
-				mark = make_flow_bookmark(ctx, box->flow_head, y);
+				mark = make_flow_bookmark(ctx, box->u.flow.head, y);
 				if (mark)
 					return mark;
 			}
@@ -246,7 +247,7 @@ make_box_bookmark(fz_context *ctx, fz_html_box *box, float y)
 fz_bookmark
 fz_make_html_bookmark(fz_context *ctx, fz_html *html, int page)
 {
-	return (fz_bookmark)make_box_bookmark(ctx, html->root, page * html->page_h);
+	return (fz_bookmark)make_box_bookmark(ctx, html->tree.root, page * html->page_h);
 }
 
 static int
@@ -268,7 +269,7 @@ lookup_box_bookmark(fz_context *ctx, fz_html_box *box, fz_html_flow *mark)
 	{
 		if (box->type == BOX_FLOW)
 		{
-			if (lookup_flow_bookmark(ctx, box->flow_head, mark))
+			if (lookup_flow_bookmark(ctx, box->u.flow.head, mark))
 				return 1;
 		}
 		else
@@ -285,7 +286,7 @@ int
 fz_lookup_html_bookmark(fz_context *ctx, fz_html *html, fz_bookmark mark)
 {
 	fz_html_flow *flow = (fz_html_flow*)mark;
-	if (flow && lookup_box_bookmark(ctx, html->root, flow))
+	if (flow && lookup_box_bookmark(ctx, html->tree.root, flow))
 		return (int)(flow->y / html->page_h);
 	return -1;
 }
@@ -328,7 +329,8 @@ cat_html_box(fz_context *ctx, fz_buffer *cat, fz_html_box *box)
 {
 	while (box)
 	{
-		cat_html_flow(ctx, cat, box->flow_head);
+		if (box->type == BOX_FLOW)
+			cat_html_flow(ctx, cat, box->u.flow.head);
 		cat_html_box(ctx, cat, box->down);
 		box = box->next;
 	}
@@ -342,7 +344,7 @@ cat_html_text(fz_context *ctx, struct outline_parser *x, fz_html_box *box)
 	else
 		fz_clear_buffer(ctx, x->cat);
 
-	cat_html_flow(ctx, x->cat, box->flow_head);
+	cat_html_flow(ctx, x->cat, box->u.flow.head);
 	cat_html_box(ctx, x->cat, box->down);
 
 	return fz_string_from_buffer(ctx, x->cat);
@@ -353,6 +355,7 @@ add_html_outline(fz_context *ctx, struct outline_parser *x, fz_html_box *box)
 {
 	fz_outline *node;
 	char buf[100];
+	int heading;
 
 	node = fz_new_outline(ctx);
 	fz_try(ctx)
@@ -361,7 +364,7 @@ add_html_outline(fz_context *ctx, struct outline_parser *x, fz_html_box *box)
 		if (!box->id)
 		{
 			fz_snprintf(buf, sizeof buf, "'%d", x->id++);
-			box->id = Memento_label(fz_pool_strdup(ctx, x->html->pool, buf), "box_id");
+			box->id = Memento_label(fz_pool_strdup(ctx, x->html->tree.pool, buf), "box_id");
 		}
 		node->uri = Memento_label(fz_asprintf(ctx, "#%s", box->id), "outline_uri");
 		node->is_open = 1;
@@ -372,19 +375,20 @@ add_html_outline(fz_context *ctx, struct outline_parser *x, fz_html_box *box)
 		fz_rethrow(ctx);
 	}
 
-	if (x->level[x->current] < (int)box->heading && x->current < 5)
+	heading = fz_html_heading_from_struct(box->structure);
+	if (x->level[x->current] < heading && x->current < 5)
 	{
 		x->tail[x->current+1] = x->down[x->current];
 		x->current += 1;
 	}
 	else
 	{
-		while (x->current > 0 && x->level[x->current] > (int)box->heading)
+		while (x->current > 0 && x->level[x->current] > heading)
 		{
 			x->current -= 1;
 		}
 	}
-	x->level[x->current] = box->heading;
+	x->level[x->current] = heading;
 
 	*(x->tail[x->current]) = node;
 	x->tail[x->current] = &node->next;
@@ -396,7 +400,8 @@ load_html_outline(fz_context *ctx, struct outline_parser *x, fz_html_box *box)
 {
 	while (box)
 	{
-		if (box->heading)
+		int heading = fz_html_heading_from_struct(box->structure);
+		if (heading)
 			add_html_outline(ctx, x, box);
 		if (box->down)
 			load_html_outline(ctx, x, box->down);
@@ -417,7 +422,7 @@ fz_load_html_outline(fz_context *ctx, fz_html *html)
 	state.current = 0;
 	state.id = 1;
 	fz_try(ctx)
-		load_html_outline(ctx, &state, html->root);
+		load_html_outline(ctx, &state, html->tree.root);
 	fz_always(ctx)
 		fz_drop_buffer(ctx, state.cat);
 	fz_catch(ctx)
