@@ -17,17 +17,20 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
+#include "cmaps/TrueType-UCS2.h"
+
 #include <string.h>
 
-pdf_cmap *
-pdf_load_embedded_cmap(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj)
+static pdf_cmap *
+pdf_load_embedded_cmap_imp(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	fz_stream *file = NULL;
 	pdf_cmap *cmap = NULL;
 	pdf_cmap *usecmap = NULL;
@@ -57,14 +60,9 @@ pdf_load_embedded_cmap(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj)
 		}
 		else if (pdf_is_indirect(ctx, obj))
 		{
-			if (pdf_mark_obj(ctx, obj))
+			if (pdf_cycle(ctx, &cycle, cycle_up, obj))
 				fz_throw(ctx, FZ_ERROR_GENERIC, "recursive CMap");
-			fz_try(ctx)
-				usecmap = pdf_load_embedded_cmap(ctx, doc, obj);
-			fz_always(ctx)
-				pdf_unmark_obj(ctx, obj);
-			fz_catch(ctx)
-				fz_rethrow(ctx);
+			usecmap = pdf_load_embedded_cmap_imp(ctx, doc, obj, &cycle);
 			pdf_set_usecmap(ctx, cmap, usecmap);
 		}
 
@@ -84,6 +82,12 @@ pdf_load_embedded_cmap(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj)
 	return cmap;
 }
 
+pdf_cmap *
+pdf_load_embedded_cmap(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj)
+{
+	return pdf_load_embedded_cmap_imp(ctx, doc, stmobj, NULL);
+}
+
 pdf_cmap *
 pdf_new_identity_cmap(fz_context *ctx, int wmode, int bytes)
 {
@@ -115,6 +119,7 @@ pdf_load_builtin_cmap(fz_context *ctx, const char *name)
 {
 	if (!strcmp(name, "Identity-H")) return pdf_new_identity_cmap(ctx, 0, 2);
 	if (!strcmp(name, "Identity-V")) return pdf_new_identity_cmap(ctx, 1, 2);
+	if (!strcmp(name, "TrueType-UCS2")) return &cmap_TrueType_UCS2;
 	return NULL;
 }
 
@@ -241,6 +246,7 @@ static pdf_cmap *table[] = {
 	&cmap_KSCms_UHC_HW_V,
 	&cmap_KSCms_UHC_V,
 	&cmap_KSCpc_EUC_H,
+	&cmap_TrueType_UCS2,
 	&cmap_UniCNS_UCS2_H,
 	&cmap_UniCNS_UCS2_V,
 	&cmap_UniCNS_UTF16_H,
