@@ -1,12 +1,43 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
+typedef struct resources_stack
+{
+	struct resources_stack *next;
+	pdf_obj *res;
+} resources_stack;
+
 typedef struct
 {
 	pdf_processor super;
 	fz_output *out;
 	int ahxencode;
 	int extgstate;
+	pdf_obj *res;
+	pdf_obj *last_res;
+	resources_stack *rstack;
 } pdf_output_processor;
 
 /* general graphics state */
@@ -63,26 +94,6 @@ pdf_out_ri(fz_context *ctx, pdf_processor *proc, const char *intent)
 		fz_write_printf(ctx, out, "%n ri\n", intent);
 }
 
-static void
-pdf_out_gs_OP(fz_context *ctx, pdf_processor *proc, int b)
-{
-}
-
-static void
-pdf_out_gs_op(fz_context *ctx, pdf_processor *proc, int b)
-{
-}
-
-static void
-pdf_out_gs_OPM(fz_context *ctx, pdf_processor *proc, int i)
-{
-}
-
-static void
-pdf_out_gs_UseBlackPtComp(fz_context *ctx, pdf_processor *proc, pdf_obj *name)
-{
-}
-
 static void
 pdf_out_i(fz_context *ctx, pdf_processor *proc, float flatness)
 {
@@ -601,7 +612,7 @@ pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img, const char *colo
 	else if (img->colorspace == fz_device_cmyk(ctx))
 		fz_write_string(ctx, out, "/CS/CMYK\n");
 	else if (colorspace)
-		fz_write_printf(ctx, out, "/CS%n/n", colorspace);
+		fz_write_printf(ctx, out, "/CS%n\n", colorspace);
 	else
 		fz_throw(ctx, FZ_ERROR_GENERIC, "BI operator can only show ImageMask, Gray, RGB, or CMYK images");
 	if (img->interpolate)
@@ -623,7 +634,7 @@ pdf_out_BI(fz_context *ctx, pdf_processor *proc, fz_image *img, const char *colo
 
 	case FZ_IMAGE_JPEG:
 		fz_write_string(ctx, out, ahx ? "/F[/AHx/DCT]\n" : "/F/DCT\n");
-		if (cbuf->params.u.jpeg.color_transform != -1)
+		if (cbuf->params.u.jpeg.color_transform >= 0)
 			fz_write_printf(ctx, out, "/DP<</ColorTransform %d>>\n",
 				cbuf->params.u.jpeg.color_transform);
 		break;
@@ -730,7 +741,7 @@ pdf_out_Do_image(fz_context *ctx, pdf_processor *proc, const char *name, fz_imag
 }
 
 static void
-pdf_out_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *xobj, pdf_obj *page_resources)
+pdf_out_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *xobj)
 {
 	fz_output *out = ((pdf_output_processor*)proc)->out;
 	fz_write_printf(ctx, out, "%n Do\n", name);
@@ -805,134 +816,161 @@ pdf_close_output_processor(fz_context *ctx, pdf_processor *proc)
 static void
 pdf_drop_output_processor(fz_context *ctx, pdf_processor *proc)
 {
-	fz_output *out = ((pdf_output_processor*)proc)->out;
-	fz_drop_output(ctx, out);
+	pdf_output_processor *p = (pdf_output_processor *)proc;
+
+	fz_drop_output(ctx, p->out);
+}
+
+static void
+pdf_out_push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *res)
+{
+	pdf_output_processor *p = (pdf_output_processor *)proc;
+	resources_stack *stk = fz_malloc_struct(ctx, resources_stack);
+
+	stk->next = p->rstack;
+	p->rstack = stk;
+	stk->res = pdf_keep_obj(ctx, res);
+}
+
+static pdf_obj *
+pdf_out_pop_resources(fz_context *ctx, pdf_processor *proc)
+{
+	pdf_output_processor *p = (pdf_output_processor *)proc;
+	resources_stack *stk = p->rstack;
+	pdf_obj *res = stk->res;
+
+	p->rstack = stk->next;
+	fz_free(ctx, stk);
+
+	return res;
 }
 
 pdf_processor *
 pdf_new_output_processor(fz_context *ctx, fz_output *out, int ahxencode)
 {
 	pdf_output_processor *proc = pdf_new_processor(ctx, sizeof *proc);
-	{
-		proc->super.close_processor = pdf_close_output_processor;
-		proc->super.drop_processor = pdf_drop_output_processor;
 
-		/* general graphics state */
-		proc->super.op_w = pdf_out_w;
-		proc->super.op_j = pdf_out_j;
-		proc->super.op_J = pdf_out_J;
-		proc->super.op_M = pdf_out_M;
-		proc->super.op_d = pdf_out_d;
-		proc->super.op_ri = pdf_out_ri;
-		proc->super.op_i = pdf_out_i;
-		proc->super.op_gs_begin = pdf_out_gs_begin;
-		proc->super.op_gs_end = pdf_out_gs_end;
+	proc->super.close_processor = pdf_close_output_processor;
+	proc->super.drop_processor = pdf_drop_output_processor;
 
-		/* transparency graphics state */
-		proc->super.op_gs_BM = NULL;
-		proc->super.op_gs_CA = NULL;
-		proc->super.op_gs_ca = NULL;
-		proc->super.op_gs_SMask = NULL;
+	proc->super.push_resources = pdf_out_push_resources;
+	proc->super.pop_resources = pdf_out_pop_resources;
 
-		/* special graphics state */
-		proc->super.op_q = pdf_out_q;
-		proc->super.op_Q = pdf_out_Q;
-		proc->super.op_cm = pdf_out_cm;
+	/* general graphics state */
+	proc->super.op_w = pdf_out_w;
+	proc->super.op_j = pdf_out_j;
+	proc->super.op_J = pdf_out_J;
+	proc->super.op_M = pdf_out_M;
+	proc->super.op_d = pdf_out_d;
+	proc->super.op_ri = pdf_out_ri;
+	proc->super.op_i = pdf_out_i;
+	proc->super.op_gs_begin = pdf_out_gs_begin;
+	proc->super.op_gs_end = pdf_out_gs_end;
 
-		/* path construction */
-		proc->super.op_m = pdf_out_m;
-		proc->super.op_l = pdf_out_l;
-		proc->super.op_c = pdf_out_c;
-		proc->super.op_v = pdf_out_v;
-		proc->super.op_y = pdf_out_y;
-		proc->super.op_h = pdf_out_h;
-		proc->super.op_re = pdf_out_re;
+	/* transparency graphics state */
+	proc->super.op_gs_BM = NULL;
+	proc->super.op_gs_CA = NULL;
+	proc->super.op_gs_ca = NULL;
+	proc->super.op_gs_SMask = NULL;
 
-		/* path painting */
-		proc->super.op_S = pdf_out_S;
-		proc->super.op_s = pdf_out_s;
-		proc->super.op_F = pdf_out_F;
-		proc->super.op_f = pdf_out_f;
-		proc->super.op_fstar = pdf_out_fstar;
-		proc->super.op_B = pdf_out_B;
-		proc->super.op_Bstar = pdf_out_Bstar;
-		proc->super.op_b = pdf_out_b;
-		proc->super.op_bstar = pdf_out_bstar;
-		proc->super.op_n = pdf_out_n;
+	/* special graphics state */
+	proc->super.op_q = pdf_out_q;
+	proc->super.op_Q = pdf_out_Q;
+	proc->super.op_cm = pdf_out_cm;
 
-		/* clipping paths */
-		proc->super.op_W = pdf_out_W;
-		proc->super.op_Wstar = pdf_out_Wstar;
+	/* path construction */
+	proc->super.op_m = pdf_out_m;
+	proc->super.op_l = pdf_out_l;
+	proc->super.op_c = pdf_out_c;
+	proc->super.op_v = pdf_out_v;
+	proc->super.op_y = pdf_out_y;
+	proc->super.op_h = pdf_out_h;
+	proc->super.op_re = pdf_out_re;
 
-		/* text objects */
-		proc->super.op_BT = pdf_out_BT;
-		proc->super.op_ET = pdf_out_ET;
+	/* path painting */
+	proc->super.op_S = pdf_out_S;
+	proc->super.op_s = pdf_out_s;
+	proc->super.op_F = pdf_out_F;
+	proc->super.op_f = pdf_out_f;
+	proc->super.op_fstar = pdf_out_fstar;
+	proc->super.op_B = pdf_out_B;
+	proc->super.op_Bstar = pdf_out_Bstar;
+	proc->super.op_b = pdf_out_b;
+	proc->super.op_bstar = pdf_out_bstar;
+	proc->super.op_n = pdf_out_n;
 
-		/* text state */
-		proc->super.op_Tc = pdf_out_Tc;
-		proc->super.op_Tw = pdf_out_Tw;
-		proc->super.op_Tz = pdf_out_Tz;
-		proc->super.op_TL = pdf_out_TL;
-		proc->super.op_Tf = pdf_out_Tf;
-		proc->super.op_Tr = pdf_out_Tr;
-		proc->super.op_Ts = pdf_out_Ts;
+	/* clipping paths */
+	proc->super.op_W = pdf_out_W;
+	proc->super.op_Wstar = pdf_out_Wstar;
 
-		/* text positioning */
-		proc->super.op_Td = pdf_out_Td;
-		proc->super.op_TD = pdf_out_TD;
-		proc->super.op_Tm = pdf_out_Tm;
-		proc->super.op_Tstar = pdf_out_Tstar;
+	/* text objects */
+	proc->super.op_BT = pdf_out_BT;
+	proc->super.op_ET = pdf_out_ET;
 
-		/* text showing */
-		proc->super.op_TJ = pdf_out_TJ;
-		proc->super.op_Tj = pdf_out_Tj;
-		proc->super.op_squote = pdf_out_squote;
-		proc->super.op_dquote = pdf_out_dquote;
+	/* text state */
+	proc->super.op_Tc = pdf_out_Tc;
+	proc->super.op_Tw = pdf_out_Tw;
+	proc->super.op_Tz = pdf_out_Tz;
+	proc->super.op_TL = pdf_out_TL;
+	proc->super.op_Tf = pdf_out_Tf;
+	proc->super.op_Tr = pdf_out_Tr;
+	proc->super.op_Ts = pdf_out_Ts;
 
-		/* type 3 fonts */
-		proc->super.op_d0 = pdf_out_d0;
-		proc->super.op_d1 = pdf_out_d1;
+	/* text positioning */
+	proc->super.op_Td = pdf_out_Td;
+	proc->super.op_TD = pdf_out_TD;
+	proc->super.op_Tm = pdf_out_Tm;
+	proc->super.op_Tstar = pdf_out_Tstar;
 
-		/* color */
-		proc->super.op_CS = pdf_out_CS;
-		proc->super.op_cs = pdf_out_cs;
-		proc->super.op_SC_color = pdf_out_SC_color;
-		proc->super.op_sc_color = pdf_out_sc_color;
-		proc->super.op_SC_pattern = pdf_out_SC_pattern;
-		proc->super.op_sc_pattern = pdf_out_sc_pattern;
-		proc->super.op_SC_shade = pdf_out_SC_shade;
-		proc->super.op_sc_shade = pdf_out_sc_shade;
+	/* text showing */
+	proc->super.op_TJ = pdf_out_TJ;
+	proc->super.op_Tj = pdf_out_Tj;
+	proc->super.op_squote = pdf_out_squote;
+	proc->super.op_dquote = pdf_out_dquote;
 
-		proc->super.op_G = pdf_out_G;
-		proc->super.op_g = pdf_out_g;
-		proc->super.op_RG = pdf_out_RG;
-		proc->super.op_rg = pdf_out_rg;
-		proc->super.op_K = pdf_out_K;
-		proc->super.op_k = pdf_out_k;
+	/* type 3 fonts */
+	proc->super.op_d0 = pdf_out_d0;
+	proc->super.op_d1 = pdf_out_d1;
 
-		/* shadings, images, xobjects */
-		proc->super.op_BI = pdf_out_BI;
-		proc->super.op_sh = pdf_out_sh;
-		proc->super.op_Do_image = pdf_out_Do_image;
-		proc->super.op_Do_form = pdf_out_Do_form;
+	/* color */
+	proc->super.op_CS = pdf_out_CS;
+	proc->super.op_cs = pdf_out_cs;
+	proc->super.op_SC_color = pdf_out_SC_color;
+	proc->super.op_sc_color = pdf_out_sc_color;
+	proc->super.op_SC_pattern = pdf_out_SC_pattern;
+	proc->super.op_sc_pattern = pdf_out_sc_pattern;
+	proc->super.op_SC_shade = pdf_out_SC_shade;
+	proc->super.op_sc_shade = pdf_out_sc_shade;
 
-		/* marked content */
-		proc->super.op_MP = pdf_out_MP;
-		proc->super.op_DP = pdf_out_DP;
-		proc->super.op_BMC = pdf_out_BMC;
-		proc->super.op_BDC = pdf_out_BDC;
-		proc->super.op_EMC = pdf_out_EMC;
+	proc->super.op_G = pdf_out_G;
+	proc->super.op_g = pdf_out_g;
+	proc->super.op_RG = pdf_out_RG;
+	proc->super.op_rg = pdf_out_rg;
+	proc->super.op_K = pdf_out_K;
+	proc->super.op_k = pdf_out_k;
 
-		/* compatibility */
-		proc->super.op_BX = pdf_out_BX;
-		proc->super.op_EX = pdf_out_EX;
+	/* shadings, images, xobjects */
+	proc->super.op_BI = pdf_out_BI;
+	proc->super.op_sh = pdf_out_sh;
+	proc->super.op_Do_image = pdf_out_Do_image;
+	proc->super.op_Do_form = pdf_out_Do_form;
 
-		/* extgstate */
-		proc->super.op_gs_OP = pdf_out_gs_OP;
-		proc->super.op_gs_op = pdf_out_gs_op;
-		proc->super.op_gs_OPM = pdf_out_gs_OPM;
-		proc->super.op_gs_UseBlackPtComp = pdf_out_gs_UseBlackPtComp;
-	}
+	/* marked content */
+	proc->super.op_MP = pdf_out_MP;
+	proc->super.op_DP = pdf_out_DP;
+	proc->super.op_BMC = pdf_out_BMC;
+	proc->super.op_BDC = pdf_out_BDC;
+	proc->super.op_EMC = pdf_out_EMC;
+
+	/* compatibility */
+	proc->super.op_BX = pdf_out_BX;
+	proc->super.op_EX = pdf_out_EX;
+
+	/* extgstate */
+	proc->super.op_gs_OP = NULL;
+	proc->super.op_gs_op = NULL;
+	proc->super.op_gs_OPM = NULL;
+	proc->super.op_gs_UseBlackPtComp = NULL;
 
 	proc->out = out;
 	proc->ahxencode = ahxencode;
