@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -74,6 +74,9 @@ struct pdf_gstate
 	pdf_material stroke;
 	pdf_material fill;
 
+	/* pattern paint type 2 */
+	int ismask;
+
 	/* text state */
 	pdf_text_state text;
 
@@ -86,14 +89,36 @@ struct pdf_gstate
 	int luminosity;
 };
 
+typedef struct resources_stack
+{
+	struct resources_stack *next;
+	pdf_obj *resources;
+} resources_stack;
+
+typedef struct marked_content_stack
+{
+	struct marked_content_stack *next;
+	pdf_obj *tag;
+	pdf_obj *val;
+} marked_content_stack;
+
+typedef struct begin_layer_stack
+{
+	struct begin_layer_stack *next;
+	char *layer;
+} begin_layer_stack;
+
 struct pdf_run_processor
 {
 	pdf_processor super;
+	pdf_document *doc;
 	fz_device *dev;
 	fz_cookie *cookie;
 
 	fz_default_colorspaces *default_cs;
 
+	resources_stack *rstack;
+
 	/* path object state */
 	fz_path *path;
 	int clip;
@@ -108,8 +133,57 @@ struct pdf_run_processor
 	int gtop;
 	int gbot;
 	int gparent;
+
+	/* xobject cycle detector */
+	pdf_cycle_list *cycle;
+
+	pdf_obj *role_map;
+
+	marked_content_stack *marked_content;
+	pdf_obj *mcid_sent;
+
+	int struct_parent;
+	int broken_struct_tree;
+
+	/* Pending begin layers */
+	begin_layer_stack *begin_layer;
+	begin_layer_stack **next_begin_layer;
 };
 
+static void
+push_begin_layer(fz_context *ctx, pdf_run_processor *proc, const char *str)
+{
+	begin_layer_stack *s = fz_malloc_struct(ctx, begin_layer_stack);
+
+	fz_try(ctx)
+		s->layer = fz_strdup(ctx, str);
+	fz_catch(ctx)
+	{
+		fz_free(ctx, s);
+		fz_rethrow(ctx);
+	}
+
+	s->next = NULL;
+	*proc->next_begin_layer = s;
+	proc->next_begin_layer = &s->next;
+}
+
+static void
+flush_begin_layer(fz_context *ctx, pdf_run_processor *proc)
+{
+	begin_layer_stack *s;
+
+	while (proc->begin_layer)
+	{
+		s = proc->begin_layer;
+		fz_begin_layer(ctx, proc->dev, s->layer);
+		proc->begin_layer = s->next;
+		fz_free(ctx, s->layer);
+		fz_free(ctx, s);
+	}
+	proc->next_begin_layer = &proc->begin_layer;
+}
+
 typedef struct
 {
 	pdf_obj *softmask;
@@ -272,10 +346,13 @@ pdf_drop_material(fz_context *ctx, pdf_material *mat)
 static void
 pdf_copy_pattern_gstate(fz_context *ctx, pdf_gstate *dst, const pdf_gstate *src)
 {
+	pdf_font_desc *old_font = dst->text.font;
+
 	dst->ctm = src->ctm;
 
-	pdf_drop_font(ctx, dst->text.font);
-	dst->text.font = pdf_keep_font(ctx, src->text.font);
+	dst->text = src->text;
+	pdf_keep_font(ctx, src->text.font);
+	pdf_drop_font(ctx, old_font);
 
 	pdf_drop_obj(ctx, dst->softmask);
 	dst->softmask = pdf_keep_obj(ctx, src->softmask);
@@ -390,6 +467,8 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 
 	if (pat->ismask)
 	{
+		/* Inhibit any changes to the color since we're drawing an uncolored pattern. */
+		gstate->ismask = 1;
 		pdf_unset_pattern(ctx, pr, PDF_FILL);
 		pdf_unset_pattern(ctx, pr, PDF_STROKE);
 		if (what == PDF_FILL)
@@ -462,7 +541,7 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 		{
 			gstate->ctm = ptm;
 			pdf_gsave(ctx, pr);
-			pdf_process_contents(ctx, (pdf_processor*)pr, pat->document, pat->resources, pat->contents, NULL);
+			pdf_process_contents(ctx, (pdf_processor*)pr, pat->document, pat->resources, pat->contents, NULL, NULL);
 			pdf_grestore(ctx, pr);
 		}
 		fz_end_tile(ctx, pr->dev);
@@ -499,7 +578,7 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 				gstate = pr->gstate + pr->gtop;
 				gstate->ctm = fz_pre_translate(ptm, x * pat->xstep, y * pat->ystep);
 				pdf_gsave(ctx, pr);
-				pdf_process_contents(ctx, (pdf_processor*)pr, pat->document, pat->resources, pat->contents, NULL);
+				pdf_process_contents(ctx, (pdf_processor*)pr, pat->document, pat->resources, pat->contents, NULL, NULL);
 				pdf_grestore(ctx, pr);
 			}
 		}
@@ -550,6 +629,8 @@ pdf_show_image(fz_context *ctx, pdf_run_processor *pr, fz_image *image)
 	if (pr->super.hidden)
 		return;
 
+	flush_begin_layer(ctx, pr);
+
 	/* PDF has images bottom-up, so flip them right side up here */
 	image_ctm = fz_pre_scale(fz_pre_translate(gstate->ctm, 0, 1), 1, -1);
 
@@ -597,6 +678,8 @@ pdf_show_path(fz_context *ctx, pdf_run_processor *pr, int doclose, int dofill, i
 	softmask_save softmask = { NULL };
 	int knockout_group = 0;
 
+	flush_begin_layer(ctx, pr);
+
 	if (dostroke) {
 		if (pr->dev->flags & (FZ_DEVFLAG_STROKECOLOR_UNDEFINED | FZ_DEVFLAG_LINEJOIN_UNDEFINED | FZ_DEVFLAG_LINEWIDTH_UNDEFINED))
 			pr->dev->flags |= FZ_DEVFLAG_UNCACHEABLE;
@@ -748,6 +831,9 @@ pdf_flush_text(fz_context *ctx, pdf_run_processor *pr)
 	if (!text)
 		return gstate;
 
+	/* If we are going to output text, we need to have flushed any begin layers first. */
+	flush_begin_layer(ctx, pr);
+
 	dofill = dostroke = doclip = doinvisible = 0;
 	switch (pr->tos.text_mode)
 	{
@@ -884,13 +970,13 @@ pdf_flush_text(fz_context *ctx, pdf_run_processor *pr)
 }
 
 static void
-pdf_show_char(fz_context *ctx, pdf_run_processor *pr, int cid)
+pdf_show_char(fz_context *ctx, pdf_run_processor *pr, int cid, fz_text_language lang)
 {
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
 	pdf_font_desc *fontdesc = gstate->text.font;
 	fz_matrix trm;
 	int gid;
-	int ucsbuf[8];
+	int ucsbuf[PDF_MRANGE_CAP];
 	int ucslen;
 	int i;
 	int render_direct;
@@ -913,7 +999,12 @@ pdf_show_char(fz_context *ctx, pdf_run_processor *pr, int cid)
 		 * type3 glyphs that seem to inherit current graphics
 		 * attributes, or type 3 glyphs within type3 glyphs). */
 		fz_matrix composed = fz_concat(trm, gstate->ctm);
+		pdf_gsave(ctx, pr);
+		gstate = pr->gstate + pr->gtop;
+		pdf_drop_font(ctx, gstate->text.font);
+		gstate->text.font = NULL; /* don't inherit the current font... */
 		fz_render_t3_glyph_direct(ctx, pr->dev, fontdesc->font, gid, composed, gstate, pr->default_cs);
+		pdf_grestore(ctx, pr);
 		/* Render text invisibly so that it can still be extracted. */
 		pr->tos.text_mode = 3;
 	}
@@ -921,6 +1012,15 @@ pdf_show_char(fz_context *ctx, pdf_run_processor *pr, int cid)
 	ucslen = 0;
 	if (fontdesc->to_unicode)
 		ucslen = pdf_lookup_cmap_full(fontdesc->to_unicode, cid, ucsbuf);
+
+	/* convert ascii whitespace control characters to spaces */
+	if (ucslen == 1 && (ucsbuf[0] >= 8 && ucsbuf[0] <= 13))
+		ucsbuf[0] = ' ';
+
+	/* ignore obviously bad values in ToUnicode, fall back to the cid_to_ucs table */
+	if (ucslen == 1 && (ucsbuf[0] < 32 || (ucsbuf[0] >= 127 && ucsbuf[0] < 160)))
+		ucslen = 0;
+
 	if (ucslen == 0 && (size_t)cid < fontdesc->cid_to_ucs_len)
 	{
 		ucsbuf[0] = fontdesc->cid_to_ucs[cid];
@@ -933,11 +1033,11 @@ pdf_show_char(fz_context *ctx, pdf_run_processor *pr, int cid)
 	}
 
 	/* add glyph to textobject */
-	fz_show_glyph(ctx, pr->tos.text, fontdesc->font, trm, gid, ucsbuf[0], fontdesc->wmode, 0, FZ_BIDI_NEUTRAL, FZ_LANG_UNSET);
+	fz_show_glyph_aux(ctx, pr->tos.text, fontdesc->font, trm, gid, ucsbuf[0], cid, fontdesc->wmode, 0, FZ_BIDI_NEUTRAL, lang);
 
 	/* add filler glyphs for one-to-many unicode mapping */
 	for (i = 1; i < ucslen; i++)
-		fz_show_glyph(ctx, pr->tos.text, fontdesc->font, trm, -1, ucsbuf[i], fontdesc->wmode, 0, FZ_BIDI_NEUTRAL, FZ_LANG_UNSET);
+		fz_show_glyph_aux(ctx, pr->tos.text, fontdesc->font, trm, -1, ucsbuf[i], -1, fontdesc->wmode, 0, FZ_BIDI_NEUTRAL, lang);
 
 	pdf_tos_move_after_char(ctx, &pr->tos);
 }
@@ -954,6 +1054,53 @@ pdf_show_space(fz_context *ctx, pdf_run_processor *pr, float tadj)
 		pr->tos.tm = fz_pre_translate(pr->tos.tm, 0, tadj);
 }
 
+static pdf_obj *
+lookup_mcid(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val)
+{
+	pdf_obj *mcid;
+	int id;
+	pdf_obj *mcids;
+
+	if (proc->struct_parent == -1)
+		return NULL;
+
+	mcid = pdf_dict_get(ctx, val, PDF_NAME(MCID));
+	if (!mcid)
+		return NULL;
+
+	if (!pdf_is_number(ctx, mcid))
+		return NULL;
+
+	id = pdf_to_int(ctx, mcid);
+	mcids = pdf_lookup_number(ctx, pdf_dict_getl(ctx, pdf_trailer(ctx, proc->doc), PDF_NAME(Root), PDF_NAME(StructTreeRoot), PDF_NAME(ParentTree), NULL), proc->struct_parent);
+	return pdf_array_get(ctx, mcids, id);
+}
+
+static fz_text_language
+find_lang_from_mc(fz_context *ctx, pdf_run_processor *pr)
+{
+	marked_content_stack *mc;
+
+	for (mc = pr->marked_content; mc != NULL; mc = mc->next)
+	{
+		size_t len;
+		const char *lang;
+
+		lang = pdf_dict_get_string(ctx, mc->val, PDF_NAME(Lang), &len);
+		if (!lang)
+			lang = pdf_dict_get_string(ctx, lookup_mcid(ctx, pr, mc->val), PDF_NAME(Lang), &len);
+		if (lang)
+		{
+			char text[8];
+			memcpy(text, lang, len < 8 ? len : 7);
+			text[len < 8 ? len : 7] = 0;
+			return fz_text_language_from_string(text);
+		}
+	}
+
+	return FZ_LANG_UNSET;
+}
+
 static void
 show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, size_t len)
 {
@@ -962,6 +1109,9 @@ show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, size_t l
 	unsigned char *end = buf + len;
 	unsigned int cpt;
 	int cid;
+	fz_text_language lang = find_lang_from_mc(ctx, pr);
+
+	flush_begin_layer(ctx, pr);
 
 	while (buf < end)
 	{
@@ -970,7 +1120,7 @@ show_string(fz_context *ctx, pdf_run_processor *pr, unsigned char *buf, size_t l
 
 		cid = pdf_lookup_cmap(fontdesc->encoding, cpt);
 		if (cid >= 0)
-			pdf_show_char(ctx, pr, cid);
+			pdf_show_char(ctx, pr, cid, lang);
 		else
 			fz_warn(ctx, "cannot encode character");
 		if (cpt == 32 && w == 1)
@@ -1073,6 +1223,8 @@ pdf_init_gstate(fz_context *ctx, pdf_gstate *gs, fz_matrix ctm)
 
 	gs->fill.color_params = fz_default_color_params;
 	gs->stroke.color_params = fz_default_color_params;
+
+	gs->ismask = 0;
 }
 
 static void
@@ -1096,6 +1248,10 @@ pdf_set_colorspace(fz_context *ctx, pdf_run_processor *pr, int what, fz_colorspa
 
 	gstate = pdf_flush_text(ctx, pr);
 
+	/* Don't change color if we're drawing an uncolored pattern tile! */
+	if (gstate->ismask)
+		return;
+
 	mat = what == PDF_FILL ? &gstate->fill : &gstate->stroke;
 
 	fz_drop_colorspace(ctx, mat->colorspace);
@@ -1124,6 +1280,10 @@ pdf_set_color(fz_context *ctx, pdf_run_processor *pr, int what, float *v)
 
 	gstate = pdf_flush_text(ctx, pr);
 
+	/* Don't change color if we're drawing an uncolored pattern tile! */
+	if (gstate->ismask)
+		return;
+
 	mat = what == PDF_FILL ? &gstate->fill : &gstate->stroke;
 
 	switch (mat->kind)
@@ -1180,10 +1340,571 @@ pdf_set_pattern(fz_context *ctx, pdf_run_processor *pr, int what, pdf_pattern *p
 	mat->gstate_num = pr->gparent;
 }
 
-static void
-pdf_run_xobject(fz_context *ctx, pdf_run_processor *proc, pdf_obj *xobj, pdf_obj *page_resources, fz_matrix transform, int is_smask)
+fz_structure
+structure_type(fz_context *ctx, pdf_run_processor *proc, pdf_obj *tag)
 {
-	pdf_run_processor *pr = (pdf_run_processor *)proc;
+	/* Perform Structure mapping to go from tag to standard. */
+	if (proc->role_map)
+	{
+		pdf_obj *o = pdf_dict_get(ctx, proc->role_map, tag);
+		if (o)
+			tag = o;
+	}
+
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Document)))
+		return FZ_STRUCTURE_DOCUMENT;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Part)))
+		return FZ_STRUCTURE_PART;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Art)))
+		return FZ_STRUCTURE_ART;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Sect)))
+		return FZ_STRUCTURE_SECT;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Div)))
+		return FZ_STRUCTURE_DIV;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(BlockQuote)))
+		return FZ_STRUCTURE_BLOCKQUOTE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Caption)))
+		return FZ_STRUCTURE_CAPTION;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TOC)))
+		return FZ_STRUCTURE_TOC;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TOCI)))
+		return FZ_STRUCTURE_TOCI;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Index)))
+		return FZ_STRUCTURE_INDEX;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(NonStruct)))
+		return FZ_STRUCTURE_NONSTRUCT;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Private)))
+		return FZ_STRUCTURE_PRIVATE;
+	/* Grouping elements (PDF 2.0 - Table 364) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(DocumentFragment)))
+		return FZ_STRUCTURE_DOCUMENTFRAGMENT;
+	/* Grouping elements (PDF 2.0 - Table 365) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Aside)))
+		return FZ_STRUCTURE_ASIDE;
+	/* Grouping elements (PDF 2.0 - Table 366) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Title)))
+		return FZ_STRUCTURE_TITLE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(FENote)))
+		return FZ_STRUCTURE_FENOTE;
+	/* Grouping elements (PDF 2.0 - Table 367) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Sub)))
+		return FZ_STRUCTURE_SUB;
+
+	/* Paragraphlike elements (PDF 1.7 - Table 10.21) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(P)))
+		return FZ_STRUCTURE_P;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H)))
+		return FZ_STRUCTURE_H;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H1)))
+		return FZ_STRUCTURE_H1;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H2)))
+		return FZ_STRUCTURE_H2;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H3)))
+		return FZ_STRUCTURE_H3;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H4)))
+		return FZ_STRUCTURE_H4;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H5)))
+		return FZ_STRUCTURE_H5;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(H6)))
+		return FZ_STRUCTURE_H6;
+
+	/* List elements (PDF 1.7 - Table 10.23) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(List)))
+		return FZ_STRUCTURE_LIST;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(LI)))
+		return FZ_STRUCTURE_LISTITEM;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Lbl)))
+		return FZ_STRUCTURE_LABEL;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(LBody)))
+		return FZ_STRUCTURE_LISTBODY;
+
+	/* Table elements (PDF 1.7 - Table 10.24) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Table)))
+		return FZ_STRUCTURE_TABLE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TR)))
+		return FZ_STRUCTURE_TR;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TH)))
+		return FZ_STRUCTURE_TH;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TD)))
+		return FZ_STRUCTURE_TD;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(THead)))
+		return FZ_STRUCTURE_THEAD;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TBody)))
+		return FZ_STRUCTURE_TBODY;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(TFoot)))
+		return FZ_STRUCTURE_TFOOT;
+
+	/* Inline elements (PDF 1.7 - Table 10.25) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Span)))
+		return FZ_STRUCTURE_SPAN;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Quote)))
+		return FZ_STRUCTURE_QUOTE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Note)))
+		return FZ_STRUCTURE_NOTE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Reference)))
+		return FZ_STRUCTURE_REFERENCE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(BibEntry)))
+		return FZ_STRUCTURE_BIBENTRY;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Code)))
+		return FZ_STRUCTURE_CODE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Link)))
+		return FZ_STRUCTURE_LINK;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Annot)))
+		return FZ_STRUCTURE_ANNOT;
+	/* Inline elements (PDF 2.0 - Table 368) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Em)))
+		return FZ_STRUCTURE_EM;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Strong)))
+		return FZ_STRUCTURE_STRONG;
+
+	/* Ruby inline element (PDF 1.7 - Table 10.26) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Ruby)))
+		return FZ_STRUCTURE_RUBY;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(RB)))
+		return FZ_STRUCTURE_RB;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(RT)))
+		return FZ_STRUCTURE_RT;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(RP)))
+		return FZ_STRUCTURE_RP;
+
+	/* Warichu inline element (PDF 1.7 - Table 10.26) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Warichu)))
+		return FZ_STRUCTURE_WARICHU;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(WT)))
+		return FZ_STRUCTURE_WT;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(WP)))
+		return FZ_STRUCTURE_WP;
+
+	/* Illustration elements (PDF 1.7 - Table 10.27) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Figure)))
+		return FZ_STRUCTURE_FIGURE;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Formula)))
+		return FZ_STRUCTURE_FORMULA;
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Form)))
+		return FZ_STRUCTURE_FORM;
+
+	/* Artifact structure type (PDF 2.0 - Table 375) */
+	if (pdf_name_eq(ctx, tag, PDF_NAME(Artifact)))
+		return FZ_STRUCTURE_ARTIFACT;
+
+	return FZ_STRUCTURE_INVALID;
+}
+
+static void
+begin_metatext(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val, pdf_obj *mcid, fz_metatext meta, pdf_obj *name)
+{
+	pdf_obj *text = pdf_dict_get(ctx, val, name);
+
+	if (!text)
+		text = pdf_dict_get(ctx, mcid, name);
+	if (!text)
+		return;
+
+	pdf_flush_text(ctx, proc);
+
+	fz_begin_metatext(ctx, proc->dev, meta, pdf_to_text_string(ctx, text));
+}
+
+static void
+end_metatext(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val, pdf_obj *mcid, pdf_obj *name)
+{
+	pdf_obj *text = pdf_dict_get(ctx, val, name);
+
+	if (!text)
+		text = pdf_dict_get(ctx, mcid, name);
+	if (!text)
+		return;
+
+	pdf_flush_text(ctx, proc);
+
+	fz_end_metatext(ctx, proc->dev);
+}
+
+static void
+begin_oc(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val, pdf_cycle_list *cycle_up)
+{
+	/* val has been resolved to a dict for us by the originally specified name
+	 * having been looked up in Properties already for us. Either there will
+	 * be a Name entry, or there will be an OCGs and it'll be a group one. */
+	pdf_cycle_list cycle;
+	pdf_obj *obj;
+	int i, n;
+
+	if (pdf_cycle(ctx, &cycle, cycle_up, val))
+		return;
+
+	obj = pdf_dict_get(ctx, val, PDF_NAME(Name));
+	if (obj)
+	{
+		const char *name = "";
+		pdf_flush_text(ctx, proc);
+		if (pdf_is_name(ctx, obj))
+			name = pdf_to_name(ctx, obj);
+		else if (pdf_is_string(ctx, obj))
+			name = pdf_to_text_string(ctx, obj);
+
+		push_begin_layer(ctx, proc, name);
+		return;
+	}
+
+	obj = pdf_dict_get(ctx, val, PDF_NAME(OCGs));
+	n = pdf_array_len(ctx, obj);
+	for (i = 0; i < n; i++)
+	{
+		begin_oc(ctx, proc, pdf_array_get(ctx, obj, i), &cycle);
+	}
+}
+
+static void
+end_oc(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val, pdf_cycle_list *cycle_up)
+{
+	/* val has been resolved to a dict for us by the originally specified name
+	 * having been looked up in Properties already for us. Either there will
+	 * be a Name entry, or there will be an OCGs and it'll be a group one. */
+	pdf_cycle_list cycle;
+	pdf_obj *obj;
+	int i, n;
+
+	if (pdf_cycle(ctx, &cycle, cycle_up, val))
+		return;
+
+	obj = pdf_dict_get(ctx, val, PDF_NAME(Name));
+	if (obj)
+	{
+		flush_begin_layer(ctx, proc);
+		fz_end_layer(ctx, proc->dev);
+		return;
+	}
+
+	obj = pdf_dict_get(ctx, val, PDF_NAME(OCGs));
+	n = pdf_array_len(ctx, obj);
+	for (i = n-1; i >= 0; i--)
+	{
+		end_oc(ctx, proc, pdf_array_get(ctx, obj, i), &cycle);
+	}
+}
+
+static void
+begin_layer(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val)
+{
+	/* val has been resolved to a dict for us by the originally specified name
+	 * having been looked up in Properties already for us. Go with the 'Title'
+	 * entry. */
+	pdf_obj *obj = pdf_dict_get(ctx, val, PDF_NAME(Title));
+	if (obj)
+	{
+		pdf_flush_text(ctx, proc);
+		push_begin_layer(ctx, proc, pdf_to_text_string(ctx, obj));
+	}
+}
+
+static void
+end_layer(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val)
+{
+	/* val has been resolved to a dict for us by the originally specified name
+	 * having been looked up in Properties already for us. Go with the 'Title'
+	 * entry. */
+	pdf_obj *obj = pdf_dict_get(ctx, val, PDF_NAME(Title));
+	if (obj)
+	{
+		fz_end_layer(ctx, proc->dev);
+	}
+}
+
+static void
+pop_structure_to(fz_context *ctx, pdf_run_processor *proc, pdf_obj *common)
+{
+	pdf_obj *struct_tree_root = pdf_dict_getl(ctx, pdf_trailer(ctx, proc->doc), PDF_NAME(Root), PDF_NAME(StructTreeRoot), NULL);
+
+	while (pdf_objcmp(ctx, proc->mcid_sent, common))
+	{
+		pdf_obj *p = pdf_dict_get(ctx, proc->mcid_sent, PDF_NAME(P));
+
+		fz_end_structure(ctx, proc->dev);
+		pdf_drop_obj(ctx, proc->mcid_sent);
+		proc->mcid_sent = pdf_keep_obj(ctx, p);
+		if (!pdf_objcmp(ctx, p, struct_tree_root))
+		{
+			pdf_drop_obj(ctx, proc->mcid_sent);
+			proc->mcid_sent = NULL;
+			break;
+		}
+	}
+}
+
+struct line
+{
+	pdf_obj *obj;
+	struct line *child;
+};
+
+static pdf_obj *
+find_most_recent_common_ancestor_imp(fz_context *ctx, pdf_obj *a, struct line *line_a, pdf_obj *b, struct line *line_b, pdf_cycle_list *cycle_up_a, pdf_cycle_list *cycle_up_b)
+{
+	struct line line;
+	pdf_obj *common = NULL;
+	pdf_cycle_list cycle;
+	pdf_obj *parent;
+
+	/* First ascend one lineage. */
+	if (pdf_is_dict(ctx, a))
+	{
+		if (pdf_cycle(ctx, &cycle, cycle_up_a, a))
+			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in structure tree");
+		line.obj = a;
+		line.child = line_a;
+		parent = pdf_dict_get(ctx, a, PDF_NAME(P));
+		return find_most_recent_common_ancestor_imp(ctx, parent, &line, b, NULL, &cycle, NULL);
+	}
+	/* Then ascend the other lineage. */
+	else if (pdf_is_dict(ctx, b))
+	{
+		if (pdf_cycle(ctx, &cycle, cycle_up_b, b))
+			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in structure tree");
+		line.obj = b;
+		line.child = line_b;
+		parent = pdf_dict_get(ctx, b, PDF_NAME(P));
+		return find_most_recent_common_ancestor_imp(ctx, a, line_a, parent, &line, cycle_up_a, &cycle);
+	}
+
+	/* Once both lineages are know, traverse top-down to find most recent common ancestor. */
+	while (line_a && line_b && !pdf_objcmp(ctx, line_a->obj, line_b->obj))
+	{
+		common = line_a->obj;
+		line_a = line_a->child;
+		line_b = line_b->child;
+	}
+	return common;
+}
+
+static pdf_obj *
+find_most_recent_ancestor(fz_context *ctx, pdf_obj *a, pdf_obj *b)
+{
+	if (!pdf_is_dict(ctx, a) || !pdf_is_dict(ctx, b))
+		return NULL;
+	return find_most_recent_common_ancestor_imp(ctx, a, NULL, b, NULL, NULL, NULL);
+}
+
+static void
+send_begin_structure(fz_context *ctx, pdf_run_processor *proc, pdf_obj *mc_dict)
+{
+	pdf_obj *common = NULL;
+
+	/* We are currently nested in A,B,C,...E,F,mcid_sent. We want to update to
+	 * being in A,B,C,...G,H,mc_dict. So we need to find the lowest common point. */
+	common = find_most_recent_ancestor(ctx, proc->mcid_sent, mc_dict);
+
+	/* So, we need to pop everything up to common (i.e. everything below common will be closed). */
+	pop_structure_to(ctx, proc, common);
+
+	/* Now we need to send everything between common (proc->mcid_sent) and mc_dict.
+	 * Again, n^2 will do... */
+	while (pdf_objcmp(ctx, proc->mcid_sent, mc_dict))
+	{
+		pdf_obj *send = mc_dict;
+		fz_structure standard;
+		pdf_obj *tag;
+		int uid;
+
+		while (1) {
+			pdf_obj *p = pdf_dict_get(ctx, send, PDF_NAME(P));
+
+			if (!pdf_objcmp(ctx, p, proc->mcid_sent))
+				break;
+			send = p;
+		}
+
+		uid = pdf_to_num(ctx, send);
+		tag = pdf_dict_get(ctx, send, PDF_NAME(S));
+		standard = structure_type(ctx, proc, tag);
+		if (standard != FZ_STRUCTURE_INVALID)
+			fz_begin_structure(ctx, proc->dev, standard, pdf_to_name(ctx, tag), uid);
+
+		pdf_drop_obj(ctx, proc->mcid_sent);
+		proc->mcid_sent = pdf_keep_obj(ctx, send);
+	}
+}
+
+static void
+push_marked_content(fz_context *ctx, pdf_run_processor *proc, const char *tagstr, pdf_obj *val)
+{
+	pdf_obj *tag;
+	marked_content_stack *mc = NULL;
+	int drop_tag = 1;
+	fz_structure standard;
+	pdf_obj *mc_dict = NULL;
+
+	/* Flush any pending text so it's not in the wrong layer. */
+	pdf_flush_text(ctx, proc);
+
+	if (!tagstr)
+		tagstr = "Untitled";
+	tag = pdf_new_name(ctx, tagstr);
+
+	fz_var(drop_tag);
+
+	fz_try(ctx)
+	{
+		/* First, push it on the stack. */
+		mc = fz_malloc_struct(ctx, marked_content_stack);
+		mc->next = proc->marked_content;
+		mc->tag = tag;
+		mc->val = pdf_keep_obj(ctx, val);
+		proc->marked_content = mc;
+		drop_tag = 0;
+
+		/* Check to see if val contains an MCID. */
+		mc_dict = lookup_mcid(ctx, proc, val);
+
+		/* Start any optional content layers. */
+		if (pdf_name_eq(ctx, tag, PDF_NAME(OC)))
+			begin_oc(ctx, proc, val, NULL);
+
+		/* Special handling for common non-spec extension. */
+		if (pdf_name_eq(ctx, tag, PDF_NAME(Layer)))
+			begin_layer(ctx, proc, val);
+
+		/* Structure */
+		if (mc_dict && !proc->broken_struct_tree)
+		{
+			fz_try(ctx)
+				send_begin_structure(ctx, proc, mc_dict);
+			fz_catch(ctx)
+			{
+				fz_warn(ctx, "structure tree broken, assume tree is missing: %s", fz_caught_message(ctx));
+				proc->broken_struct_tree = 1;
+			}
+		}
+
+		if (!mc_dict || proc->broken_struct_tree)
+		{
+			standard = structure_type(ctx, proc, tag);
+			if (standard != FZ_STRUCTURE_INVALID)
+			{
+				pdf_flush_text(ctx, proc);
+				fz_begin_structure(ctx, proc->dev, standard, pdf_to_name(ctx, tag), 0);
+			}
+		}
+
+		/* ActualText */
+		begin_metatext(ctx, proc, val, mc_dict, FZ_METATEXT_ACTUALTEXT, PDF_NAME(ActualText));
+
+		/* Alt */
+		begin_metatext(ctx, proc, val, mc_dict, FZ_METATEXT_ALT, PDF_NAME(Alt));
+
+		/* Abbreviation */
+		begin_metatext(ctx, proc, val, mc_dict, FZ_METATEXT_ABBREVIATION, PDF_NAME(E));
+
+		/* Title */
+		begin_metatext(ctx, proc, val, mc_dict, FZ_METATEXT_TITLE, PDF_NAME(T));
+	}
+	fz_catch(ctx)
+	{
+		if (drop_tag)
+			pdf_drop_obj(ctx, tag);
+		fz_rethrow(ctx);
+	}
+}
+
+static void
+pop_marked_content(fz_context *ctx, pdf_run_processor *proc, int neat)
+{
+	marked_content_stack *mc = proc->marked_content;
+	pdf_obj *val, *tag;
+	fz_structure standard;
+	pdf_obj *mc_dict = NULL;
+
+	if (mc == NULL)
+		return;
+
+	proc->marked_content = mc->next;
+	tag = mc->tag;
+	val = mc->val;
+	fz_free(ctx, mc);
+
+	/* If we're not interested in neatly closing any open layers etc
+	 * in the processor, (maybe we've had errors already), then just
+	 * exit here. */
+	if (!neat)
+	{
+		pdf_drop_obj(ctx, tag);
+		pdf_drop_obj(ctx, val);
+		return;
+	}
+
+	/* Make sure that any pending text is written into the correct layer. */
+	pdf_flush_text(ctx, proc);
+
+	/* Close structure/layers here, in reverse order to how we opened them. */
+	fz_try(ctx)
+	{
+		/* Check to see if val contains an MCID. */
+		mc_dict = lookup_mcid(ctx, proc, val);
+
+		/* Title */
+		end_metatext(ctx, proc, val, mc_dict, PDF_NAME(T));
+
+		/* Abbreviation */
+		end_metatext(ctx, proc, val, mc_dict, PDF_NAME(E));
+
+		/* Alt */
+		end_metatext(ctx, proc, val, mc_dict, PDF_NAME(Alt));
+
+		/* ActualText */
+		end_metatext(ctx, proc, val, mc_dict, PDF_NAME(ActualText));
+
+		/* Structure */
+		if (mc_dict)
+		{
+			/* Do nothing for now. */
+		}
+		else
+		{
+			/* Maybe drop this entirely? */
+			standard = structure_type(ctx, proc, tag);
+			if (standard != FZ_STRUCTURE_INVALID)
+			{
+				pdf_flush_text(ctx, proc);
+				fz_end_structure(ctx, proc->dev);
+			}
+		}
+
+		/* Finally, close any layers. */
+		if (pdf_name_eq(ctx, tag, PDF_NAME(Layer)))
+			end_layer(ctx, proc, val);
+
+		if (pdf_name_eq(ctx, tag, PDF_NAME(OC)))
+			end_oc(ctx, proc, val, NULL);
+	}
+	fz_always(ctx)
+	{
+		pdf_drop_obj(ctx, tag);
+		pdf_drop_obj(ctx, val);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
+static void
+clear_marked_content(fz_context *ctx, pdf_run_processor *pr)
+{
+	if (pr->marked_content == NULL)
+		return;
+
+	fz_try(ctx)
+		while (pr->marked_content)
+			pop_marked_content(ctx, pr, 1);
+	fz_always(ctx)
+		while (pr->marked_content)
+			pop_marked_content(ctx, pr, 0);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
+static void
+pdf_run_xobject(fz_context *ctx, pdf_run_processor *pr, pdf_obj *xobj, pdf_obj *page_resources, fz_matrix transform, int is_smask)
+{
+	pdf_cycle_list cycle_here;
 	pdf_gstate *gstate = NULL;
 	int oldtop = 0;
 	int oldbot = -1;
@@ -1198,10 +1919,17 @@ pdf_run_xobject(fz_context *ctx, pdf_run_processor *proc, pdf_obj *xobj, pdf_obj
 	fz_colorspace *cs = NULL;
 	fz_default_colorspaces *save_default_cs = NULL;
 	fz_default_colorspaces *xobj_default_cs = NULL;
+	marked_content_stack *save_marked_content = NULL;
+	int save_struct_parent;
+	pdf_obj *struct_parent;
 
 	/* Avoid infinite recursion */
-	if (xobj == NULL || pdf_mark_obj(ctx, xobj))
+	pdf_cycle_list *cycle_up = pr->cycle;
+	if (xobj == NULL || pdf_cycle(ctx, &cycle_here, cycle_up, xobj))
 		return;
+	pr->cycle = &cycle_here;
+
+	flush_begin_layer(ctx, pr);
 
 	fz_var(cs);
 	fz_var(xobj_default_cs);
@@ -1211,9 +1939,18 @@ pdf_run_xobject(fz_context *ctx, pdf_run_processor *proc, pdf_obj *xobj, pdf_obj
 	oldtop = pr->gtop;
 
 	save_default_cs = pr->default_cs;
+	save_marked_content = pr->marked_content;
+	pr->marked_content = NULL;
+	save_struct_parent = pr->struct_parent;
+
+	pr->struct_parent = -1;
 
 	fz_try(ctx)
 	{
+		struct_parent = pdf_dict_get(ctx, xobj, PDF_NAME(StructParent));
+		if (pdf_is_number(ctx, struct_parent))
+			pr->struct_parent = pdf_to_int(ctx, struct_parent);
+
 		pdf_gsave(ctx, pr);
 
 		gstate = pr->gstate + pr->gtop;
@@ -1289,7 +2026,7 @@ pdf_run_xobject(fz_context *ctx, pdf_run_processor *proc, pdf_obj *xobj, pdf_obj
 		oldbot = pr->gbot;
 		pr->gbot = pr->gtop;
 
-		pdf_process_contents(ctx, (pdf_processor*)pr, doc, resources, xobj, pr->cookie);
+		pdf_process_contents(ctx, (pdf_processor*)pr, doc, resources, xobj, pr->cookie, NULL);
 
 		/* Undo any gstate mismatches due to the pdf_process_contents call */
 		if (oldbot != -1)
@@ -1323,10 +2060,13 @@ pdf_run_xobject(fz_context *ctx, pdf_run_processor *proc, pdf_obj *xobj, pdf_obj
 	}
 	fz_always(ctx)
 	{
+		clear_marked_content(ctx, pr);
+		pr->marked_content = save_marked_content;
 		pr->default_cs = save_default_cs;
 		fz_drop_default_colorspaces(ctx, xobj_default_cs);
 		fz_drop_colorspace(ctx, cs);
-		pdf_unmark_obj(ctx, xobj);
+		pr->cycle = cycle_up;
+		pr->struct_parent = save_struct_parent;
 	}
 	fz_catch(ctx)
 	{
@@ -1348,6 +2088,8 @@ static void pdf_run_w(fz_context *ctx, pdf_processor *proc, float linewidth)
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pdf_gstate *gstate = pdf_flush_text(ctx, pr);
 
+	flush_begin_layer(ctx, pr);
+
 	pr->dev->flags &= ~FZ_DEVFLAG_LINEWIDTH_UNDEFINED;
 	gstate->stroke_state = fz_unshare_stroke_state(ctx, gstate->stroke_state);
 	gstate->stroke_state->linewidth = linewidth;
@@ -1476,7 +2218,7 @@ static void pdf_run_gs_ca(fz_context *ctx, pdf_processor *proc, float alpha)
 	gstate->fill.alpha = fz_clamp(alpha, 0, 1);
 }
 
-static void pdf_run_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, pdf_obj *page_resources, float *bc, int luminosity)
+static void pdf_run_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, float *bc, int luminosity)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pdf_gstate *gstate = pdf_flush_text(ctx, pr);
@@ -1498,7 +2240,7 @@ static void pdf_run_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smas
 			cs_n = fz_colorspace_n(ctx, cs);
 		gstate->softmask_ctm = gstate->ctm;
 		gstate->softmask = pdf_keep_obj(ctx, smask);
-		gstate->softmask_resources = pdf_keep_obj(ctx, page_resources);
+		gstate->softmask_resources = pdf_keep_obj(ctx, pr->rstack->resources);
 		for (i = 0; i < cs_n; ++i)
 			gstate->softmask_bc[i] = bc[i];
 		gstate->luminosity = luminosity;
@@ -1511,6 +2253,7 @@ static void pdf_run_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smas
 static void pdf_run_q(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
+	flush_begin_layer(ctx, pr);
 	pdf_gsave(ctx, pr);
 }
 
@@ -1941,6 +2684,8 @@ static void pdf_run_BI(fz_context *ctx, pdf_processor *proc, fz_image *image, co
 static void pdf_run_sh(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *shade)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
+
+	flush_begin_layer(ctx, pr);
 	pdf_show_shade(ctx, pr, shade);
 }
 
@@ -1950,9 +2695,10 @@ static void pdf_run_Do_image(fz_context *ctx, pdf_processor *proc, const char *n
 	pdf_show_image(ctx, pr, image);
 }
 
-static void pdf_run_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *xobj, pdf_obj *page_resources)
+static void pdf_run_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *xobj)
 {
-	pdf_run_xobject(ctx, (pdf_run_processor*)proc, xobj, page_resources, fz_identity, 0);
+	pdf_run_processor *pr = (pdf_run_processor *)proc;
+	pdf_run_xobject(ctx, (pdf_run_processor*)proc, xobj, pr->rstack->resources, fz_identity, 0);
 }
 
 /* marked content */
@@ -1960,33 +2706,20 @@ static void pdf_run_Do_form(fz_context *ctx, pdf_processor *proc, const char *na
 static void pdf_run_BMC(fz_context *ctx, pdf_processor *proc, const char *tag)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
-
-	if (!tag)
-		tag = "Untitled";
-
-	fz_begin_layer(ctx, pr->dev, tag);
+	push_marked_content(ctx, pr, tag, NULL);
 }
 
 static void pdf_run_BDC(fz_context *ctx, pdf_processor *proc, const char *tag, pdf_obj *raw, pdf_obj *cooked)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
-	const char *str;
-
-	if (!tag)
-		tag = "Untitled";
-
-	str = pdf_dict_get_text_string(ctx, cooked, PDF_NAME(Name));
-	if (strlen(str) == 0)
-		str = tag;
-
-	fz_begin_layer(ctx, pr->dev, str);
+	push_marked_content(ctx, pr, tag, cooked);
 }
 
 static void pdf_run_EMC(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 
-	fz_end_layer(ctx, pr->dev);
+	pop_marked_content(ctx, pr, 1);
 }
 
 static void pdf_run_MP(fz_context *ctx, pdf_processor *proc, const char *tag)
@@ -2030,6 +2763,10 @@ pdf_close_run_processor(fz_context *ctx, pdf_processor *proc)
 		fz_pop_clip(ctx, pr->dev);
 		pr->gstate[0].clip_depth--;
 	}
+
+	pop_structure_to(ctx, pr, NULL);
+
+	clear_marked_content(ctx, pr);
 }
 
 static void
@@ -2049,6 +2786,57 @@ pdf_drop_run_processor(fz_context *ctx, pdf_processor *proc)
 	fz_drop_default_colorspaces(ctx, pr->default_cs);
 
 	fz_free(ctx, pr->gstate);
+
+	while (pr->rstack)
+	{
+		resources_stack *stk = pr->rstack;
+		pr->rstack = stk->next;
+		pdf_drop_obj(ctx, stk->resources);
+		fz_free(ctx, stk);
+	}
+
+	while (pr->begin_layer)
+	{
+		begin_layer_stack *stk = pr->begin_layer;
+		pr->begin_layer = stk->next;
+		fz_free(ctx, stk->layer);
+		fz_free(ctx, stk);
+	}
+
+	while (pr->marked_content)
+		pop_marked_content(ctx, pr, 0);
+
+	pdf_drop_obj(ctx, pr->mcid_sent);
+
+	pdf_drop_document(ctx, pr->doc);
+	pdf_drop_obj(ctx, pr->role_map);
+}
+
+static void
+pdf_run_push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *resources)
+{
+	pdf_run_processor *pr = (pdf_run_processor *)proc;
+	resources_stack *stk = fz_malloc_struct(ctx, resources_stack);
+
+	stk->next = pr->rstack;
+	pr->rstack = stk;
+	stk->resources = pdf_keep_obj(ctx, resources);
+}
+
+static pdf_obj *
+pdf_run_pop_resources(fz_context *ctx, pdf_processor *proc)
+{
+	pdf_run_processor *pr = (pdf_run_processor *)proc;
+	resources_stack *stk = pr->rstack;
+
+	if (stk)
+	{
+		pr->rstack = stk->next;
+		pdf_drop_obj(ctx, stk->resources);
+		fz_free(ctx, stk);
+	}
+
+	return NULL;
 }
 
 /*
@@ -2067,7 +2855,7 @@ pdf_drop_run_processor(fz_context *ctx, pdf_processor *proc)
 	gstate: The initial graphics state.
 */
 pdf_processor *
-pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char *usage, pdf_gstate *gstate, fz_default_colorspaces *default_cs, fz_cookie *cookie)
+pdf_new_run_processor(fz_context *ctx, pdf_document *doc, fz_device *dev, fz_matrix ctm, int struct_parent, const char *usage, pdf_gstate *gstate, fz_default_colorspaces *default_cs, fz_cookie *cookie)
 {
 	pdf_run_processor *proc = pdf_new_processor(ctx, sizeof *proc);
 	{
@@ -2076,6 +2864,9 @@ pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char
 		proc->super.close_processor = pdf_close_run_processor;
 		proc->super.drop_processor = pdf_drop_run_processor;
 
+		proc->super.push_resources = pdf_run_push_resources;
+		proc->super.pop_resources = pdf_run_pop_resources;
+
 		/* general graphics state */
 		proc->super.op_w = pdf_run_w;
 		proc->super.op_j = pdf_run_j;
@@ -2198,6 +2989,7 @@ pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char
 		proc->super.op_END = pdf_run_END;
 	}
 
+	proc->doc = pdf_keep_document(ctx, doc);
 	proc->dev = dev;
 	proc->cookie = cookie;
 
@@ -2214,6 +3006,10 @@ pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char
 
 	proc->gtop = -1;
 
+	proc->marked_content = NULL;
+
+	proc->next_begin_layer = &proc->begin_layer;
+
 	fz_try(ctx)
 	{
 		proc->path = fz_new_path(ctx);
@@ -2241,5 +3037,8 @@ pdf_new_run_processor(fz_context *ctx, fz_device *dev, fz_matrix ctm, const char
 	/* We need to save an extra level to allow for level 0 to be the parent gstate level. */
 	pdf_gsave(ctx, proc);
 
+	proc->struct_parent = struct_parent;
+	proc->role_map = pdf_keep_obj(ctx, pdf_dict_getl(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root), PDF_NAME(StructTreeRoot), PDF_NAME(RoleMap), NULL));
+
 	return (pdf_processor*)proc;
 }
