@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -85,10 +85,23 @@ pdf_to_matrix(fz_context *ctx, pdf_obj *array)
 	}
 }
 
+char *
+pdf_format_date(fz_context *ctx, int64_t time, char *s, size_t n)
+{
+	time_t secs = time;
+#ifdef _POSIX_SOURCE
+	struct tm tmbuf, *tm = gmtime_r(&secs, &tmbuf);
+#else
+	struct tm *tm = gmtime(&secs);
+#endif
+	if (time < 0 || !tm || !strftime(s, n, "D:%Y%m%d%H%M%SZ", tm))
+		return NULL;
+	return s;
+}
+
 int64_t
-pdf_to_date(fz_context *ctx, pdf_obj *time)
+pdf_parse_date(fz_context *ctx, const char *s)
 {
-	const char *s = pdf_to_str_buf(ctx, time);
 	int tz_sign, tz_hour, tz_min, tz_adj;
 	struct tm tm;
 	time_t utc;
@@ -154,7 +167,20 @@ pdf_to_date(fz_context *ctx, pdf_obj *time)
 
 	if (s[0] == 'Z')
 	{
-		s += 1;
+		if (s[1] == '0' && s[2] == '0')
+		{
+			s += 3;
+			if (s[0] == '\'' && s[1] == '0' && s[2] == '0')
+			{
+				s += 3;
+				if (s[0] == '\'')
+					s += 1;
+			}
+		}
+		else
+		{
+			s += 1;
+		}
 	}
 	else if ((s[0] == '-' || s[0] == '+') && isdigit(s[1]) && isdigit(s[2]))
 	{
@@ -196,6 +222,12 @@ pdf_to_date(fz_context *ctx, pdf_obj *time)
 	return utc - tz_adj;
 }
 
+int64_t
+pdf_to_date(fz_context *ctx, pdf_obj *time)
+{
+	return pdf_parse_date(ctx, pdf_to_str_buf(ctx, time));
+}
+
 static int
 rune_from_utf16be(int *out, const unsigned char *s, const unsigned char *end)
 {
