@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -26,22 +26,21 @@
 /* Sample various functions into lookup tables */
 
 static void
-pdf_sample_composite_shade_function(fz_context *ctx, fz_shade *shade, pdf_function *func, float t0, float t1)
+pdf_sample_composite_shade_function(fz_context *ctx, float shade[256][FZ_MAX_COLORS+1], int n, pdf_function *func, float t0, float t1)
 {
-	int i, n;
+	int i;
 	float t;
 
-	n = fz_colorspace_n(ctx, shade->colorspace);
 	for (i = 0; i < 256; i++)
 	{
 		t = t0 + (i / 255.0f) * (t1 - t0);
-		pdf_eval_function(ctx, func, &t, 1, shade->function[i], n);
-		shade->function[i][n] = 1;
+		pdf_eval_function(ctx, func, &t, 1, shade[i], n);
+		shade[i][n] = 1;
 	}
 }
 
 static void
-pdf_sample_component_shade_function(fz_context *ctx, fz_shade *shade, int funcs, pdf_function **func, float t0, float t1)
+pdf_sample_component_shade_function(fz_context *ctx, float shade[256][FZ_MAX_COLORS+1], int funcs, pdf_function **func, float t0, float t1)
 {
 	int i, k;
 	float t;
@@ -50,17 +49,16 @@ pdf_sample_component_shade_function(fz_context *ctx, fz_shade *shade, int funcs,
 	{
 		t = t0 + (i / 255.0f) * (t1 - t0);
 		for (k = 0; k < funcs; k++)
-			pdf_eval_function(ctx, func[k], &t, 1, &shade->function[i][k], 1);
-		shade->function[i][k] = 1;
+			pdf_eval_function(ctx, func[k], &t, 1, &shade[i][k], 1);
+		shade[i][k] = 1;
 	}
 }
 
-static void
-pdf_sample_shade_function(fz_context *ctx, fz_shade *shade, int funcs, pdf_function **func, float t0, float t1)
+void
+pdf_sample_shade_function(fz_context *ctx, float shade[256][FZ_MAX_COLORS+1], int n, int funcs, pdf_function **func, float t0, float t1)
 {
-	shade->use_function = 1;
 	if (funcs == 1)
-		pdf_sample_composite_shade_function(ctx, shade, func[0], t0, t1);
+		pdf_sample_composite_shade_function(ctx, shade, n, func[0], t0, t1);
 	else
 		pdf_sample_component_shade_function(ctx, shade, funcs, func, t0, t1);
 }
@@ -70,12 +68,12 @@ pdf_sample_shade_function(fz_context *ctx, fz_shade *shade, int funcs, pdf_funct
 #define FUNSEGS 64 /* size of sampled mesh for function-based shadings */
 
 static void
-pdf_load_function_based_shading(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf_obj *dict, pdf_function *func)
+pdf_load_function_based_shading(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf_obj *dict, int funcs, pdf_function **func)
 {
 	pdf_obj *obj;
 	float x0, y0, x1, y1;
 	float fv[2];
-	int xx, yy;
+	int xx, yy, zz;
 	float *p;
 	int n = fz_colorspace_n(ctx, shade->colorspace);
 
@@ -100,16 +98,40 @@ pdf_load_function_based_shading(fz_context *ctx, pdf_document *doc, fz_shade *sh
 	shade->u.f.domain[1][1] = y1;
 
 	p = shade->u.f.fn_vals;
-	for (yy = 0; yy <= FUNSEGS; yy++)
+	if (funcs == 1)
 	{
-		fv[1] = y0 + (y1 - y0) * yy / FUNSEGS;
-
-		for (xx = 0; xx <= FUNSEGS; xx++)
+		for (yy = 0; yy <= FUNSEGS; yy++)
 		{
-			fv[0] = x0 + (x1 - x0) * xx / FUNSEGS;
+			fv[1] = y0 + (y1 - y0) * yy / FUNSEGS;
 
-			pdf_eval_function(ctx, func, fv, 2, p, n);
-			p += n;
+			for (xx = 0; xx <= FUNSEGS; xx++)
+			{
+				fv[0] = x0 + (x1 - x0) * xx / FUNSEGS;
+
+				pdf_eval_function(ctx, func[0], fv, 2, p, n);
+				p += n;
+			}
+		}
+	}
+	else
+	{
+		if (funcs != n)
+			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 1 2in, n-out function, or n 2 in, 1-out functions");
+
+		for (yy = 0; yy <= FUNSEGS; yy++)
+		{
+			fv[1] = y0 + (y1 - y0) * yy / FUNSEGS;
+
+			for (xx = 0; xx <= FUNSEGS; xx++)
+			{
+				fv[0] = x0 + (x1 - x0) * xx / FUNSEGS;
+
+				for (zz = 0; zz < n; zz++)
+				{
+					pdf_eval_function(ctx, func[zz], fv, 2, p, 1);
+					p ++;
+				}
+			}
 		}
 	}
 }
@@ -144,7 +166,8 @@ pdf_load_linear_shading(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf
 		e1 = pdf_array_get_bool(ctx, obj, 1);
 	}
 
-	pdf_sample_shade_function(ctx, shade, funcs, func, d0, d1);
+	shade->use_function = 1;
+	pdf_sample_shade_function(ctx, shade->function, shade->colorspace->n, funcs, func, d0, d1);
 
 	shade->u.l_or_r.extend[0] = e0;
 	shade->u.l_or_r.extend[1] = e1;
@@ -182,7 +205,8 @@ pdf_load_radial_shading(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf
 		e1 = pdf_array_get_bool(ctx, obj, 1);
 	}
 
-	pdf_sample_shade_function(ctx, shade, funcs, func, d0, d1);
+	shade->use_function = 1;
+	pdf_sample_shade_function(ctx, shade->function, shade->colorspace->n, funcs, func, d0, d1);
 
 	shade->u.l_or_r.extend[0] = e0;
 	shade->u.l_or_r.extend[1] = e1;
@@ -271,9 +295,12 @@ pdf_load_type4_shade(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf_ob
 	pdf_load_mesh_params(ctx, doc, shade, dict);
 
 	if (funcs > 0)
-		pdf_sample_shade_function(ctx, shade, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	{
+		shade->use_function = 1;
+		pdf_sample_shade_function(ctx, shade->function, shade->colorspace->n, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	}
 
-	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict));
+	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict), 0);
 }
 
 static void
@@ -282,9 +309,12 @@ pdf_load_type5_shade(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf_ob
 	pdf_load_mesh_params(ctx, doc, shade, dict);
 
 	if (funcs > 0)
-		pdf_sample_shade_function(ctx, shade, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	{
+		shade->use_function = 1;
+		pdf_sample_shade_function(ctx, shade->function, shade->colorspace->n, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	}
 
-	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict));
+	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict), 0);
 }
 
 /* Type 6 & 7 -- Patch mesh shadings */
@@ -295,9 +325,12 @@ pdf_load_type6_shade(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf_ob
 	pdf_load_mesh_params(ctx, doc, shade, dict);
 
 	if (funcs > 0)
-		pdf_sample_shade_function(ctx, shade, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	{
+		shade->use_function = 1;
+		pdf_sample_shade_function(ctx, shade->function, shade->colorspace->n, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	}
 
-	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict));
+	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict), 0);
 }
 
 static void
@@ -306,9 +339,12 @@ pdf_load_type7_shade(fz_context *ctx, pdf_document *doc, fz_shade *shade, pdf_ob
 	pdf_load_mesh_params(ctx, doc, shade, dict);
 
 	if (funcs > 0)
-		pdf_sample_shade_function(ctx, shade, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	{
+		shade->use_function = 1;
+		pdf_sample_shade_function(ctx, shade->function, shade->colorspace->n, funcs, func, shade->u.m.c0[0], shade->u.m.c1[0]);
+	}
 
-	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict));
+	shade->buffer = pdf_load_compressed_stream(ctx, doc, pdf_to_num(ctx, dict), 0);
 }
 
 /* Load all of the shading dictionary parameters, then switch on the shading type. */
@@ -414,7 +450,7 @@ pdf_load_shading_dict(fz_context *ctx, pdf_document *doc, pdf_obj *dict, fz_matr
 		shade->type = type;
 		switch (type)
 		{
-		case 1: pdf_load_function_based_shading(ctx, doc, shade, dict, func[0]); break;
+		case 1: pdf_load_function_based_shading(ctx, doc, shade, dict, funcs, func); break;
 		case 2: pdf_load_linear_shading(ctx, doc, shade, dict, funcs, func); break;
 		case 3: pdf_load_radial_shading(ctx, doc, shade, dict, funcs, func); break;
 		case 4: pdf_load_type4_shade(ctx, doc, shade, dict, funcs, func); break;
