@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -197,7 +197,7 @@ build_compression_params(fz_context *ctx, pdf_obj *f, pdf_obj *p, fz_compression
  * Create a filter given a name and param dictionary.
  */
 static fz_stream *
-build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params)
+build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params, int might_be_image)
 {
 	fz_compression_params local_params;
 
@@ -205,6 +205,18 @@ build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, p
 	if (params == NULL)
 		params = &local_params;
 
+	if (!might_be_image &&
+		(pdf_name_eq(ctx, f, PDF_NAME(CCITTFaxDecode)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(CCF)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(DCTDecode)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(DCT)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(JBIG2Decode)) ||
+			pdf_name_eq(ctx, f, PDF_NAME(JPXDecode))))
+	{
+		fz_warn(ctx, "Can't open image only stream for non-image purposes");
+		return fz_open_memory(ctx, (unsigned char *)"", 0);
+	}
+
 	build_compression_params(ctx, f, p, params);
 
 	/* If we were using params we were passed in, and we successfully
@@ -257,11 +269,11 @@ build_filter(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *f, p
 
 /* Build filter, and assume ownership of chain */
 static fz_stream *
-build_filter_drop(fz_context *ctx, fz_stream *tail, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params)
+build_filter_drop(fz_context *ctx, fz_stream *tail, pdf_document *doc, pdf_obj *f, pdf_obj *p, int num, int gen, fz_compression_params *params, int might_be_image)
 {
 	fz_stream *head;
 	fz_try(ctx)
-		head = build_filter(ctx, tail, doc, f, p, num, gen, params);
+		head = build_filter(ctx, tail, doc, f, p, num, gen, params, might_be_image);
 	fz_always(ctx)
 		fz_drop_stream(ctx, tail);
 	fz_catch(ctx)
@@ -275,7 +287,7 @@ build_filter_drop(fz_context *ctx, fz_stream *tail, pdf_document *doc, pdf_obj *
  * Assume ownership of chain.
  */
 static fz_stream *
-build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params)
+build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params, int might_be_image)
 {
 	fz_var(chain);
 	fz_try(ctx)
@@ -285,7 +297,7 @@ build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pd
 		{
 			pdf_obj *f = pdf_array_get(ctx, fs, i);
 			pdf_obj *p = pdf_array_get(ctx, ps, i);
-			chain = build_filter_drop(ctx, chain, doc, f, p, num, gen, (i == n-1 ? params : NULL));
+			chain = build_filter_drop(ctx, chain, doc, f, p, num, gen, (i == n-1 ? params : NULL), might_be_image);
 		}
 	}
 	fz_catch(ctx)
@@ -294,9 +306,9 @@ build_filter_chain_drop(fz_context *ctx, fz_stream *chain, pdf_document *doc, pd
 }
 
 static fz_stream *
-build_filter_chain(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params)
+build_filter_chain(fz_context *ctx, fz_stream *chain, pdf_document *doc, pdf_obj *fs, pdf_obj *ps, int num, int gen, fz_compression_params *params, int might_be_image)
 {
-	return build_filter_chain_drop(ctx, fz_keep_stream(ctx, chain), doc, fs, ps, num, gen, params);
+	return build_filter_chain_drop(ctx, fz_keep_stream(ctx, chain), doc, fs, ps, num, gen, params, might_be_image);
 }
 
 /*
@@ -313,7 +325,7 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
 	pdf_xref_entry *x = NULL;
 	fz_stream *null_stm, *crypt_stm;
 	int hascrypt;
-	int len;
+	int64_t len;
 
 	if (num > 0 && num < pdf_xref_len(ctx, doc))
 	{
@@ -335,10 +347,10 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
 	}
 
 	hascrypt = pdf_stream_has_crypt(ctx, stmobj);
-	len = pdf_dict_get_int(ctx, stmobj, PDF_NAME(Length));
+	len = pdf_dict_get_int64(ctx, stmobj, PDF_NAME(Length));
 	if (len < 0)
 		len = 0;
-	null_stm = fz_open_endstream_filter(ctx, file_stm, len, offset);
+	null_stm = fz_open_endstream_filter(ctx, file_stm, (uint64_t)len, offset);
 	if (doc->crypt && !hascrypt)
 	{
 		fz_try(ctx)
@@ -357,7 +369,7 @@ pdf_open_raw_filter(fz_context *ctx, fz_stream *file_stm, pdf_document *doc, pdf
  * to stream length and decrypting.
  */
 static fz_stream *
-pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj *stmobj, int num, int64_t offset, fz_compression_params *imparams)
+pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj *stmobj, int num, int64_t offset, fz_compression_params *imparams, int might_be_image)
 {
 	pdf_obj *filters = pdf_dict_geta(ctx, stmobj, PDF_NAME(Filter), PDF_NAME(F));
 	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME(DecodeParms), PDF_NAME(DP));
@@ -368,9 +380,9 @@ pdf_open_filter(fz_context *ctx, pdf_document *doc, fz_stream *file_stm, pdf_obj
 	fz_try(ctx)
 	{
 		if (pdf_is_name(ctx, filters))
-			fstm = build_filter(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams);
+			fstm = build_filter(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams, might_be_image);
 		else if (pdf_array_len(ctx, filters) > 0)
-			fstm = build_filter_chain(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams);
+			fstm = build_filter_chain(ctx, rstm, doc, filters, params, orig_num, orig_gen, imparams, might_be_image);
 		else
 		{
 			if (imparams)
@@ -393,9 +405,9 @@ pdf_open_inline_stream(fz_context *ctx, pdf_document *doc, pdf_obj *stmobj, int
 	pdf_obj *params = pdf_dict_geta(ctx, stmobj, PDF_NAME(DecodeParms), PDF_NAME(DP));
 
 	if (pdf_is_name(ctx, filters))
-		return build_filter(ctx, file_stm, doc, filters, params, 0, 0, imparams);
+		return build_filter(ctx, file_stm, doc, filters, params, 0, 0, imparams, 1);
 	else if (pdf_array_len(ctx, filters) > 0)
-		return build_filter_chain(ctx, file_stm, doc, filters, params, 0, 0, imparams);
+		return build_filter_chain(ctx, file_stm, doc, filters, params, 0, 0, imparams, 1);
 
 	if (imparams)
 		imparams->type = FZ_IMAGE_RAW;
@@ -453,7 +465,7 @@ pdf_open_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 }
 
 static fz_stream *
-pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params)
+pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compression_params *params, int might_be_image)
 {
 	pdf_xref_entry *x;
 
@@ -461,13 +473,13 @@ pdf_open_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 	if (x->stm_ofs == 0 && x->stm_buf == NULL)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
 
-	return pdf_open_filter(ctx, doc, doc->file, x->obj, num, x->stm_ofs, params);
+	return pdf_open_filter(ctx, doc, doc->file, x->obj, num, x->stm_ofs, params, might_be_image);
 }
 
 fz_stream *
 pdf_open_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
-	return pdf_open_image_stream(ctx, doc, num, NULL);
+	return pdf_open_image_stream(ctx, doc, num, NULL, 1);
 }
 
 fz_stream *
@@ -475,7 +487,7 @@ pdf_open_stream_with_offset(fz_context *ctx, pdf_document *doc, int num, pdf_obj
 {
 	if (stm_ofs == 0)
 		fz_throw(ctx, FZ_ERROR_GENERIC, "object is not a stream");
-	return pdf_open_filter(ctx, doc, doc->file, dict, num, stm_ofs, NULL);
+	return pdf_open_filter(ctx, doc, doc->file, dict, num, stm_ofs, NULL, 1);
 }
 
 fz_buffer *
@@ -483,7 +495,7 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 {
 	fz_stream *stm;
 	pdf_obj *dict;
-	int len;
+	int64_t len;
 	fz_buffer *buf = NULL;
 	pdf_xref_entry *x;
 
@@ -497,7 +509,7 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 	dict = pdf_load_object(ctx, doc, num);
 
 	fz_try(ctx)
-		len = pdf_dict_get_int(ctx, dict, PDF_NAME(Length));
+		len = pdf_dict_get_int64(ctx, dict, PDF_NAME(Length));
 	fz_always(ctx)
 		pdf_drop_obj(ctx, dict);
 	fz_catch(ctx)
@@ -505,8 +517,11 @@ pdf_load_raw_stream_number(fz_context *ctx, pdf_document *doc, int num)
 
 	stm = pdf_open_raw_stream_number(ctx, doc, num);
 
+	if (len < 0)
+		len = 1024;
+
 	fz_try(ctx)
-		buf = fz_read_all(ctx, stm, len);
+		buf = fz_read_all(ctx, stm, (size_t)len);
 	fz_always(ctx)
 		fz_drop_stream(ctx, stm);
 	fz_catch(ctx)
@@ -639,7 +654,7 @@ pdf_load_image_stream(fz_context *ctx, pdf_document *doc, int num, fz_compressio
 		fz_rethrow(ctx);
 	}
 
-	stm = pdf_open_image_stream(ctx, doc, num, params);
+	stm = pdf_open_image_stream(ctx, doc, num, params, 1);
 
 	fz_try(ctx)
 	{
@@ -718,7 +733,7 @@ pdf_open_contents_stream(fz_context *ctx, pdf_document *doc, pdf_obj *obj)
 
 	num = pdf_to_num(ctx, obj);
 	if (pdf_is_stream(ctx, obj))
-		return pdf_open_image_stream(ctx, doc, num, NULL);
+		return pdf_open_image_stream(ctx, doc, num, NULL, 0);
 
 	fz_warn(ctx, "content stream is not a stream (%d 0 R)", num);
 	return fz_open_memory(ctx, (unsigned char *)"", 0);
