@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "svg-imp.h"
@@ -52,7 +52,7 @@ const char *
 svg_lex_number(float *fp, const char *ss)
 {
 	const char *s = ss;
-	if (*s == '-')
+	if (*s == '+' || *s == '-')
 		++s;
 	while (*s >= '0' && *s <= '9')
 		++s;
