@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 /*
  * muraster -- Convert a document to a raster file.
@@ -501,13 +501,11 @@ static struct {
 	char *maxfilename;
 } timing;
 
-#define stringify(A) #A
-
 static int usage(void)
 {
 	fprintf(stderr,
 		"muraster version " FZ_VERSION "\n"
-		"Usage: muraster [options] file [pages]\n"
+		"usage: muraster [options] file [pages]\n"
 		"\t-p -\tpassword\n"
 		"\n"
 		"\t-o -\toutput file name\n"
@@ -520,9 +518,9 @@ static int usage(void)
 		"\n"
 		"\t-R {auto,0,90,180,270}\n"
 		"\t\trotate clockwise (default: auto)\n"
-		"\t-r -{,_}\tx and y resolution in dpi (default: " stringify(X_RESOLUTION) "x" stringify(Y_RESOLUTION) ")\n"
-		"\t-w -\tprintable width (in inches) (default: " stringify(PAPER_WIDTH) ")\n"
-		"\t-h -\tprintable height (in inches) (default: " stringify(PAPER_HEIGHT) "\n"
+		"\t-r -{,_}\tx and y resolution in dpi (default: %dx%d)\n"
+		"\t-w -\tprintable width (in inches) (default: %.2f)\n"
+		"\t-h -\tprintable height (in inches) (default: %.2f)\n"
 		"\t-f\tfit file to page if too large\n"
 		"\t-B -\tminimum band height (e.g. 32)\n"
 		"\t-M -\tmax bandmemory (e.g. 655360)\n"
@@ -540,7 +538,8 @@ static int usage(void)
 		"\t-A -\tnumber of bits of antialiasing (0 to 8)\n"
 		"\t-A -/-\tnumber of bits of antialiasing (0 to 8) (graphics, text)\n"
 		"\n"
-		"\tpages\tcomma separated list of page numbers and ranges\n"
+		"\tpages\tcomma separated list of page numbers and ranges\n",
+		X_RESOLUTION, Y_RESOLUTION, PAPER_WIDTH, PAPER_HEIGHT
 		);
 	return 1;
 }
@@ -565,6 +564,8 @@ static int drawband(fz_context *ctx, fz_page *page, fz_display_list *list, fz_ma
 
 	*bit = NULL;
 
+	fz_var(dev);
+
 	fz_try(ctx)
 	{
 		fz_clear_pixmap_with_value(ctx, pix, 255);
@@ -635,11 +636,11 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 					ibounds.y1 = ibounds.y0 + remaining_height;
 				remaining_height -= band_height;
 				w->pix = fz_new_pixmap_with_bbox(ctx, colorspace, ibounds, NULL, 0);
+				w->pix->y += band * band_height;
 				fz_set_pixmap_resolution(ctx, w->pix, x_resolution, y_resolution);
 				DEBUG_THREADS(("Worker %d, Pre-triggering band %d\n", band, band));
 				w->started = 1;
 				mu_trigger_semaphore(&w->start);
-				ctm.f -= band_height;
 			}
 			pix = workers[0].pix;
 		}
@@ -699,7 +700,8 @@ static int dodrawpage(fz_context *ctx, int pagenum, fz_cookie *cookie, render_de
 				w->started = 1;
 				mu_trigger_semaphore(&w->start);
 			}
-			ctm.f -= draw_height;
+			if (render->num_workers <= 0)
+				pix += draw_height;
 		}
 	}
 	fz_always(ctx)
@@ -802,6 +804,8 @@ static int try_render_page(fz_context *ctx, int pagenum, fz_cookie *cookie, int
 		break;
 	}
 
+	fz_close_band_writer(ctx, render->bander);
+
 	fz_drop_page(ctx, render->page);
 	fz_drop_display_list(ctx, render->list);
 	fz_drop_band_writer(ctx, render->bander);
@@ -999,16 +1003,21 @@ initialise_banding(fz_context *ctx, render_details *render, int color)
 	}
 
 	w = render->ibounds.x1 - render->ibounds.x0;
+	h = render->ibounds.y1 - render->ibounds.y0;
+	if (w <= 0 || h <= 0)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Invalid page dimensions");
+
+
 	min_band_mem = (size_t)bpp * w * min_band_height;
-	reps = (int)(max_band_memory / min_band_mem);
-	if (reps < 1)
+	if (min_band_mem > 0)
+		reps = (int)(max_band_memory / min_band_mem);
+	if (min_band_mem == 0 || reps < 1)
 		reps = 1;
 
 	/* Adjust reps to even out the work between threads */
 	if (render->num_workers > 0)
 	{
 		int runs, num_bands;
-		h = render->ibounds.y1 - render->ibounds.y0;
 		num_bands = (h + min_band_height - 1) / min_band_height;
 		/* num_bands = number of min_band_height bands */
 		runs = (num_bands + reps-1) / reps;
@@ -1415,6 +1424,13 @@ read_resolution(const char *arg)
 		y_resolution = fz_atoi(arg);
 	else
 		y_resolution = x_resolution;
+
+	if (x_resolution <= 0 || y_resolution <= 0)
+	{
+		fprintf(stderr, "Ignoring invalid resolution\n");
+		x_resolution =  X_RESOLUTION;
+		y_resolution =  Y_RESOLUTION;
+	}
 }
 
 static int
@@ -1714,7 +1730,7 @@ int main(int argc, char **argv)
 	fz_catch(ctx)
 	{
 		fz_drop_document(ctx, doc);
-		fprintf(stderr, "error: cannot draw '%s'\n", filename);
+		fz_log_error_printf(ctx, "cannot draw '%s'", filename);
 		errored = 1;
 	}
 
