@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 /*
  * PDF creation tool: Tool for creating pdf content.
  *
@@ -11,7 +33,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 
-static void usage(void)
+static int usage(void)
 {
 	fprintf(stderr,
 		"usage: mutool create [-o output.pdf] [-O options] page.txt [page2.txt ...]\n"
@@ -26,7 +48,7 @@ static void usage(void)
 		"\t%%%%Image Name Filename\n\n"
 		);
 	fputs(fz_pdf_write_options_usage, stderr);
-	exit(1);
+	return 1;
 }
 
 static fz_context *ctx = NULL;
@@ -143,75 +165,87 @@ static void create_page(char *input)
 
 	char line[4096];
 	char *s, *p;
-	fz_stream *stm;
+	fz_stream *stm = NULL;
 
-	fz_buffer *contents;
+	fz_buffer *contents = NULL;
 	pdf_obj *resources;
-	pdf_obj *page;
+	pdf_obj *page = NULL;
 
 	resources = pdf_new_dict(ctx, doc, 2);
-	contents = fz_new_buffer(ctx, 1024);
 
-	stm = fz_open_file(ctx, input);
-	while (fz_read_line(ctx, stm, line, sizeof line))
+	fz_var(stm);
+	fz_var(page);
+	fz_var(contents);
+
+	fz_try(ctx)
 	{
-		if (line[0] == '%' && line[1] == '%')
+		contents = fz_new_buffer(ctx, 1024);
+
+		stm = fz_open_file(ctx, input);
+		while (fz_read_line(ctx, stm, line, sizeof line))
 		{
-			p = line;
-			s = fz_strsep(&p, " ");
-			if (!strcmp(s, "%%MediaBox"))
+			if (line[0] == '%' && line[1] == '%')
 			{
-				mediabox.x0 = fz_atoi(fz_strsep(&p, " "));
-				mediabox.y0 = fz_atoi(fz_strsep(&p, " "));
-				mediabox.x1 = fz_atoi(fz_strsep(&p, " "));
-				mediabox.y1 = fz_atoi(fz_strsep(&p, " "));
+				p = line;
+				s = fz_strsep(&p, " ");
+				if (!strcmp(s, "%%MediaBox"))
+				{
+					mediabox.x0 = fz_atoi(fz_strsep(&p, " "));
+					mediabox.y0 = fz_atoi(fz_strsep(&p, " "));
+					mediabox.x1 = fz_atoi(fz_strsep(&p, " "));
+					mediabox.y1 = fz_atoi(fz_strsep(&p, " "));
+				}
+				else if (!strcmp(s, "%%Rotate"))
+				{
+					rotate = fz_atoi(fz_strsep(&p, " "));
+				}
+				else if (!strcmp(s, "%%Font"))
+				{
+					char *name = fz_strsep(&p, " ");
+					char *path = fz_strsep(&p, " ");
+					char *enc = fz_strsep(&p, " ");
+					if (!name || !path)
+						fz_throw(ctx, FZ_ERROR_GENERIC, "Font directive missing arguments");
+					add_font_res(resources, name, path, enc);
+				}
+				else if (!strcmp(s, "%%CJKFont"))
+				{
+					char *name = fz_strsep(&p, " ");
+					char *lang = fz_strsep(&p, " ");
+					char *wmode = fz_strsep(&p, " ");
+					char *style = fz_strsep(&p, " ");
+					if (!name || !lang)
+						fz_throw(ctx, FZ_ERROR_GENERIC, "CJKFont directive missing arguments");
+					add_cjkfont_res(resources, name, lang, wmode, style);
+				}
+				else if (!strcmp(s, "%%Image"))
+				{
+					char *name = fz_strsep(&p, " ");
+					char *path = fz_strsep(&p, " ");
+					if (!name || !path)
+						fz_throw(ctx, FZ_ERROR_GENERIC, "Image directive missing arguments");
+					add_image_res(resources, name, path);
+				}
 			}
-			else if (!strcmp(s, "%%Rotate"))
+			else
 			{
-				rotate = fz_atoi(fz_strsep(&p, " "));
-			}
-			else if (!strcmp(s, "%%Font"))
-			{
-				char *name = fz_strsep(&p, " ");
-				char *path = fz_strsep(&p, " ");
-				char *enc = fz_strsep(&p, " ");
-				if (!name || !path)
-					fz_throw(ctx, FZ_ERROR_GENERIC, "Font directive missing arguments");
-				add_font_res(resources, name, path, enc);
-			}
-			else if (!strcmp(s, "%%CJKFont"))
-			{
-				char *name = fz_strsep(&p, " ");
-				char *lang = fz_strsep(&p, " ");
-				char *wmode = fz_strsep(&p, " ");
-				char *style = fz_strsep(&p, " ");
-				if (!name || !lang)
-					fz_throw(ctx, FZ_ERROR_GENERIC, "CJKFont directive missing arguments");
-				add_cjkfont_res(resources, name, lang, wmode, style);
-			}
-			else if (!strcmp(s, "%%Image"))
-			{
-				char *name = fz_strsep(&p, " ");
-				char *path = fz_strsep(&p, " ");
-				if (!name || !path)
-					fz_throw(ctx, FZ_ERROR_GENERIC, "Image directive missing arguments");
-				add_image_res(resources, name, path);
+				fz_append_string(ctx, contents, line);
+				fz_append_byte(ctx, contents, '\n');
 			}
 		}
-		else
-		{
-			fz_append_string(ctx, contents, line);
-			fz_append_byte(ctx, contents, '\n');
-		}
+
+		page = pdf_add_page(ctx, doc, mediabox, rotate, resources, contents);
+		pdf_insert_page(ctx, doc, -1, page);
 	}
-	fz_drop_stream(ctx, stm);
-
-	page = pdf_add_page(ctx, doc, mediabox, rotate, resources, contents);
-	pdf_insert_page(ctx, doc, -1, page);
-	pdf_drop_obj(ctx, page);
-
-	fz_drop_buffer(ctx, contents);
-	pdf_drop_obj(ctx, resources);
+	fz_always(ctx)
+	{
+		fz_drop_stream(ctx, stm);
+		pdf_drop_obj(ctx, page);
+		fz_drop_buffer(ctx, contents);
+		pdf_drop_obj(ctx, resources);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 int pdfcreate_main(int argc, char **argv)
@@ -220,6 +254,7 @@ int pdfcreate_main(int argc, char **argv)
 	char *output = "out.pdf";
 	char *flags = "compress";
 	int i, c;
+	int error = 0;
 
 	while ((c = fz_getopt(argc, argv, "o:O:")) != -1)
 	{
@@ -227,12 +262,12 @@ int pdfcreate_main(int argc, char **argv)
 		{
 		case 'o': output = fz_optarg; break;
 		case 'O': flags = fz_optarg; break;
-		default: usage(); break;
+		default: return usage();
 		}
 	}
 
 	if (fz_optind == argc)
-		usage();
+		return usage();
 
 	ctx = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
 	if (!ctx)
@@ -243,16 +278,26 @@ int pdfcreate_main(int argc, char **argv)
 
 	pdf_parse_write_options(ctx, &opts, flags);
 
-	doc = pdf_create_document(ctx);
+	fz_var(doc);
 
-	for (i = fz_optind; i < argc; ++i)
-		create_page(argv[i]);
+	fz_try(ctx)
+	{
+		doc = pdf_create_document(ctx);
 
-	pdf_save_document(ctx, doc, output, &opts);
+		for (i = fz_optind; i < argc; ++i)
+			create_page(argv[i]);
 
-	pdf_drop_document(ctx, doc);
+		pdf_save_document(ctx, doc, output, &opts);
+	}
+	fz_always(ctx)
+		pdf_drop_document(ctx, doc);
+	fz_catch(ctx)
+	{
+		fz_log_error(ctx, fz_caught_message(ctx));
+		error = 1;
+	}
 
 	fz_flush_warnings(ctx);
 	fz_drop_context(ctx);
-	return 0;
+	return error;
 }
