@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "xps-imp.h"
@@ -546,7 +546,7 @@ xps_parse_glyphs(fz_context *ctx, xps_document *doc, fz_matrix ctm,
 	int is_sideways = 0;
 	int bidi_level = 0;
 
-	fz_text *text;
+	fz_text *text = NULL;
 	fz_rect area;
 
 	/*
@@ -614,6 +614,8 @@ xps_parse_glyphs(fz_context *ctx, xps_document *doc, fz_matrix ctm,
 	if (!font)
 		font = fz_new_base14_font(ctx, "Times-Roman");
 
+	fz_var(text);
+
 	fz_try(ctx)
 	{
 		/*
