@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "xps-imp.h"
@@ -246,7 +246,7 @@ xps_drop_document(fz_context *ctx, fz_document *doc_)
 }
 
 static int
-xps_lookup_metadata(fz_context *ctx, fz_document *doc_, const char *key, char *buf, int size)
+xps_lookup_metadata(fz_context *ctx, fz_document *doc_, const char *key, char *buf, size_t size)
 {
 	if (!strcmp(key, FZ_META_FORMAT))
 		return 1 + (int)fz_strlcpy(buf, "XPS", size);
@@ -259,7 +259,7 @@ xps_init_document(fz_context *ctx, xps_document *doc)
 	doc->super.refs = 1;
 	doc->super.drop_document = xps_drop_document;
 	doc->super.load_outline = xps_load_outline;
-	doc->super.resolve_link = xps_lookup_link_target;
+	doc->super.resolve_link_dest = xps_lookup_link_target;
 	doc->super.count_pages = xps_count_pages;
 	doc->super.load_page = xps_load_page;
 	doc->super.lookup_metadata = xps_lookup_metadata;
